# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 30.0 `specfun` Chapter.

``specfun`` - Approximations of Special Functions

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the provision of some commonly occurring
physical and mathematical functions.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/sintro.html
"""

# NAG Copyright 2017-2024.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import _nag_s_consts
from ..base import utils

def log_shifted(x):
    u"""
    ln(1+x).

    ``log_shifted`` returns a value of the shifted logarithmic function,
    ln(1+x).

    For full information please refer to the NAG Library document for
    s01ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s01baf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    lnx1 : float
        The value of the shifted logarithmic function, ln(1+x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > -1.0.

    Notes
    -----
    ``log_shifted`` computes values of ln(1+x), retaining full relative
    precision even when |x| is small.
    The function is based on the Chebyshev expansion

        ln((1+p^2+2px_)/(1+p^2-2px_)) = 4sum_k = 0^infinity
        p^(2k+1)/(2k+1)T_{2k+1}(x_).

    Setting x_= x(1+p^2)/(2p(x+2)), and choosing p = (q-1)/(q+1), q =
    fourthroot(2) the expansion is valid in the domain x in
    [1/(sqrt(2))-1,sqrt(2)-1].

    Outside this domain, ln(1+x) is computed by the standard logarithmic
    function.

    References
    ----------
    Lyusternik, L A, Chervonenkis, O A and Yanpolskii, A R, 1965,
    `Handbook for Computing Elementary Functions`, p. 57, Pergamon Press
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.log_shifted'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s01baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, lnx1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _lnx1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _lnx1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > -1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lnx1_py = utils._EngineFloat64ScalarType.to_py(_lnx1)
        return_data = _lnx1_py
    _err_class.manage()
    return return_data

def exp_complex(z):
    u"""
    Complex exponential, e^z.

    ``exp_complex`` evaluates the exponential function e^z, for
    `complex` z.

    For full information please refer to the NAG Library document for
    s01ea

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s01eaf.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    Returns
    -------
    ez : complex
        The value of the exponential function.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The imaginary part of argument `z` is so large that the
            result has no precision: `z` = (*<value>*,*<value>*).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Argument `z` causes overflow in real part of result: `z` =
            (*<value>*,*<value>*).

        (`errno` 2)
            Argument `z` causes overflow in imaginary part of result:
            `z` = (*<value>*,*<value>*).

        (`errno` 3)
            Argument `z` causes overflow in both real and imaginary
            parts of result: `z` = (*<value>*,*<value>*).

        (`errno` 4)
            The imaginary part of argument `z` is so large that the
            result is accurate to less than half precision: `z` =
            (*<value>*,*<value>*).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``exp_complex`` evaluates the exponential function e^z, taking care
    to avoid machine overflow, and giving a warning if the result cannot
    be computed to more than half precision.
    The function is evaluated as e^z = e^x(cos(y)+isin(y)), where x and
    y are the real and imaginary parts respectively of z.

    Since cos(y) and sin(y) are less than or equal to 1 in magnitude, it
    is possible that e^x may overflow although e^xcos(y) or e^xsin(y)
    does not.
    In this case the alternative formula sign(cos(y))e^(x+ln(|cos(y)|))
    is used for the real part of the result, and
    sign(sin(y))e^(x+ln(|sin(y)|)) for the imaginary part.
    If either part of the result still overflows, a warning is returned
    through argument errno.

    If Im(z) is too large, precision may be lost in the evaluation of
    sin(y) and cos(y).
    Again, a warning is returned through errno.
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.exp_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s01eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_COMPLEX128, # Output, ez.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _ez = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _ez,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Argument z causes overflow in real part of result:',
                    'z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Argument z causes overflow in imaginary part of result:',
                    'z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Argument z causes overflow in both real and imaginary parts of result:',
                    'z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The imaginary part of argument z is so large that the result is',
                    'accurate to less than half precision:',
                    'z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The imaginary part of argument z is so large that the result has no',
                    'precision: z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ez_py = utils._EngineComplex128ScalarType.to_py(_ez)
        return_data = _ez_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ez'],
            )(
                _ez_py,
            )
    _err_class.manage()
    return return_data

def tan(x):
    u"""
    tan(x).

    ``tan`` returns the value of the circular tangent, tan(x).

    For full information please refer to the NAG Library document for
    s07aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s07aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    tx : float
        The value of the circular tangent, tan(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>* and the constant F_1 = *<value>*.

            Constraint: abs(`x`) <= F_1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `x` = *<value>* and the constant F_2 = *<value>*.

            The function has been called with an argument that is too
            close to an odd multiple of pi/2, at which the function is
            infinite; the function has returned a value with the correct
            sign but a more or less arbitrary but large magnitude.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``tan`` calculates an approximate value for the circular tangent of
    its argument, tan(x).
    It is based on the Chebyshev expansion

        tan(theta) = theta y(t) = theta sum'_{r = 0}c_r T_r(t)

    where -pi/4 < theta < pi/4 and -1 < t < +1, t = 2(4 theta/pi)^2-1.

    The reduction to the standard range is accomplished by taking

        x = N pi/2+theta

    where N is an integer and -pi/4 < theta < pi/4,

    i.e., theta = x-(2x/pi)pi/2 where N = [2x/pi] = the nearest integer
    to 2x/pi.

    From the properties of tan(x) it follows that

        tan(x) = {[table omitted]}

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.tan'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s07aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, tx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _tx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _tx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'the constant F_1 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(x) <= F_1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'the constant F_2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The function has been called with an argument that is too close to',
                    'an odd multiple of pi/2, at which the function is infinite;',
                    'the function has returned a value with the correct sign but a more or less',
                    'arbitrary but large magnitude.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tx_py = utils._EngineFloat64ScalarType.to_py(_tx)
        return_data = _tx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['tx'],
            )(
                _tx_py,
            )
    _err_class.manage()
    return return_data

def arcsin(x):
    u"""
    arcsin(x).

    ``arcsin`` returns the value of the inverse circular sine,
    arcsin(x).
    The value is in the principal range (-pi/2,pi/2).

    For full information please refer to the NAG Library document for
    s09aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s09aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    asx : float
        The value of the inverse circular sine, arcsin(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``arcsin`` calculates an approximate value for the inverse circular
    sine, arcsin(x).
    It is based on the Chebyshev expansion

        arcsin(x) = x*y(x) = xsum'_{r = 0}a_r T_r(t)

    where -1/(sqrt(2)) <= x <= 1/(sqrt(2)) and t = 4x^2-1.

    For x^2 <= 1/2, arcsin(x) = x*y(x).

    For 1/2 < x^2 <= 1, arcsin(x) = sign(x){pi/2-arcsin(sqrt(1-x^2))}.

    For x^2 > 1, arcsin(x) is undefined and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.arcsin'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s09aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, asx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _asx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _asx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _asx_py = utils._EngineFloat64ScalarType.to_py(_asx)
        return_data = _asx_py
    _err_class.manage()
    return return_data

def arccos(x):
    u"""
    arccos(x).

    ``arccos`` returns the value of the inverse circular cosine,
    arccos(x); the result is in the principal range (0,pi).

    For full information please refer to the NAG Library document for
    s09ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s09abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    acx : float
        The value of the inverse circular cosine, arccos(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``arccos`` calculates an approximate value for the inverse circular
    cosine, arccos(x).
    It is based on the Chebyshev expansion

        arcsin(x) = x*y(t) = xsum'_{r = 0}a_r T_r(t)

    where (-1)/(sqrt(2)) <= x <= 1/(sqrt(2)), and t = 4x^2-1.

    For x^2 <= 1/2, arccos(x) = pi/2-arcsin(x).

    For -1 <= x < (-1)/(sqrt(2)), arccos(x) = pi-arcsin(sqrt(1-x^2)).

    For 1/(sqrt(2)) < x <= 1, arccos(x) = arcsin(sqrt(1-x^2)).

    For |x| > 1, arccos(x) is undefined and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.arccos'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s09abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, acx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _acx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _acx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _acx_py = utils._EngineFloat64ScalarType.to_py(_acx)
        return_data = _acx_py
    _err_class.manage()
    return return_data

def tanh(x):
    u"""
    tanh(x).

    ``tanh`` returns a value for the hyperbolic tangent, tanh(x).

    For full information please refer to the NAG Library document for
    s10aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s10aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    thx : float
        The value of the hyperbolic tangent, tanh(x).

    Notes
    -----
    ``tanh`` calculates an approximate value for the hyperbolic tangent
    of its argument, tanh(x).

    For |x| <= 1 it is based on the Chebyshev expansion

        tanh(x) = x*y(t) = xsum'_{r = 0}a_r T_r(t)

    where -1 <= x <= 1, -1 <= t <= 1, and t = 2x^2-1.

    For 1 < |x| < E_1

        tanh(x) = (e^(2x)-1)/(e^(2x)+1).

    For |x| >= E_1, tanh(x) = sign(x) to within the representation
    accuracy of the machine and so this approximation is used.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.tanh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s10aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, thx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _thx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _thx,
    )
    utils._handle_cb_exception(excs)
    _thx_py = utils._EngineFloat64ScalarType.to_py(_thx)
    return_data = _thx_py
    return return_data

def sinh(x):
    u"""
    sinh(x).

    ``sinh`` returns the value of the hyperbolic sine, sinh(x).

    For full information please refer to the NAG Library document for
    s10ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s10abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    shx : float
        The value of the hyperbolic sine, sinh(x).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= E_1.

            The function has been called with an argument too large in
            absolute magnitude. There is a danger of overflow. The
            result returned is the value of sinh(x) at the closest
            argument for which a valid call could be made.

    Notes
    -----
    ``sinh`` calculates an approximate value for the hyperbolic sine of
    its argument, sinh(x).

    For |x| <= 1 it uses the Chebyshev expansion

        sinh(x) = x*y(t) = xsum'_{r = 0}a_r T_r(t)

    where t = 2x^2-1.

    For 1 < |x| <= E_1, sinh(x) = 1/2(e^x-e^(-x))

    where E_1 is a machine-dependent constant.

    For |x| > E_1, the function fails owing to the danger of setting
    overflow in calculating e^x.
    The result returned for such calls is sinh(sign(x)E_1), i.e., it
    returns the result for the nearest valid argument.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.sinh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s10abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, shx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _shx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _shx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= E_1.',
                    'The function has been called with an argument too large in absolute',
                    'magnitude. There is a danger of overflow. The result returned is the value',
                    'of sinh(x) at the closest argument for which a valid call could be made.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _shx_py = utils._EngineFloat64ScalarType.to_py(_shx)
        return_data = _shx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['shx'],
            )(
                _shx_py,
            )
    _err_class.manage()
    return return_data

def cosh(x):
    u"""
    cosh(x).

    ``cosh`` returns the value of the hyperbolic cosine, cosh(x).

    For full information please refer to the NAG Library document for
    s10ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s10acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    chx : float
        The value of the hyperbolic cosine, cosh(x).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= E_1.

            The function has been called with an argument too large in
            absolute magnitude. There is a danger of overflow. The
            result returned is the value of cosh(x) at the nearest valid
            argument.

    Notes
    -----
    ``cosh`` calculates an approximate value for the hyperbolic cosine,
    cosh(x).

    For |x| <= E_1, cosh(x) = 1/2(e^x+e^(-x)).

    For |x| > E_1, the function fails owing to danger of setting
    overflow in calculating e^x.
    The result returned for such calls is cosh(E_1), i.e., it returns
    the result for the nearest valid argument.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.cosh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s10acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, chx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _chx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _chx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= E_1.',
                    'The function has been called with an argument too large in absolute',
                    'magnitude. There is a danger of overflow. The result returned is the value',
                    'of cosh(x) at the nearest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _chx_py = utils._EngineFloat64ScalarType.to_py(_chx)
        return_data = _chx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['chx'],
            )(
                _chx_py,
            )
    _err_class.manage()
    return return_data

def arctanh(x):
    u"""
    arctanh(x).

    ``arctanh`` returns the value of the inverse hyperbolic tangent,
    arctanh(x).

    For full information please refer to the NAG Library document for
    s11aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s11aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    athx : float
        The value of the inverse hyperbolic tangent, arctanh(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) < 1.

            The function has been called with an argument greater than
            or equal to 1.0 in magnitude, for which arctanh is not
            defined.

    Notes
    -----
    ``arctanh`` calculates an approximate value for the inverse
    hyperbolic tangent of its argument, arctanh(x).

    For x^2 <= 1/2 it is based on the Chebyshev expansion

        arctanh(x) = x*y(t) = xsum'_{r = 0}a_r T_r(t)

    where -1/(sqrt(2)) <= x <= 1/(sqrt(2)),  -1 <= t <= 1,  and t =
    4x^2-1.

    For 1/2 < x^2 < 1, it uses

        arctanh(x) = 1/2ln((1+x)/(1-x)).

    For |x| >= 1, the function fails as arctanh(x) is undefined.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.arctanh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s11aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, athx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _athx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _athx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) < 1.',
                    'The function has been called with an argument greater than or equal to',
                    '1.0 in magnitude, for which arctanh is not defined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _athx_py = utils._EngineFloat64ScalarType.to_py(_athx)
        return_data = _athx_py
    _err_class.manage()
    return return_data

def arcsinh(x):
    u"""
    arcsinh(x).

    ``arcsinh`` returns the value of the inverse hyperbolic sine,
    arcsinh(x).

    For full information please refer to the NAG Library document for
    s11ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s11abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    ashx : float
        The value of the inverse hyperbolic sine, arcsinh(x).

    Notes
    -----
    ``arcsinh`` calculates an approximate value for the inverse
    hyperbolic sine of its argument, arcsinh(x).

    For |x| <= 1 it is based on the Chebyshev expansion

        arcsinh(x) = x*y(t) = xsum'_{r = 0}c_r T_r(t), where t = 2x^2-1.

    For |x| > 1 it uses the fact that

        arcsinh(x) = sign(x)*ln(|x|+sqrt(x^2+1)).

    This form is used directly for 1 < |x| < 10^k, where k = n/2+1, and
    the machine uses approximately n decimal place arithmetic.

    For |x| >= 10^k, sqrt(x^2+1) is equal to |x| to within the accuracy
    of the machine and hence we can guard against premature overflow
    and, without loss of accuracy, calculate

        arcsinh(x) = sign(x)*(ln(2)+ln(|x|)).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.arcsinh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s11abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ashx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ashx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ashx,
    )
    utils._handle_cb_exception(excs)
    _ashx_py = utils._EngineFloat64ScalarType.to_py(_ashx)
    return_data = _ashx_py
    return return_data

def arccosh(x):
    u"""
    arccosh(x).

    ``arccosh`` returns the value of the inverse hyperbolic cosine,
    arccosh(x).
    The result is in the principal positive branch.

    For full information please refer to the NAG Library document for
    s11ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s11acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    achx : float
        The value of the inverse hyperbolic cosine, arccosh(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 1.0.

            The function has been called with an argument less than 1.0,
            for which arccosh(x) is not defined.

    Notes
    -----
    ``arccosh`` calculates an approximate value for the inverse
    hyperbolic cosine, arccosh(x).
    It is based on the relation

        arccosh(x) = ln(x+sqrt(x^2-1)).

    This form is used directly for 1 < x < 10^k, where k = n/2+1, and
    the machine uses approximately n decimal place arithmetic.

    For x >= 10^k, sqrt(x^2-1) is equal to sqrt(x) to within the
    accuracy of the machine and hence we can guard against premature
    overflow and, without loss of accuracy, calculate

        arccosh(x) = ln(2)+ln(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.arccosh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s11acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, achx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _achx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _achx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 1.0.',
                    'The function has been called with an argument less than 1.0, for which',
                    'arccosh(x) is not defined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _achx_py = utils._EngineFloat64ScalarType.to_py(_achx)
        return_data = _achx_py
    _err_class.manage()
    return return_data

def integral_exp(x):
    u"""
    Exponential integral E_1(x).

    ``integral_exp`` returns the value of the exponential integral
    E_1(x).

    For full information please refer to the NAG Library document for
    s13aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s13aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    e1x : float
        The value of the exponential integral E_1(x).

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `x` = *<value>* and the constant x_hi = *<value>*.
            The evaluation has been abandoned due to the likelihood of
            overflow.

            Constraint: `x` >= -x_hi.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `x` = 0.0 and the function is infinite.

    Notes
    -----
    ``integral_exp`` calculates an approximate value for

        E_1(x) = -Ei(-x) = int_x^infinity (e^(-u))/udu.

    using Chebyshev expansions, where x is real.
    For x < 0, the real part of the principal value of the integral is
    taken.
    The value E_1(0) is infinite, and so, when x = 0, ``integral_exp``
    exits with an error and returns the largest representable machine
    number.

    For 0 < x <= 4,

        E_1(x) = y(t)-ln(x) = sum'_r a_r T_r(t)-ln(x),

    where t = 1/2x-1.

    For x > 4,

        E_1(x) = (e^(-x))/xy(t) = (e^(-x))/xsum'_r a_r T_r(t),

    where t = -1.0+14.5/((x+3.25)) = (11.25-x)/(3.25+x).

    In both cases, -1 <= t <= +1.

    For x < 0, the approximation is based on expansions proposed by Cody
    and Thatcher Jr. (1969).
    Precautions are taken to maintain good relative accuracy in the
    vicinity of x_0≈-0.372507..., which corresponds to a simple zero of
    Ei(-x).

    ``integral_exp`` guards against producing underflows and overflows
    by using the argument x_hi.
    To guard against overflow, if x < -x_hi the function terminates and
    returns the negative of the largest representable machine number.
    To guard against underflow, if x > x_hi the result is set directly
    to zero.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J and Thatcher Jr., H C, 1969, `Rational Chebyshev
    approximations for the exponential integral Ei(x)`, Math. Comp.
    (23), 289--303
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.integral_exp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s13aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, e1x.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _e1x = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _e1x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = 0.0 and the function is infinite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'the constant x_hi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The evaluation has been abandoned due to the likelihood of overflow.',
                    'Constraint: x >= -x_hi.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _e1x_py = utils._EngineFloat64ScalarType.to_py(_e1x)
        return_data = _e1x_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['e1x'],
            )(
                _e1x_py,
            )
    _err_class.manage()
    return return_data

def integral_cos(x):
    u"""
    Cosine integral Ci(x).

    ``integral_cos`` returns the value of the cosine integral

        Ci(x) = gamma+ln(x)+int_0^x(cos(u)-1)/udu, x > 0

    where gamma denotes Euler's constant.

    For full information please refer to the NAG Library document for
    s13ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s13acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    cix : float
        The value of the cosine integral

            Ci(x) = gamma+ln(x)+int_0^x(cos(u)-1)/udu, x > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            The function has been called with an argument less than or
            equal to zero for which Ci(x) is not defined.

    Notes
    -----
    ``integral_cos`` calculates an approximate value for Ci(x).

    For 0 < x <= 16 it is based on the Chebyshev expansion

        Ci(x) = ln(x)+sum'_r = 0a_r T_r(t),t = 2(x/16)^2-1.

    For 16 < x < x_hi,

        Ci(x) = f(x)sin(x)/x-g(x)cos(x)/(x^2)

    where f(x) = sum'_{r = 0}f_r T_r(t) and g(x) = sum'_{r = 0}g_r
    T_r(t), t = 2(16/x)^2-1.

    For x >= x_hi, Ci(x) = 0 to within the accuracy possible (see
    [Accuracy]).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.integral_cos'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s13acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, cix.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _cix = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _cix,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'The function has been called with an argument less than or equal to zero',
                    'for which Ci(x) is not defined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _cix_py = utils._EngineFloat64ScalarType.to_py(_cix)
        return_data = _cix_py
    _err_class.manage()
    return return_data

def integral_sin(x):
    u"""
    Sine integral Si(x).

    ``integral_sin`` returns the value of the sine integral

        Si(x) = int_0^xsin(u)/udu,

    .

    For full information please refer to the NAG Library document for
    s13ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s13adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    six : float
        The value of the sine integral

            Si(x) = int_0^xsin(u)/udu.

    Notes
    -----
    ``integral_sin`` calculates an approximate value for Si(x).

    For |x| <= 16.0 it is based on the Chebyshev expansion

        Si(x) = xsum'_r = 0a_r T_r(t),t = 2(x/16)^2-1.

    For 16 < |x| < x_hi, where x_hi is an implementation-dependent
    number,

        Si(x) = sign(x){pi/2-f(x)cos(x)/x-g(x)sin(x)/(x^2)}

    where f(x) = sum'_{r = 0}f_r T_r(t) and g(x) = sum'_{r = 0}g_r
    T_r(t), t = 2(16/x)^2-1.

    For |x| >= x_hi, Si(x) = 1/2 pi sign(x) to within machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.integral_sin'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s13adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, six.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _six = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _six,
    )
    utils._handle_cb_exception(excs)
    _six_py = utils._EngineFloat64ScalarType.to_py(_six)
    return_data = _six_py
    return return_data

def gamma(x):
    u"""
    Gamma function.

    ``gamma`` returns the value of the gamma function Gamma(x).

    For full information please refer to the NAG Library document for
    s14aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    gx : float
        The value of the gamma function Gamma(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            The argument is too large, the function returns the
            approximate value of Gamma(x) at the nearest valid argument.

        (`errno` 2)
            On entry, `x` = *<value>*. The function returns zero.

            Constraint: `x` >= *<value>*.

            The argument is too large and negative, the function returns
            zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) >= *<value>*.

            The argument is too close to zero, the function returns the
            approximate value of Gamma(x) at the nearest valid argument.

        (`errno` 4)
            On entry, `x` = *<value>*.

            Constraint: `x` must not be a negative integer.

            The argument is a negative integer, at which values Gamma(x)
            is infinite. The function returns a large positive value.

    Notes
    -----
    ``gamma`` evaluates an approximation to the gamma function Gamma(x).
    The function is based on the Chebyshev expansion:

        Gamma(1+u) = sum'_r = 0a_r T_r(t)

    where 0 <= u < 1,t = 2u-1, and uses the property Gamma(1+x) = x
    Gamma(x).
    If x = N+1+u where N is integral and 0 <= u < 1 then it follows
    that:

    [table omitted]

    There are four possible failures for this function:

    (i) if x is too large, there is a danger of overflow since Gamma(x)
        could become too large to be represented in the machine;

    (#) if x is too large and negative, there is a danger of underflow;

    (#) if x is equal to a negative integer, Gamma(x) would overflow
        since it has poles at such points;

    (#) if x is too near zero, there is again the danger of overflow on
        some machines. For small x, Gamma(x) ~= 1/x, and on some
        machines there exists a range of nonzero but small values of x
        for which 1/x is larger than the greatest representable value.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, gx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _gx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _gx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The argument is too large, the function returns the approximate value of',
                    'Gamma(x) at the nearest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'The function returns zero.',
                    'Constraint: x >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The argument is too large and negative, the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The argument is too close to zero, the function returns the approximate',
                    'value of Gamma(x) at the nearest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x must not be a negative integer.',
                    'The argument is a negative integer, at which values Gamma(x) is infinite.',
                    'The function returns a large positive value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _gx_py = utils._EngineFloat64ScalarType.to_py(_gx)
        return_data = _gx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['gx'],
            )(
                _gx_py,
            )
    _err_class.manage()
    return return_data

def gamma_log_real(x):
    u"""
    Log gamma function, real argument.

    ``gamma_log_real`` returns the value of the logarithm of the gamma
    function, ln(Gamma)(x).

    For full information please refer to the NAG Library document for
    s14ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    lgx : float
        The value of ln(Gamma)(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `x` = *<value>* and the constant x_big =
            *<value>*.

            Constraint: `x` <= x_big.

    Notes
    -----
    ``gamma_log_real`` calculates an approximate value for ln(Gamma)(x).
    It is based on rational Chebyshev expansions.

    Denote by R_n,m^i(x) = P_n^i(x)/Q_m^i(x) a ratio of polynomials of
    degree n in the numerator and m in the denominator.
    Then:

        for 0 < x <= 1/2,

                ln(Gamma)(x)≈-ln(x)+xR_n,m^1(x+1);

        for 1/2 < x <= 3/2,

                ln(Gamma)(x)≈(x-1)R_n,m^1(x);

        for 3/2 < x <= 4,

                ln(Gamma)(x)≈(x-2)R_n,m^2(x);

        for 4 < x <= 12,

                ln(Gamma)(x)≈R_n,m^3(x);

        and for 12 < x,

ln(Gamma)(x)≈(x-1/2)ln(x)-x+ln(sqrt(2π))+1/xR_n,m^4(1/x^2).

    For each expansion, the specific values of n and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    Let epsilon denote machine precision and let x_huge denote the
    largest positive model number (see ``machine.real_largest``).
    For x < 0.0 the value ln(Gamma)(x) is not defined;
    ``gamma_log_real`` returns zero and exits with `errno` = 1.
    It also exits with `errno` = 1 when x = 0.0, and in this case the
    value x_huge is returned.
    For x in the interval (0.0,epsilon], the function ln(Gamma)(x) =
    -ln(x) to machine accuracy.

    Now denote by x_big the largest allowable argument for ln(Gamma)(x)
    on the machine.
    For (x_big)^(1/4) < x <= x_big the R_n,m^4(1/x^2) term in Equation
    (1) is negligible.
    For x > x_big there is a danger of setting overflow, and so
    ``gamma_log_real`` exits with `errno` = 2 and returns x_huge.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J and Hillstrom, K E, 1967, `Chebyshev approximations for
    the natural logarithm of the gamma function`, Math.Comp. (21),
    198--203
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_log_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, lgx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _lgx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _lgx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1021:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'the constant x_big = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x <= x_big.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _lgx_py = utils._EngineFloat64ScalarType.to_py(_lgx)
        return_data = _lgx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['lgx'],
            )(
                _lgx_py,
            )
    _err_class.manage()
    return return_data

def polygamma(x):
    u"""
    psi(x)-ln(x).

    ``polygamma`` returns a value of the function psi(x)-ln(x), where
    psi is the psi function psi(x) = d/dxln(Gamma)(x) =
    Gamma'(x)/Gamma(x).

    For full information please refer to the NAG Library document for
    s14ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    pgx : float
        The value of the function psi(x)-ln(x), where psi is the psi
        function psi(x) = d/dxln(Gamma)(x) = Gamma'(x)/Gamma(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

        (`errno` 2)
            Computation halted due to likelihood of underflow. `x` may
            be too large. `x` = *<value>*.

        (`errno` 3)
            Computation halted due to likelihood of overflow. `x` may be
            too small. `x` = *<value>*.

    Notes
    -----
    ``polygamma`` returns a value of the function psi(x)-ln(x).
    The psi function is computed without the logarithmic term so that
    when x is large, sums or differences of psi functions may be
    computed without unnecessary loss of precision, by analytically
    combining the logarithmic terms.
    For example, the difference d = psi(x+1/2)-psi(x) has an asymptotic
    behaviour for large x given by
    d~ln(x+1/2)-ln(x)+O(1/(x^2))~ln(1+1/2x)~1/2x.

    Computing d directly would amount to subtracting two large numbers
    which are close to ln(x+1/2) and ln(x) to produce a small number
    close to 1/2x, resulting in a loss of significant digits.
    However, using this function to compute f(x) = psi(x)-ln(x), we can
    compute d = f(x+1/2)-f(x)+ln(1+1/2x), and the dominant logarithmic
    term may be computed accurately from its power series when x is
    large.
    Thus we avoid the unnecessary loss of precision.

    The function is derived from the function PSIFN in Amos (1983).

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1983, `Algorithm 610: A portable FORTRAN subroutine for
    derivatives of the psi function`, ACM Trans. Math. Software (9),
    494--502
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.polygamma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, pgx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _pgx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _pgx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation halted due to likelihood of underflow. x may be',
                    'too large. x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation halted due to likelihood of overflow. x may be',
                    'too small. x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _pgx_py = utils._EngineFloat64ScalarType.to_py(_pgx)
        return_data = _pgx_py
    _err_class.manage()
    return return_data

def polygamma_deriv(x, n, m, ans):
    u"""
    Scaled derivatives of psi(x).

    ``polygamma_deriv`` returns a sequence of values of scaled
    derivatives of the psi function psi(x) (also known as the digamma
    function).

    For full information please refer to the NAG Library document for
    s14ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    n : int
        The index of the first member n of the sequence of functions.

    m : int
        The number of members m required in the sequence w(k,x), for k =
        n,...,n+m-1.

    ans : float, shape (`m`), modified in place
        `On exit`: the first m elements of `ans` contain the required
        values w(k,x), for k = n,...,n+m-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 4)
            Computation abandoned due to the likelihood of underflow.

        (`errno` 5)
            Computation abandoned due to the likelihood of overflow.

        (`errno` 6)
            There is not enough internal workspace to continue
            computation. `m` is probably too large.

    Notes
    -----
    ``polygamma_deriv`` computes m values of the function

        w(k,x) = (-1)^(k+1)psi^(k)(x)/k!,

    for x > 0, k = n, n+1,...,n+m-1, where psi is the psi function

        psi(x) = d/dxln(Gamma)(x) = Gamma'(x)/Gamma(x),

    and psi^(k) denotes the kth derivative of psi.

    The function is derived from the function PSIFN in Amos (1983).
    The basic method of evaluation of w(k,x) is the asymptotic series

        w(k,x)~epsilon(k,x)+1/2x^(k+1)+1/(x^k)sum_j = 1^infinity
        B_{2j}((2j+k-1)!)/((2j)!k!x^(2j))

    for large x greater than a machine-dependent value x_min, followed
    by backward recurrence using

        w(k,x) = w(k,x+1)+x^(-k-1)

    for smaller values of x, where epsilon(k,x) = -ln(x) when k = 0,
    epsilon(k,x) = 1/kx^k when k > 0, and B_{2j}, j = 1,2,..., are the
    Bernoulli numbers.

    When k is large, the above procedure may be inefficient, and the
    expansion

        w(k,x) = sum_j = 1^infinity 1/((x+j)^(k+1)),

    which converges rapidly for large k, is used instead.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1983, `Algorithm 610: A portable FORTRAN subroutine for
    derivatives of the psi function`, ACM Trans. Math. Software (9),
    494--502
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.polygamma_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Output, ans[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _ans_cls = utils._EngineFloat64ArrayType(
        ans, {'fun_name': fun_name, 'entity': 'ans'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _ans = _ans_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _n,
        _m,
        _ans,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation abandoned due to the likelihood of underflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation abandoned due to the likelihood of overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There is not enough internal workspace to continue computation.',
                    'm is probably too large.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ans_cls.refresh_py(_ans)
        return_data = None
    _err_class.manage()
    return return_data

def psi_deriv_real(x, k):
    u"""
    Polygamma function psi^(n)(x) for real x.

    ``psi_deriv_real`` returns the value of the kth derivative of the
    psi function psi(x) for real x and k = 0,1,...,6.

    For full information please refer to the NAG Library document for
    s14ae

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14aef.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    k : int
        The function psi^(k)(x) to be evaluated.

    Returns
    -------
    pkx : float
        The value of the kth derivative of the psi function psi(x) for
        real x and k = 0,1,...,6.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` is 'too close' to a non-positive integer: `x`
            = *<value>* and nint(`x`) = *<value>*.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` <= 6.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            Evaluation abandoned due to likelihood of underflow.

        (`errno` 3)
            Evaluation abandoned due to likelihood of overflow.

    Notes
    -----
    ``psi_deriv_real`` evaluates an approximation to the kth derivative
    of the psi function psi(x) given by

        psi^(k)(x) = (d^k)/dx^k psi(x) = (d^k)/dx^k(d/dxloge(Gamma(x))),

    where x is real with x != 0,-1,-2,... and k = 0,1,...,6.
    For negative noninteger values of x, the recurrence relationship

        psi^(k)(x+1) = psi^(k)(x)+(d^k)/dx^k(1/x)

    is used.
    The value of (-1)^(k+1)psi^(k)(x)/k! is obtained by a call to
    ``polygamma_deriv``, which is based on the function PSIFN in Amos
    (1983).

    Note that psi^(k)(x) is also known as the `polygamma` function.
    Specifically, psi^(0)(x) is often referred to as the `digamma`
    function and psi^(1)(x) as the `trigamma` function in the
    literature.
    Further details can be found in Abramowitz and Stegun (1972).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Amos, D E, 1983, `Algorithm 610: A portable FORTRAN subroutine for
    derivatives of the psi function`, ACM Trans. Math. Software (9),
    494--502
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.psi_deriv_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Output, pkx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _pkx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _k,
        _pkx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x is \'too close\' to a non-positive integer:',
                    'x = {:s} and nint(x) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Evaluation abandoned due to likelihood of underflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Evaluation abandoned due to likelihood of overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _pkx_py = utils._EngineFloat64ScalarType.to_py(_pkx)
        return_data = _pkx_py
    _err_class.manage()
    return return_data

def psi_deriv_complex(z, k):
    u"""
    Polygamma function psi^(n)(z) for complex z.

    ``psi_deriv_complex`` returns the value of the kth derivative of the
    psi function psi(z) for complex z and k = 0,1,...,4.

    For full information please refer to the NAG Library document for
    s14af

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14aff.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    k : int
        The function psi^(k)(z) to be evaluated.

    Returns
    -------
    pkz : complex
        The value of the kth derivative of the psi function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, Re(`z`) is 'too close' to a non-positive integer
            when Im(`z`) = 0.0: Re(`z`) = *<value>*, nint(Re(`z`)) =
            *<value>*.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` <= 4.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            Evaluation abandoned due to likelihood of overflow.

    Notes
    -----
    ``psi_deriv_complex`` evaluates an approximation to the kth
    derivative of the psi function psi(z) given by

        psi^(k)(z) = (d^k)/dz^k psi(z) = (d^k)/dz^k(d/dzloge(Gamma(z))),

    where z = x+iy is complex provided y != 0 and k = 0,1,...,4.
    If y = 0, z is real and thus psi^(k)(z) is singular when z =
    0,-1,-2,....

    Note that psi^(k)(z) is also known as the `polygamma` function.
    Specifically, psi^(0)(z) is often referred to as the `digamma`
    function and psi^(1)(z) as the `trigamma` function in the
    literature.
    Further details can be found in Abramowitz and Stegun (1972).

    ``psi_deriv_complex`` is based on a modification of the method
    proposed by Kölbig (1972).

    To obtain the value of psi^(k)(z) when z is real, ``psi_deriv_real``
    can be used.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Kölbig, K S, 1972, `Programs for computing the logarithm of the
    gamma function, and the digamma function, for complex arguments`,
    Comp. Phys. Comm. (4), 221--226
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.psi_deriv_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Output, pkz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _pkz = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _k,
        _pkz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, Re(z) is \'too close\' to a non-positive integer when',
                    'Im(z) = 0.0: Re(z) = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'nint(Re(z)) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k <= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Evaluation abandoned due to likelihood of overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _pkz_py = utils._EngineComplex128ScalarType.to_py(_pkz)
        return_data = _pkz_py
    _err_class.manage()
    return return_data

def gamma_log_complex(z):
    u"""
    Logarithm of the gamma function ln(Gamma)(z), complex argument.

    ``gamma_log_complex`` returns the value of the logarithm of the
    gamma function ln(Gamma)(z) for complex z, .

    For full information please refer to the NAG Library document for
    s14ag

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14agf.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    Returns
    -------
    lngz : complex
        The value of ln(Gamma)(z).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, Re(`z`) is 'too close' to a non-positive integer
            when Im(`z`) = 0.0.

    Notes
    -----
    ``gamma_log_complex`` evaluates an approximation to the logarithm of
    the gamma function ln(Gamma)(z) defined for Re(z) > 0 by

        ln(Gamma)(z) = ln(int_0^infinity)e^(-t)t^(z-1)dt

    where z = x+iy is complex.
    It is extended to the rest of the complex plane by analytic
    continuation unless y = 0, in which case z is real and each of the
    points z = 0,-1,-2,... is a singularity and a branch point.

    ``gamma_log_complex`` is based on the method proposed by Kölbig
    (1972) in which the value of ln(Gamma)(z) is computed in the
    different regions of the z plane by means of the formulae

        [table omitted]

    where n = [x_0]-[x], {B_{2k}} are Bernoulli numbers (see Abramowitz
    and Stegun (1972)) and [x] is the largest integer <= x.
    Note that care is taken to ensure that the imaginary part is
    computed correctly, and not merely modulo 2 pi.

    The function uses the values K = 10 and x_0 = 7.
    The remainder term R_K(z) is discussed in [Accuracy].

    To obtain the value of ln(Gamma)(z) when z is real and positive,
    ``gamma_log_real`` can be used.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Kölbig, K S, 1972, `Programs for computing the logarithm of the
    gamma function, and the digamma function, for complex arguments`,
    Comp. Phys. Comm. (4), 221--226
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_log_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_COMPLEX128, # Output, lngz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _lngz = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _lngz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, Re(z) is \'too close\' to a non-positive integer when Im(z) = 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lngz_py = utils._EngineComplex128ScalarType.to_py(_lngz)
        return_data = _lngz_py
    _err_class.manage()
    return return_data

def gamma_log_scaled_real(x):
    u"""
    Scaled log gamma function.

    ``gamma_log_scaled_real`` returns the value of ln(G(x)), the scaled
    logarithm of the gamma function Gamma(x).

    For full information please refer to the NAG Library document for
    s14ah

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14ahf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    slgx : float
        The value of ln(G(x)), the scaled logarithm of the gamma
        function Gamma(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` <= 0.0. On failure, the function value
            returned is zero.

    Notes
    -----
    ``gamma_log_scaled_real`` calculates an approximate value for
    ln(G(x)), where G(x) = Gamma(x+1)/(x/e)^x.
    This is a variant of the ln(Gamma(x)) function (see also
    ``gamma_log_real``), which avoids rounding problems for very large
    arguments by computing ln(Gamma(x)) with the Stirling approximation
    factored out.

    For 0 < x < 15, ln(G(x)) = ln(Gamma(x+1))-x*ln(x)+x;

    and for 15 <= x, ln(G(x)) = 1/2ln(x)+ln(sqrt(2 pi))+1/xR(1/x^2),
    where R is a suitable Remez approximation.

    For x <= 0.0, the value ln(G(x)) is undefined;
    ``gamma_log_scaled_real`` returns zero and exits with `errno` = 1.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_log_scaled_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, slgx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _slgx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _slgx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x <= 0.0. On failure, the function value returned is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _slgx_py = utils._EngineFloat64ScalarType.to_py(_slgx)
        return_data = _slgx_py
    _err_class.manage()
    return return_data

def gamma_vector(n, x, f, ivalid):
    u"""
    Gamma function, vectorized Gamma(x).

    ``gamma_vector`` returns an array of values of the gamma function
    Gamma(x).

    For full information please refer to the NAG Library document for
    s14an

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14anf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Gamma(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. `f`[i-1] contains the
            approximate value of Gamma(x_i) at the nearest valid
            argument. The threshold value is the same as for `errno` = 1
            in ``gamma``.

        `ivalid`[i-1] = 2

            x_i is too large and negative. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 2 in ``gamma``.

        `ivalid`[i-1] = 3

            x_i is too close to zero. `f`[i-1] contains the approximate
            value of Gamma(x_i) at the nearest valid argument. The
            threshold value is the same as for `errno` = 2 in ``gamma``.

        `ivalid`[i-1] = 4

            x_i is a negative integer, at which values Gamma(x_i) are
            infinite. `f`[i-1] contains a large positive value.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``gamma_vector`` evaluates an approximation to the gamma function
    Gamma(x) for an array of arguments x_i, for i = 1,2,...,n.
    The function is based on the Chebyshev expansion:

        Gamma(1+u) = sum'_r = 0a_r T_r(t)

    where 0 <= u < 1,t = 2u-1, and uses the property Gamma(1+x) = x
    Gamma(x).
    If x = N+1+u where N is integral and 0 <= u < 1 then it follows
    that:

    [table omitted]

    There are four possible failures for this function:

    (i) if x is too large, there is a danger of overflow since Gamma(x)
        could become too large to be represented in the machine;

    (#) if x is too large and negative, there is a danger of underflow;

    (#) if x is equal to a negative integer, Gamma(x) would overflow
        since it has poles at such points;

    (#) if x is too near zero, there is again the danger of overflow on
        some machines. For small x, Gamma(x) ~= 1/x, and on some
        machines there exists a range of nonzero but small values of x
        for which 1/x is larger than the greatest representable value.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14anft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def gamma_log_real_vector(n, x, f, ivalid):
    u"""
    Log gamma function, vectorized ln(Gamma(x)).

    ``gamma_log_real_vector`` returns an array of values of the
    logarithm of the gamma function, ln(Gamma)(x).

    For full information please refer to the NAG Library document for
    s14ap

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14apf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: ln(Gamma)(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i <= 0.

        `ivalid`[i-1] = 2

            x_i is too large and positive. The threshold value is the
            same as for `errno` = 2 in ``gamma_log_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``gamma_log_real_vector`` calculates an approximate value for
    ln(Gamma)(x) for an array of arguments x_i, for i = 1,2,...,n.
    It is based on rational Chebyshev expansions.

    Denote by R_n,m^i(x) = P_n^i(x)/Q_m^i(x) a ratio of polynomials of
    degree n in the numerator and m in the denominator.
    Then:

        for 0 < x <= 1/2,

                ln(Gamma)(x)≈-ln(x)+xR_n,m^1(x+1);

        for 1/2 < x <= 3/2,

                ln(Gamma)(x)≈(x-1)R_n,m^1(x);

        for 3/2 < x <= 4,

                ln(Gamma)(x)≈(x-2)R_n,m^2(x);

        for 4 < x <= 12,

                ln(Gamma)(x)≈R_n,m^3(x);

        and for 12 < x,

ln(Gamma)(x)≈(x-1/2)ln(x)-x+ln(sqrt(2π))+1/xR_n,m^4(1/x^2).

    For each expansion, the specific values of n and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    Let epsilon denote machine precision and let x_huge denote the
    largest positive model number (see ``machine.real_largest``).
    For x < 0.0 the value ln(Gamma)(x) is not defined;
    ``gamma_log_real_vector`` returns zero and exits with `errno` = 1.
    It also exits with `errno` = 1 when x = 0.0, and in this case the
    value x_huge is returned.
    For x in the interval (0.0,epsilon], the function ln(Gamma)(x) =
    -ln(x) to machine accuracy.

    Now denote by x_big the largest allowable argument for ln(Gamma)(x)
    on the machine.
    For (x_big)^(1/4) < x <= x_big the R_n,m^4(1/x^2) term in Equation
    (1) is negligible.
    For x > x_big there is a danger of setting overflow, and so
    ``gamma_log_real_vector`` exits with `errno` = 2 and returns x_huge.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J and Hillstrom, K E, 1967, `Chebyshev approximations for
    the natural logarithm of the gamma function`, Math.Comp. (21),
    198--203
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_log_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14apft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def gamma_incomplete(a, x, tol):
    u"""
    Incomplete gamma functions P(a,x) and Q(a,x).

    ``gamma_incomplete`` computes values for the incomplete gamma
    functions P(a,x) and Q(a,x).

    For full information please refer to the NAG Library document for
    s14ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14baf.html

    Parameters
    ----------
    a : float
        The argument a of the functions.

    x : float
        The argument x of the functions.

    tol : float
        The relative accuracy required by you in the results. If
        ``gamma_incomplete`` is entered with `tol` greater than 1.0 or
        less than machine precision, then the value of machine precision
        is used instead.

    Returns
    -------
    p : float
        The value of P(a,x)

    q : float
        The value of P(a,x)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 3)
            Algorithm fails to terminate in *<value>* iterations.

    Notes
    -----
    ``gamma_incomplete`` evaluates the incomplete gamma functions in the
    normalized form

        P(a,x) = 1/Gamma(a)int_0^xt^(a-1)e^(-t)dt,

        Q(a,x) = 1/Gamma(a)int_x^infinity t^(a-1)e^(-t)dt,

    with x >= 0 and a > 0, to a user-specified accuracy.
    With this normalization, P(a,x)+Q(a,x) = 1.

    Several methods are used to evaluate the functions depending on the
    arguments a and x, the methods including Taylor expansion for
    P(a,x), Legendre's continued fraction for Q(a,x), and power series
    for Q(a,x).
    When both a and x are large, and a ~= x, the uniform asymptotic
    expansion of Temme (1987) is employed for greater efficiency --
    specifically, this expansion is used when a >= 20 and 0.7a <= x <=
    1.4a.

    Once either P or Q is computed, the other is obtained by subtraction
    from 1.
    In order to avoid loss of relative precision in this subtraction,
    the smaller of P and Q is computed first.

    This function is derived from the function GAM in Gautschi (1979b).

    References
    ----------
    Gautschi, W, 1979, `A computational procedure for incomplete gamma
    functions`, ACM Trans. Math. Software (5), 466--481

    Gautschi, W, 1979, `Algorithm 542: Incomplete gamma functions`, ACM
    Trans. Math. Software (5), 482--489

    Temme, N M, 1987, `On the computation of the incomplete gamma
    functions for large values of the parameters`, Algorithms for
    Approximation, (eds J C Mason and M G Cox), Oxford University Press
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_incomplete'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Output, p.
        utils._BYREF_FLOAT64, # Output, q.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    _q = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _x,
        _tol,
        _p,
        _q,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithm fails to terminate in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        _q_py = utils._EngineFloat64ScalarType.to_py(_q)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['p', 'q'],
        )(
            _p_py,
            _q_py,
        )
    _err_class.manage()
    return return_data

def gamma_incomplete_vector(n, a, x, tol, p, q, ivalid):
    u"""
    Incomplete gamma functions, vectorized P(a,x) and Q(a,x).

    ``gamma_incomplete_vector`` computes an array of values for the
    incomplete gamma functions P(a,x) and Q(a,x).

    For full information please refer to the NAG Library document for
    s14bn

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14bnf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    a : float, shape (`n`)
        The argument a_i of the function, for i = 1,2,...,`n`.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    tol : float
        The relative accuracy required by you in the results. If
        ``gamma_incomplete_vector`` is entered with `tol` greater than
        1.0 or less than machine precision, then the value of machine
        precision is used instead.

    p : float, shape (`n`), modified in place
        `On exit`: P(a_i,x_i), the function values.

    q : float, shape (`n`), modified in place
        `On exit`: Q(a_i,x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for a_i and
        x_i, for i = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            a_i <= 0.

        `ivalid`[i-1] = 2

            x_i < 0.

        `ivalid`[i-1] = 3

            Algorithm fails to terminate.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``gamma_incomplete_vector`` evaluates the incomplete gamma functions
    in the normalized form, for an array of arguments a_i,x_i, for i =
    1,2,...,n.

        P(a,x) = 1/Gamma(a)int_0^xt^(a-1)e^(-t)dt,

        Q(a,x) = 1/Gamma(a)int_x^infinity t^(a-1)e^(-t)dt,

    with x >= 0 and a > 0, to a user-specified accuracy.
    With this normalization, P(a,x)+Q(a,x) = 1.

    Several methods are used to evaluate the functions depending on the
    arguments a and x, the methods including Taylor expansion for
    P(a,x), Legendre's continued fraction for Q(a,x), and power series
    for Q(a,x).
    When both a and x are large, and a ~= x, the uniform asymptotic
    expansion of Temme (1987) is employed for greater efficiency --
    specifically, this expansion is used when a >= 20 and 0.7a <= x <=
    1.4a.

    Once either P or Q is computed, the other is obtained by subtraction
    from 1.
    In order to avoid loss of relative precision in this subtraction,
    the smaller of P and Q is computed first.

    This function is derived from the function GAM in Gautschi (1979b).

    References
    ----------
    Gautschi, W, 1979, `A computational procedure for incomplete gamma
    functions`, ACM Trans. Math. Software (5), 466--481

    Gautschi, W, 1979, `Algorithm 542: Incomplete gamma functions`, ACM
    Trans. Math. Software (5), 482--489

    Temme, N M, 1987, `On the computation of the incomplete gamma
    functions for large values of the parameters`, Algorithms for
    Approximation, (eds J C Mason and M G Cox), Oxford University Press
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.gamma_incomplete_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14bnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[n].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Output, p[n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _x,
        _tol,
        _p,
        _q,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        _q_cls.refresh_py(_q)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def beta_log_real(a, b):
    u"""
    Logarithm of the beta function ln(B)(a,b).

    ``beta_log_real`` returns the value of the logarithm of the beta
    function, ln(B)(a,b).

    For full information please refer to the NAG Library document for
    s14cb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14cbf.html

    Parameters
    ----------
    a : float
        The argument a of the function.

    b : float
        The argument b of the function.

    Returns
    -------
    lbab : float
        The value of the logarithm of the beta function, ln(B)(a,b).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

    Notes
    -----
    ``beta_log_real`` calculates values for ln(B)(a,b) where B is the
    beta function given by

        B(a,b) = int_0^1t^(a-1)(1-t)^(b-1)dt

    or equivalently

        B(a,b) = Gamma(a)Gamma(b)/Gamma(a+b)

    and Gamma(x) is the gamma function.
    Note that the beta function is symmetric, so that B(a,b) = B(b,a).

    In order to efficiently obtain accurate results several methods are
    used depending on the parameters a and b.

    Let a_0 = min(a,b) and b_0 = max(a,b).
    Then:

    for a_0 >= 8,

        ln(B) = 0.5 ln((2 pi)-0.5) ln(b_0) + Delta(a_0) + Delta (b_0) -
        Delta (a_0+b_0) - u - v ;

    where

        Delta(a_0) = ln(Gamma(a_0))-(a_0-0.5)ln(a_0)+a_0-0.5ln(2 pi),

        u = -(a_0-0.5)ln[(a_0)/(a_0+b_0)]  and

        v = b_0 ln(1+(a_0)/(b_0)).

    for a_0 < 1,

        for b_0 >= 8,

                ln(B) = ln(Gamma)(a_0)+ln(Gamma(b_0)/Gamma(a_0+b_0));

        for b_0 < 8,

                ln(B) =
                ln(Gamma)(a_0)+ln(Gamma)(b_0)-ln(Gamma)(a_0+b_0);

    for 2 < a_0 < 8, a_0 is reduced to the interval [1,2] by B(a,b) =
    (a_0-1)/(a_0+b_0-1)B(a_0-1,b_0);

    for 1 <= a_0 <= 2,

        for b_0 >= 8,

                ln(B) = ln(Gamma)(a_0)+ln(Gamma(b_0)/Gamma(a_0+b_0));

        for 2 < b_0 < 8, b_0 is reduced to the interval [1,2];

        for b_0 <= 2,

                ln(B) =
                ln(Gamma)(a_0)+ln(Gamma)(b_0)-ln(Gamma)(a_0+b_0).

    ``beta_log_real`` is derived from BETALN in DiDonato and Morris
    (1992).

    References
    ----------
    DiDonato, A R and Morris, A H, 1992, `Algorithm 708: Significant
    digit computation of the incomplete beta function ratios`, ACM
    Trans. Math. Software (18), 360--373
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.beta_log_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Output, lbab.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _lbab = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _lbab,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lbab_py = utils._EngineFloat64ScalarType.to_py(_lbab)
        return_data = _lbab_py
    _err_class.manage()
    return return_data

def beta_incomplete(a, b, x):
    u"""
    Regularized incomplete beta function I_x(a,b) and its complement
    1-I_x.

    ``beta_incomplete`` computes values for the regularized incomplete
    beta function I_x(a,b) and its complement 1-I_x(a,b).

    For full information please refer to the NAG Library document for
    s14cc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14ccf.html

    Parameters
    ----------
    a : float
        The argument a of the function.

    b : float
        The argument b of the function.

    x : float
        x, upper limit of integration.

    Returns
    -------
    w : float
        The value of the incomplete beta function I_x(a,b) evaluated
        from zero to x.

    w1 : float
        The value of the complement of the incomplete beta function
        1-I_x(a,b), i.e., the incomplete beta function evaluated from x
        to one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `b` = *<value>*.

            Constraint: `b` >= 0.0.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` >= 0.0.

        (`errno` 2)
            On entry, `a` and `b` were zero.

            Constraint: `a` or `b` must be nonzero.

        (`errno` 3)
            On entry, `x` = *<value>*.

            Constraint: 0.0 <= `x` <= 1.0.

        (`errno` 4)
            On entry, `x` and `a` were zero.

            Constraint: `x` or `a` must be nonzero.

        (`errno` 5)
            On entry, 1.0-`x` and `b` were zero.

            Constraint: 1.0-`x` or `b` must be nonzero.

    Notes
    -----
    ``beta_incomplete`` evaluates the regularized incomplete beta
    function and its complement in the normalized form

        [table omitted]

    with

        0 <= x <= 1,

        a >= 0 and b >= 0,

        and the beta function B(a,b) is defined as B(a,b) =
        int_0^1t^(a-1)(1-t)^(b-1)dt = Gamma(a)Gamma(b)/Gamma(a+b) where
        Gamma(y) is the gamma function.

    Several methods are used to evaluate the functions depending on the
    arguments a, b and x.
    The methods include Wise's asymptotic expansion (see Wise (1950))
    when a > b, continued fraction derived by DiDonato and Morris (1992)
    when a, b > 1, and power series when b <= 1 or b*x <= 0.7.
    When both a and b are large, specifically a, b >= 15, the DiDonato
    and Morris (1992) asymptotic expansion is employed for greater
    efficiency.

    Once either I_x(a,b) or I_y(b,a) is computed, the other is obtained
    by subtraction from 1.
    In order to avoid loss of relative precision in this subtraction,
    the smaller of I_x(a,b) and I_y(b,a) is computed first.

    ``beta_incomplete`` is derived from BRATIO in DiDonato and Morris
    (1992).

    References
    ----------
    DiDonato, A R and Morris, A H, 1992, `Algorithm 708: Significant
    digit computation of the incomplete beta function ratios`, ACM
    Trans. Math. Software (18), 360--373

    Wise, M E, 1950, `The incomplete beta function as a contour integral
    and a quickly converging series for its inverse`, Biometrika (37),
    208--218
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.beta_incomplete'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, w.
        utils._BYREF_FLOAT64, # Output, w1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _w = utils._EngineFloat64ScalarType.empty_ctypes()
    _w1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _x,
        _w,
        _w1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a and b were zero.',
                    'Constraint: a or b must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= x <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x and a were zero.',
                    'Constraint: x or a must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, 1.0-x and b were zero.',
                    'Constraint: 1.0-x or b must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _w_py = utils._EngineFloat64ScalarType.to_py(_w)
        _w1_py = utils._EngineFloat64ScalarType.to_py(_w1)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['w', 'w1'],
        )(
            _w_py,
            _w1_py,
        )
    _err_class.manage()
    return return_data

def beta_log_real_vector(n, a, b, f, ivalid):
    u"""
    Logarithm of the beta function, vectorized ln(B)(a,b).

    ``beta_log_real_vector`` returns an array of values of the logarithm
    of the beta function, ln(B)(a,b).

    For full information please refer to the NAG Library document for
    s14cp

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14cpf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    a : float, shape (`n`)
        The argument a_i of the function, for i = 1,2,...,`n`.

    b : float, shape (`n`)
        The argument b_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: ln(B)(a_i,b_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for a_i and
        b_i, for i = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            a_i or b_i <= 0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `a` or `b` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``beta_log_real_vector`` calculates values for ln(B)(a,b), for
    arrays of arguments a_i and b_i, for i = 1,2,...,n, where B is the
    beta function given by

        B(a,b) = int_0^1t^(a-1)(1-t)^(b-1)dt

    or equivalently

        B(a,b) = Gamma(a)Gamma(b)/Gamma(a+b)

    and Gamma(x) is the gamma function.
    Note that the beta function is symmetric, so that B(a,b) = B(b,a).

    In order to efficiently obtain accurate results several methods are
    used depending on the parameters a and b.

    Let a_0 = min(a,b) and b_0 = max(a,b).
    Then:

    for a_0 >= 8,

        ln(B) = 0.5 ln((2 pi)-0.5) ln(b_0) + Delta(a_0) + Delta (b_0) -
        Delta (a_0+b_0) - u - v ;

    where

        Delta(a_0) = ln(Gamma(a_0))-(a_0-0.5)ln(a_0)+a_0-0.5ln(2 pi),

        u = -(a_0-0.5)ln[(a_0)/(a_0+b_0)]  and

        v = b_0 ln(1+(a_0)/(b_0));

    for a_0 < 1,

        for b_0 >= 8,

                ln(B) = ln(Gamma)(a_0)+ln(Gamma(b_0)/Gamma(a_0+b_0));

        for b_0 < 8,

                ln(B) =
                ln(Gamma)(a_0)+ln(Gamma)(b_0)-ln(Gamma)(a_0+b_0);

    for 2 < a_0 < 8, a_0 is reduced to the interval [1,2] by B(a,b) =
    (a_0-1)/(a_0+b_0-1)B(a_0-1,b_0);

    for 1 <= a_0 <= 2,

        for b_0 >= 8,

                ln(B) = ln(Gamma)(a_0)+ln(Gamma(b_0)/Gamma(a_0+b_0));

        for 2 < b_0 < 8, b_0 is reduced to the interval [1,2];

        for b_0 <= 2,

                ln(B) =
                ln(Gamma)(a_0)+ln(Gamma)(b_0)-ln(Gamma)(a_0+b_0).

    ``beta_log_real_vector`` is derived from BETALN in DiDonato and
    Morris (1992).

    References
    ----------
    DiDonato, A R and Morris, A H, 1992, `Algorithm 708: Significant
    digit computation of the incomplete beta function ratios`, ACM
    Trans. Math. Software (18), 360--373
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.beta_log_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14cpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of a or b was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def beta_incomplete_vector(n, a, b, x, w, w1, ivalid):
    u"""
    Regularized incomplete beta function, vectorized I_x(a,b) and its
    complement 1-I_x.

    ``beta_incomplete_vector`` computes an array of values for the
    regularized incomplete beta function I_x(a,b) and its complement
    1-I_x(a,b).

    For full information please refer to the NAG Library document for
    s14cq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s14cqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    a : float, shape (`n`)
        The argument a_i of the function, for i = 1,2,...,`n`.

    b : float, shape (`n`)
        The argument b_i of the function, for i = 1,2,...,`n`.

    x : float, shape (`n`)
        x_i, the upper limit of integration, for i = 1,2,...,`n`.

    w : float, shape (`n`), modified in place
        `On exit`: the values of the incomplete beta function
        I_x_i(a_i,b_i) evaluated from zero to x_i.

    w1 : float, shape (`n`), modified in place
        `On exit`: the values of the complement of the incomplete beta
        function 1-I_x_i(a_i,b_i), i.e., the incomplete beta function
        evaluated from x_i to one.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for the ith
        evaluation, for i = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            a_i or b_i < 0.

        `ivalid`[i-1] = 2

            Both a_i and b_i = 0.

        `ivalid`[i-1] = 3

            x_i not in [0,1].

        `ivalid`[i-1] = 4

            Both x_i and a_i = 0.

        `ivalid`[i-1] = 5

            Both 1-x_i and b_i = 0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one argument had an invalid value.

            Check `ivalid` for more information.

    Notes
    -----
    ``beta_incomplete_vector`` evaluates the regularized incomplete beta
    function I_x(a,b) and its complement 1-I_x(a,b) in the normalized
    form, for arrays of arguments x_i, a_i and b_i, for i = 1,2,...,n.
    The incomplete beta function and its complement are given by

        [table omitted]

    with

        0 <= x <= 1,

        a >= 0 and b >= 0,

        and the beta function B(a,b) is defined as B(a,b) =
        int_0^1t^(a-1)(1-t)^(b-1)dt = Gamma(a)Gamma(b)/Gamma(a+b) where
        Gamma(y) is the gamma function.

    Several methods are used to evaluate the functions depending on the
    arguments a, b and x.
    The methods include Wise's asymptotic expansion (see Wise (1950))
    when a > b, continued fraction derived by DiDonato and Morris (1992)
    when a, b > 1, and power series when b <= 1 or b*x <= 0.7.
    When both a and b are large, specifically a, b >= 15, the DiDonato
    and Morris (1992) asymptotic expansion is employed for greater
    efficiency.

    Once either I_x(a,b) or I_y(b,a) is computed, the other is obtained
    by subtraction from 1.
    In order to avoid loss of relative precision in this subtraction,
    the smaller of I_x(a,b) and I_y(b,a) is computed first.

    ``beta_incomplete_vector`` is derived from BRATIO in DiDonato and
    Morris (1992).

    References
    ----------
    DiDonato, A R and Morris, A H, 1992, `Algorithm 708: Significant
    digit computation of the incomplete beta function ratios`, ACM
    Trans. Math. Software (18), 360--373

    Wise, M E, 1950, `The incomplete beta function as a contour integral
    and a quickly converging series for its inverse`, Biometrika (37),
    208--218
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.beta_incomplete_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s14cqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, w[n].
        utils._EngineFloat64ArrayType.c_type, # Output, w1[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _w = _w_cls.to_ctypes(can_reference_contents=False)
    _w1_cls = utils._EngineFloat64ArrayType(
        w1, {'fun_name': fun_name, 'entity': 'w1'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _w1 = _w1_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _x,
        _w,
        _w1,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one argument had an invalid value.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _w_cls.refresh_py(_w)
        _w1_cls.refresh_py(_w1)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def cdf_normal(x):
    u"""
    Cumulative Normal distribution function P(x).

    ``cdf_normal`` returns the value of the cumulative Normal
    distribution function, P(x).

    For full information please refer to the NAG Library document for
    s15ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    p : float
        The value of the cumulative Normal distribution function, P(x).

    Notes
    -----
    ``cdf_normal`` evaluates an approximate value for the cumulative
    Normal distribution function

        P(x) = 1/(sqrt(2 pi))int_- infinity^xe^(-u^2/2)du.

    The function is based on the fact that

        P(x) = 1/2erfc((-x)/(sqrt(2)))

    and it calls ``erfc_real`` to obtain a value of erfc for the
    appropriate argument.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.cdf_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, p.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _p,
    )
    utils._handle_cb_exception(excs)
    _p_py = utils._EngineFloat64ScalarType.to_py(_p)
    return_data = _p_py
    return return_data

def compcdf_normal(x):
    u"""
    Complement of cumulative Normal distribution function Q(x).

    ``compcdf_normal`` returns the value of the complement of the
    cumulative Normal distribution function, Q(x).

    For full information please refer to the NAG Library document for
    s15ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    q : float
        The value of the complement of the cumulative Normal
        distribution function, Q(x).

    Notes
    -----
    ``compcdf_normal`` evaluates an approximate value for the complement
    of the cumulative Normal distribution function

        Q(x) = 1/(sqrt(2 pi))int_x^infinity e^(-u^2/2)du.

    The function is based on the fact that

        Q(x) = 1/2erfc(x/(sqrt(2)))

    and it calls ``erfc_real`` to obtain the necessary value of erfc,
    the complementary error function.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.compcdf_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, q.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _q,
    )
    utils._handle_cb_exception(excs)
    _q_py = utils._EngineFloat64ScalarType.to_py(_q)
    return_data = _q_py
    return return_data

def erfc_real(x):
    u"""
    Complement of error function erfc(x).

    ``erfc_real`` returns the value of the complementary error function,
    erfc(x).

    For full information please refer to the NAG Library document for
    s15ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    res : float
        The value of the complementary error function, erfc(x).

    Notes
    -----
    ``erfc_real`` calculates an approximate value for the complement of
    the error function

        erfc(x) = 2/(sqrt(pi))int_x^infinity e^(-t^2)dt = 1-erf(x).

    Let x^ be the root of the equation erfc(x)-erf(x) = 0 (then
    x^≈0.46875).
    For |x| <= x^ the value of erfc(x) is based on the following
    rational Chebyshev expansion for erf(x):

        erf(x)≈xR_{ l,m}(x^2),

    where R_{ l,m} denotes a rational function of degree l in the
    numerator and m in the denominator.

    For |x| > x^ the value of erfc(x) is based on a rational Chebyshev
    expansion for erfc(x): for x^< |x| <= 4 the value is based on the
    expansion

        erfc(x)≈e^(x^2)R_{ l,m}(x);

    and for |x| > 4 it is based on the expansion

        erfc(x)≈e^(x^2)/x(1/(sqrt(pi))+1/x^2R_{ l,m}(1/x^2)).

    For each expansion, the specific values of l and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    For |x| >= x_hi there is a danger of setting underflow in erfc(x).
    For x >= x_hi, ``erfc_real`` returns erfc(x) = 0; for x <= -x_hi it
    returns erfc(x) = 2.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J, 1969, `Rational Chebyshev approximations for the error
    function`, Math.Comp. (23), 631--637
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfc_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, res.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _res = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _res,
    )
    utils._handle_cb_exception(excs)
    _res_py = utils._EngineFloat64ScalarType.to_py(_res)
    return_data = _res_py
    return return_data

def erf_real(x):
    u"""
    Error function erf(x).

    ``erf_real`` returns the value of the error function erf(x).

    For full information please refer to the NAG Library document for
    s15ae

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15aef.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    res : float
        The value of the error function erf(x).

    Notes
    -----
    ``erf_real`` calculates an approximate value for the error function

        erf(x) = 2/(sqrt(pi))int_0^xe^(-t^2)dt = 1-erfc(x).

    Let x^ be the root of the equation erfc(x)-erf(x) = 0 (then
    x^≈0.46875).
    For |x| <= x^ the value of erf(x) is based on the following rational
    Chebyshev expansion for erf(x):

        erf(x)≈xR_{ l,m}(x^2),

    where R_{ l,m} denotes a rational function of degree l in the
    numerator and m in the denominator.

    For |x| > x^ the value of erf(x) is based on a rational Chebyshev
    expansion for erfc(x): for x^< |x| <= 4 the value is based on the
    expansion

        erfc(x)≈e^(x^2)R_{ l,m}(x);

    and for |x| > 4 it is based on the expansion

        erfc(x)≈e^(x^2)/x(1/(sqrt(pi))+1/x^2R_{ l,m}(1/x^2)).

    For each expansion, the specific values of l and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    For |x| >= x_hi there is a danger of setting underflow in erfc(x).
    For x >= x_hi, ``erf_real`` returns erf(x) = 1; for x <= -x_hi it
    returns erf(x) = -1.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J, 1969, `Rational Chebyshev approximations for the error
    function`, Math.Comp. (23), 631--637
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erf_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, res.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _res = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _res,
    )
    utils._handle_cb_exception(excs)
    _res_py = utils._EngineFloat64ScalarType.to_py(_res)
    return_data = _res_py
    return return_data

def dawson(x):
    u"""
    Dawson's integral.

    ``dawson`` returns a value for Dawson's Integral, F(x).

    For full information please refer to the NAG Library document for
    s15af

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15aff.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    f : float
        The value of Dawson's Integral, F(x).

    Notes
    -----
    ``dawson`` evaluates an approximation for Dawson's Integral

        F(x) = e^(-x^2)int_0^xe^(t^2)dt.

    The function is based on two Chebyshev expansions:

    For 0 < |x| <= 4,

        F(x) = xsum'_r = 0a_r T_r(t), where t = 2(x/4)^2-1.

    For |x| > 4,

        F(x) = 1/xsum'_r = 0b_r T_r(t), where t = 2(4/x)^2-1.

    For |x| near zero, F(x) ~= x, and for |x| large, F(x) ~= 1/2x.
    These approximations are used for those values of x for which the
    result is correct to machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.dawson'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, f.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _f = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _f,
    )
    utils._handle_cb_exception(excs)
    _f_py = utils._EngineFloat64ScalarType.to_py(_f)
    return_data = _f_py
    return return_data

def erfcx_real(x):
    u"""
    Scaled complement of error function, erfcx(x).

    ``erfcx_real`` returns the value of the scaled complementary error
    function erfcx(x).

    For full information please refer to the NAG Library document for
    s15ag

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15agf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    ecx : float
        The value of the scaled complementary error function erfcx(x).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `x` = *<value>* and the constant x_hi = *<value>*.

            Constraint: `x` < x_hi.

        (`errno` 2)
            On entry, abs(`x`) was in the interval [*<value>*,*<value>*)
            where erfcx(`x`) is approximately 1/(sqrt(pi)*abs(`x`)): `x`
            = *<value>*.

        (`errno` 3)
            On entry, `x` = *<value>* and the constant x_neg =
            *<value>*.

            Constraint: `x` >= x_neg.

    Notes
    -----
    ``erfcx_real`` calculates an approximate value for the scaled
    complementary error function

        erfcx(x) = e^(x^2)erfc(x) = 2/(sqrt(pi))e^(x^2)int_x^infinity
        e^(-t^2)dt = e^(x^2)(1-erf(x)).

    Let x^ be the root of the equation erfc(x)-erf(x) = 0 (then
    x^≈0.46875).
    For |x| <= x^ the value of erfcx(x) is based on the following
    rational Chebyshev expansion for erf(x):

        erf(x)≈xR_{ l,m}(x^2),

    where R_{ l,m} denotes a rational function of degree l in the
    numerator and m in the denominator.

    For |x| > x^ the value of erfcx(x) is based on a rational Chebyshev
    expansion for erfc(x): for x^< |x| <= 4 the value is based on the
    expansion

        erfc(x)≈e^(x^2)R_{ l,m}(x);

    and for |x| > 4 it is based on the expansion

        erfc(x)≈e^(x^2)/x(1/(sqrt(pi))+1/x^2R_{ l,m}(1/x^2)).

    For each expansion, the specific values of l and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    Asymptotically, erfcx(x)~1/(sqrt(pi)|x|).
    There is a danger of setting underflow in erfcx(x) whenever x >=
    x_hi = min(x_huge, 1/(sqrt(pi)x_tiny)), where x_huge is the largest
    positive model number (see ``machine.real_largest``) and x_tiny is
    the smallest positive model number (see ``machine.real_smallest``).
    In this case ``erfcx_real`` exits with `errno` = 1 and returns
    erfcx(x) = 0.
    For x in the range 1/(2sqrt(epsilon)) <= x < x_hi, where epsilon is
    the machine precision, the asymptotic value 1/(sqrt(pi)|x|) is
    returned for erfcx(x) and ``erfcx_real`` exits with `errno` = 2.

    There is a danger of setting overflow in e^(x^2) whenever x < x_neg
    = -sqrt(log(x_huge/2)).
    In this case ``erfcx_real`` exits with `errno` = 3 and returns
    erfcx(x) = x_huge.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J, 1969, `Rational Chebyshev approximations for the error
    function`, Math.Comp. (23), 631--637
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfcx_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ecx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ecx = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ecx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and the constant'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'x_hi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x < x_hi.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1021:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(x) was in the interval',
                    '[{:s},{:s}) where erfcx(x) is'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'approximately 1/(sqrt(pi)*abs(x)): x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1031:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and the constant'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'x_neg = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x >= x_neg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ecx_py = utils._EngineFloat64ScalarType.to_py(_ecx)
        return_data = _ecx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ecx'],
            )(
                _ecx_py,
            )
    _err_class.manage()
    return return_data

def cdf_normal_vector(n, x, f):
    u"""
    Cumulative Normal distribution function, vectorized P(x).

    ``cdf_normal_vector`` returns an array of values of the cumulative
    Normal distribution function, P(x).

    For full information please refer to the NAG Library document for
    s15ap

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15apf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: P(x_i), the function values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``cdf_normal_vector`` evaluates approximate values of the cumulative
    Normal distribution function

        P(x) = 1/(sqrt(2 pi))int_- infinity^xe^(-u^2/2)du,

    for an array of arguments x_i, for i = 1,2,...,n.

    The function is based on the fact that

        P(x) = 1/2erfc((-x)/(sqrt(2)))

    and it calls ``erfc_real`` to obtain a value of erfc for the
    appropriate argument.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.cdf_normal_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15apft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        return_data = None
    _err_class.manage()
    return return_data

def compcdf_normal_vector(n, x, f):
    u"""
    Complement of cumulative Normal distribution function, vectorized
    Q(x).

    ``compcdf_normal_vector`` returns an array of values of the
    complement of the cumulative Normal distribution function, Q(x).

    For full information please refer to the NAG Library document for
    s15aq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15aqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Q(x_i), the function values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``compcdf_normal_vector`` evaluates approximate values for the
    complement of the cumulative Normal distribution function

        Q(x) = 1/(sqrt(2 pi))int_x^infinity e^(-u^2/2)du,

    for an array of arguments x_i, for i = 1,2,...,n.

    The function is based on the fact that

        Q(x) = 1/2erfc(x/(sqrt(2)))

    and it calls ``erfc_real`` to obtain the necessary value of erfc,
    the complementary error function.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.compcdf_normal_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15aqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        return_data = None
    _err_class.manage()
    return return_data

def erfc_real_vector(n, x, f):
    u"""
    Complement of error function, vectorized erfc(x).

    ``erfc_real_vector`` returns an array of values of the complementary
    error function, erfc(x).

    For full information please refer to the NAG Library document for
    s15ar

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15arf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: erfc(x_i), the function values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``erfc_real_vector`` calculates approximate values for the
    complement of the error function

        erfc(x) = 2/(sqrt(pi))int_x^infinity e^(-t^2)dt = 1-erf(x),

    for an array of arguments x_i, for i = 1,2,...,n.

    See ``erfc_real`` for details on the algorithm used.
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfc_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15arft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        return_data = None
    _err_class.manage()
    return return_data

def erf_real_vector(n, x, f):
    u"""
    Error function, vectorized erf(x).

    ``erf_real_vector`` returns an array of values of the error function
    erf(x)

    For full information please refer to the NAG Library document for
    s15as

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15asf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: erf(x_i), the function values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``erf_real_vector`` calculates approximate values for the error
    function

        erf(x) = 2/(sqrt(pi))int_0^xe^(-t^2)dt = 1-erfc(x),

    for an array of arguments x_i, for i = 1,2,...,n.

    See ``erf_real`` for details on the algorithm used.
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erf_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15asft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        return_data = None
    _err_class.manage()
    return return_data

def dawson_vector(n, x, f):
    u"""
    Dawson's integral, vectorized.

    ``dawson_vector`` returns an array of values for Dawson's Integral,
    F(x).

    For full information please refer to the NAG Library document for
    s15at

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15atf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: F(x_i), the function values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``dawson_vector`` evaluates approximations for Dawson's Integral

        F(x) = e^(-x^2)int_0^xe^(t^2)dt,

    for an array of arguments x_i, for i = 1,2,...,n.

    The function is based on two Chebyshev expansions:

    For 0 < |x| <= 4,

        F(x) = xsum'_r = 0a_r T_r(t), where t = 2(x/4)^2-1.

    For |x| > 4,

        F(x) = 1/xsum'_r = 0b_r T_r(t), where t = 2(4/x)^2-1.

    For |x| near zero, F(x) ~= x, and for |x| large, F(x) ~= 1/2x.
    These approximations are used for those values of x for which the
    result is correct to machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.dawson_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15atft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        return_data = None
    _err_class.manage()
    return return_data

def erfcx_real_vector(n, x, f, ivalid):
    u"""
    Scaled complement of error function, vectorized erfcx(x).

    ``erfcx_real_vector`` returns an array of values of the scaled
    complementary error function erfcx(x).

    For full information please refer to the NAG Library document for
    s15au

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15auf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: erfcx(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. The threshold value is the
            same as for `errno` = 1 in ``erfcx_real``.

        `ivalid`[i-1] = 2

            x_i was in the interval [1/(2sqrt(epsilon)),x_hi). The
            threshold values are the same as for `errno` = 2 in
            ``erfcx_real``.

        `ivalid`[i-1] = 3

            x_i is too small and positive. The threshold value is the
            same as for `errno` = 3 in ``erfcx_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` produced a result with
            reduced accuracy.

            Check `ivalid` for more information.

    Notes
    -----
    ``erfcx_real_vector`` calculates approximate values for the scaled
    complementary error function

        erfcx(x) = e^(x^2)erfc(x) = 2/(sqrt(pi))e^(x^2)int_x^infinity
        e^(-t^2)dt = e^(x^2)(1-erf(x)),

    for an array of arguments x_i, for i = 1,2,...,n.

    Let x^ be the root of the equation erfc(x)-erf(x) = 0 (then
    x^≈0.46875).
    For |x| <= x^ the value of erfcx(x) is based on the following
    rational Chebyshev expansion for erf(x):

        erf(x)≈xR_{ l,m}(x^2),

    where R_{ l,m} denotes a rational function of degree l in the
    numerator and m in the denominator.

    For |x| > x^ the value of erfcx(x) is based on a rational Chebyshev
    expansion for erfc(x): for x^< |x| <= 4 the value is based on the
    expansion

        erfc(x)≈e^(x^2)R_{ l,m}(x);

    and for |x| > 4 it is based on the expansion

        erfc(x)≈e^(x^2)/x(1/(sqrt(pi))+1/x^2R_{ l,m}(1/x^2)).

    For each expansion, the specific values of l and m are selected to
    be minimal such that the maximum relative error in the expansion is
    of the order 10^(-d), where d is the maximum number of decimal
    digits that can be accurately represented for the particular
    implementation (see ``machine.decimal_digits``).

    Asymptotically, erfcx(x)~1/(sqrt(pi)|x|).
    There is a danger of setting underflow in erfcx(x) whenever x >=
    x_hi = min(x_huge, 1/(sqrt(pi)x_tiny)), where x_huge is the largest
    positive model number (see ``machine.real_largest``) and x_tiny is
    the smallest positive model number (see ``machine.real_smallest``).
    In this case ``erfcx_real_vector`` exits with `errno` = 1 and
    returns `ivalid`[i-1] = 1 with erfcx(x_i) = 0.
    For x in the range 1/(2sqrt(epsilon)) <= x < x_hi, where epsilon is
    the machine precision, the asymptotic value 1/(sqrt(pi)|x|) is
    returned for erfcx(x_i), `ivalid`[i-1] = 2, and
    ``erfcx_real_vector`` exits with `errno` = 1.

    There is a danger of setting overflow in e^(x^2) whenever x < x_neg
    = -sqrt(log(x_huge/2)).
    In this case ``erfcx_real_vector`` exits with `errno` = 1 and
    returns `ivalid`[i-1] = 3 with erfcx(x_i) = x_huge.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Cody, W J, 1969, `Rational Chebyshev approximations for the error
    function`, Math.Comp. (23), 631--637
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfcx_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15auft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x produced a result with reduced accuracy.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def erfc_complex(z):
    u"""
    Scaled complex complement of error function, exp(-z^2)erfc(-iz).

    ``erfc_complex`` computes values of the function w(z) =
    e^(-z^2)erfc(-iz), for `complex` z.

    For full information please refer to the NAG Library document for
    s15dd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15ddf.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    Returns
    -------
    erfcz : complex
        The value of the function evaluated at z, w(z) =
        e^(-z^2)erfc(-iz).

    Raises
    ------
    NagValueError
        (`errno` 5)
            Result has no precision when entered with argument `z` =
            (*<value>*,*<value>*).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Real part of result overflows when entered with argument `z`
            = (*<value>*,*<value>*).

        (`errno` 2)
            Imaginary part of result overflows when entered with
            argument `z` = (*<value>*,*<value>*).

        (`errno` 3)
            Both real and imaginary parts of result overflow when
            entered with argument `z` = (*<value>*,*<value>*).

        (`errno` 4)
            Result has less than half precision when entered with
            argument `z` = (*<value>*,*<value>*).

    Notes
    -----
    ``erfc_complex`` computes values of the function w(z) =
    e^(-z^2)erfc(-iz), where erfc(z) is the complementary error function

        erfc(z) = 2/(sqrt(pi))int_z^infinity e^(-t^2)dt,

    for `complex` z.
    The method used is that in Gautschi (1970) for z in the first
    quadrant of the complex plane, and is extended for z in other
    quadrants via the relations w(-z) = 2e^(-z^2)-w(z) and w(-z-) =
    -w(-z)-.
    Following advice in Gautschi (1970) and van der Laan and Temme
    (1984), the code in Gautschi (1969) has been adapted to work in
    various precisions up to 18 decimal places.
    The real part of w(z) is sometimes known as the Voigt function.

    References
    ----------
    Gautschi, W, 1969, `Algorithm 363: Complex error function`, Comm.
    ACM (12), 635

    Gautschi, W, 1970, `Efficient computation of the complex error
    function`, SIAM J. Numer. Anal. (7), 187--198

    van der Laan, C G and Temme, N M, 1984, `Calculation of special
    functions: the gamma function, the exponential integrals and
    error-like functions`, CWI Tract 10, Centre for Mathematics and
    Computer Science, Amsterdam
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfc_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_COMPLEX128, # Output, erfcz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _erfcz = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _erfcz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Real part of result overflows when entered with',
                    'argument z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Imaginary part of result overflows when entered with',
                    'argument z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Both real and imaginary parts of result overflow when entered with',
                    'argument z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Result has less than half precision when entered with',
                    'argument z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Result has no precision when entered with',
                    'argument z = ({:s},{:s}).'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _erfcz_py = utils._EngineComplex128ScalarType.to_py(_erfcz)
        return_data = _erfcz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['erfcz'],
            )(
                _erfcz_py,
            )
    _err_class.manage()
    return return_data

def erfc_complex_vector(n, z, f, ivalid):
    u"""
    Scaled complex complement of error function, vectorized
    exp(-z^2)erfc(-iz).

    ``erfc_complex_vector`` computes values of the function w(z) =
    e^(-z^2)erfc(-iz), for an array of `complex` values z.

    For full information please refer to the NAG Library document for
    s15dr

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s15drf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    z : complex, shape (`n`)
        The argument z_i of the function, for i = 1,2,...,`n`.

    f : complex, shape (`n`), modified in place
        `On exit`: w(z_i) = e^(-z_i^2), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for z_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            Real part of result overflows.

        `ivalid`[i-1] = 2

            Imaginary part of result overflows.

        `ivalid`[i-1] = 3

            Both real and imaginary part of result overflows.

        `ivalid`[i-1] = 4

            Result has less than half precision.

        `ivalid`[i-1] = 5

            Result has no precision.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `z` produced a result with
            reduced accuracy.

            Check `ivalid` for more information.

    Notes
    -----
    ``erfc_complex_vector`` computes values of the function w(z_i) =
    e^(-z_i^2)erfc(-iz_i), for i = 1,2,...,n, where erfc(z_i) is the
    complementary error function

        erfc(z) = 2/(sqrt(pi))int_z^infinity e^(-t^2)dt,

    for `complex` z.
    The method used is that in Gautschi (1970) for z in the first
    quadrant of the complex plane, and is extended for z in other
    quadrants via the relations w(-z) = 2e^(-z^2)-w(z) and w(-z-) =
    -w(-z)-.
    Following advice in Gautschi (1970) and van der Laan and Temme
    (1984), the code in Gautschi (1969) has been adapted to work in
    various precisions up to 18 decimal places.
    The real part of w(z) is sometimes known as the Voigt function.

    References
    ----------
    Gautschi, W, 1969, `Algorithm 363: Complex error function`, Comm.
    ACM (12), 635

    Gautschi, W, 1970, `Efficient computation of the complex error
    function`, SIAM J. Numer. Anal. (7), 187--198

    van der Laan, C G and Temme, N M, 1984, `Calculation of special
    functions: the gamma function, the exponential integrals and
    error-like functions`, CWI Tract 10, Centre for Mathematics and
    Computer Science, Amsterdam
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.erfc_complex_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s15drft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, z[n].
        utils._EngineComplex128ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineComplex128ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineComplex128ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _z,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of z produced a result with reduced accuracy.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_y0_real(x):
    u"""
    Bessel function Y_0(x).

    ``bessel_y0_real`` returns the value of the Bessel function Y_0(x).

    For full information please refer to the NAG Library document for
    s17ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    y0 : float
        The value of the Bessel function Y_0(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

    Notes
    -----
    ``bessel_y0_real`` evaluates an approximation to the Bessel function
    of the second kind Y_0(x).

    **Note:**  Y_0(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on four Chebyshev expansions:

    For 0 < x <= 8,

        Y_0(x) = 2/piln(x)sum'_r = 0a_r T_r(t)+sum'_r = 0b_r T_r(t),
        with t = 2(x/8)^2-1.

    For x > 8,

        Y_0(x) = sqrt(2/pi x){P_0(x)sin(x-pi/4)+Q_0(x)cos(x-pi/4)}

    where P_0(x) = sum'_{r = 0}c_r T_r(t),

    and Q_0(x) = 8/xsum'_{r = 0}d_r T_r(t),with t = 2(8/x)^2-1.

    For x near zero, Y_0(x) ~= 2/pi(ln(x/2)+gamma), where gamma denotes
    Euler's constant.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of Y_0(x); only the amplitude, sqrt(2/pi n),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if x≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_y0_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, y0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y0_py = utils._EngineFloat64ScalarType.to_py(_y0)
        return_data = _y0_py
    _err_class.manage()
    return return_data

def bessel_y1_real(x):
    u"""
    Bessel function Y_1(x).

    ``bessel_y1_real`` returns the value of the Bessel function Y_1(x).

    For full information please refer to the NAG Library document for
    s17ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    y1 : float
        The value of the Bessel function Y_1(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            x is too large, the function returns the amplitude of the
            Y_1 oscillation, sqrt(2/(pi*x)).

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            Y_1 is undefined, the function returns zero.

        (`errno` 3)
            `x` is too close to zero and there is danger of overflow,
            `x` = *<value>*.

            Constraint: `x` > *<value>*.

            The function returns the value of Y_1(x) at the smallest
            valid argument.

    Notes
    -----
    ``bessel_y1_real`` evaluates an approximation to the Bessel function
    of the second kind Y_1(x).

    **Note:**  Y_1(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on four Chebyshev expansions:

    For 0 < x <= 8,

        Y_1(x) = 2/piln(x)x/8sum'_{r = 0}a_r T_r(t)-2/pi x+x/8sum'_{r =
        0}b_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        Y_1(x) = sqrt(2/pi x){P_1(x)sin(x-3pi/4)+Q_1(x)cos(x-3pi/4)}

    where P_1(x) = sum'_{r = 0}c_r T_r(t),

    and Q_1(x) = 8/xsum'_{r = 0}d_r T_r(t), with t = 2(8/x)^2-1.

    For x near zero, Y_1(x) ~= -2/pi x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For extremely small x, there is a danger of overflow in calculating
    -2/pi x and for such arguments the function will fail.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of Y_1(x); only the amplitude, sqrt(2/pi x),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if x≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_y1_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, y1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large, the function returns the amplitude of the Y_1 oscillation,',
                    'sqrt(2/(pi*x)).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'Y_1 is undefined, the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'x is too close to zero and there is danger of overflow,',
                    'x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The function returns the value of Y_1(x) at the smallest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y1_py = utils._EngineFloat64ScalarType.to_py(_y1)
        return_data = _y1_py
    _err_class.manage()
    return return_data

def bessel_j0_real(x):
    u"""
    Bessel function J_0(x).

    ``bessel_j0_real`` returns the value of the Bessel function J_0(x).

    For full information please refer to the NAG Library document for
    s17ae

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17aef.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    j0 : float
        The value of the Bessel function J_0(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large, the function returns the amplitude of
            the J_0 oscillation, sqrt(2/(pi*abs(x))).

    Notes
    -----
    ``bessel_j0_real`` evaluates an approximation to the Bessel function
    of the first kind J_0(x).

    **Note:**  J_0(-x) = J_0(x), so the approximation need only consider
    x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 8,

        J_0(x) = sum'_{r = 0}a_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        J_0(x) = sqrt(2/pi x){P_0(x)cos(x-pi/4)-Q_0(x)sin(x-pi/4)},

    where P_0(x) = sum'_{r = 0}b_r T_r(t),

    and Q_0(x) = 8/xsum'_{r = 0}c_r T_r(t),

    with t = 2(8/x)^2-1.

    For x near zero, J_0(x) ~= 1.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of J_0(x); only the amplitude, sqrt(2/pi|x|),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if |x|≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j0_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, j0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _j0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _j0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large, the function returns the amplitude of the',
                    'J_0 oscillation, sqrt(2/(pi*abs(x))).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _j0_py = utils._EngineFloat64ScalarType.to_py(_j0)
        return_data = _j0_py
    _err_class.manage()
    return return_data

def bessel_j1_real(x):
    u"""
    Bessel function J_1(x).

    ``bessel_j1_real`` returns the value of the Bessel function J_1(x).

    For full information please refer to the NAG Library document for
    s17af

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17aff.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    j1 : float
        The value of the Bessel function J_1(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large, the function returns the amplitude of
            the J_1 oscillation, sqrt(2/(pi*abs(x))).

    Notes
    -----
    ``bessel_j1_real`` evaluates an approximation to the Bessel function
    of the first kind J_1(x).

    **Note:**  J_1(-x) = -J_1(x), so the approximation need only
    consider x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 8,

        J_1(x) = x/8sum'_{r = 0}a_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        J_1(x) = sqrt(2/pi x){P_1(x)cos(x-3 pi/4)-Q_1(x)sin(x-3 pi/4)}

    where P_1(x) = sum'_{r = 0}b_r T_r(t),

    and Q_1(x) = 8/xsum'_{r = 0}c_r T_r(t),

    with t = 2(8/x)^2-1.

    For x near zero, J_1(x) ~= x/2.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of J_1(x); only the amplitude, sqrt(2/pi|x|),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if |x|≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j1_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, j1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _j1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _j1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large, the function returns the amplitude of the J_1',
                    'oscillation, sqrt(2/(pi*abs(x))).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _j1_py = utils._EngineFloat64ScalarType.to_py(_j1)
        return_data = _j1_py
    _err_class.manage()
    return return_data

def airy_ai_real(x):
    u"""
    Airy function Ai(x).

    ``airy_ai_real`` returns a value for the Airy function, Ai(x).

    For full information please refer to the NAG Library document for
    s17ag

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17agf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    ai : float
        The value of the Airy function, Ai(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            `x` is too large and positive. The function returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= *<value>*.

            `x` is too large and negative. The function returns zero.

    Notes
    -----
    ``airy_ai_real`` evaluates an approximation to the Airy function,
    Ai(x).
    It is based on a number of Chebyshev expansions:

    For x < -5,

        Ai(x) = (a(t)sin(z)-b(t)cos(z))/((-x)^(1/4))

    where z = pi/4+2/3sqrt(-x^3), and a(t) and b(t) are expansions in
    the variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Ai(x) = f(t)-xg(t),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Ai(x) = e^(-3x/2)y(t),

    where y is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Ai(x) = e^(-5x/2)u(t),

    where u is an expansion in t = 4x/9-3.

    For x >= 9,

        Ai(x) = e^(-z)v(t)/(x^(1/4)),

    where z = 2/3sqrt(x^3) and v is an expansion in t = 2*(18/z)-1.

    For |x| < machine precision, the result is set directly to Ai(0).
    This both saves time and guards against underflow in intermediate
    calculations.

    For large negative arguments, it becomes impossible to calculate the
    phase of the oscillatory function with any precision and so the
    function must fail.
    This occurs if x < -(3/2 epsilon)^(2/3), where epsilon is the
    machine precision.

    For large positive arguments, where Ai decays in an essentially
    exponential manner, there is a danger of underflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_ai_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ai.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ai = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ai,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and positive. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and negative. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ai_py = utils._EngineFloat64ScalarType.to_py(_ai)
        return_data = _ai_py
    _err_class.manage()
    return return_data

def airy_bi_real(x):
    u"""
    Airy function Bi(x).

    ``airy_bi_real`` returns a value of the Airy function, Bi(x).

    For full information please refer to the NAG Library document for
    s17ah

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17ahf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    bi : float
        The value of the Airy function, Bi(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            `x` is too large and positive. The function returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= *<value>*.

            `x` is too large and negative. The function returns zero.

    Notes
    -----
    ``airy_bi_real`` evaluates an approximation to the Airy function
    Bi(x).
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Bi(x) = (a(t)cos(z)+b(t)sin(z))/((-x)^(1/4)),

    where z = pi/4+2/3sqrt(-x^3) and a(t) and b(t) are expansions in the
    variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Bi(x) = sqrt(3)(f(t)+xg(t)),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Bi(x) = e^(11x/8)y(t),

    where y is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Bi(x) = e^(5x/2)v(t),

    where v is an expansion in t = 4x/9-3.

    For x >= 9,

        Bi(x) = e^zu(t)/(x^(1/4)),

    where z = 2/3sqrt(x^3) and u is an expansion in t = 2*(18/z)-1.

    For |x| < machine precision, the result is set directly to Bi(0).
    This both saves time and avoids possible intermediate underflows.

    For large negative arguments, it becomes impossible to calculate the
    phase of the oscillating function with any accuracy so the function
    must fail.
    This occurs if x < -(3/2 epsilon)^(2/3), where epsilon is the
    machine precision.

    For large positive arguments, there is a danger of causing overflow
    since Bi grows in an essentially exponential manner, so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_bi_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, bi.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _bi = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _bi,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and positive. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and negative. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _bi_py = utils._EngineFloat64ScalarType.to_py(_bi)
        return_data = _bi_py
    _err_class.manage()
    return return_data

def airy_ai_deriv(x):
    u"""
    Airy function Ai'(x).

    ``airy_ai_deriv`` returns a value of the derivative of the Airy
    function Ai(x).

    For full information please refer to the NAG Library document for
    s17aj

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17ajf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    aid : float
        The value of the derivative of the Airy function Ai(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            `x` is too large and positive. The function returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= *<value>*.

            `x` is too large and negative. The function returns zero.

    Notes
    -----
    ``airy_ai_deriv`` evaluates an approximation to the derivative of
    the Airy function Ai(x).
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Ai'(x) = fourthroot(-x)[a(t)cos(z)+b(t)/zetasin(z)],

    where z = pi/4+zeta, zeta = 2/3sqrt(-x^3) and a(t) and b(t) are
    expansions in variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Ai'(x) = x^2f(t)-g(t),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Ai'(x) = e^(-11x/8)y(t),

    where y(t) is an expansion in t = 4*(x/9)-1.

    For 4.5 <= x < 9,

        Ai'(x) = e^(-5x/2)v(t),

    where v(t) is an expansion in t = 4*(x/9)-3.

    For x >= 9,

        Ai'(x) = fourthroot(x)e^(-z)u(t),

    where z = 2/3sqrt(x^3) and u(t) is an expansion in t = 2*(18/z)-1.

    For |x| < the square of the machine precision, the result is set
    directly to Ai'(0).
    This both saves time and avoids possible intermediate underflows.

    For large negative arguments, it becomes impossible to calculate a
    result for the oscillating function with any accuracy and so the
    function must fail.
    This occurs for x < -((sqrt(pi))/epsilon)^(4/7), where epsilon is
    the machine precision.

    For large positive arguments, where Ai' decays in an essentially
    exponential manner, there is a danger of underflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_ai_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, aid.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _aid = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _aid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and positive. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and negative. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _aid_py = utils._EngineFloat64ScalarType.to_py(_aid)
        return_data = _aid_py
    _err_class.manage()
    return return_data

def airy_bi_deriv(x):
    u"""
    Airy function Bi'(x).

    ``airy_bi_deriv`` returns a value for the derivative of the Airy
    function Bi(x).

    For full information please refer to the NAG Library document for
    s17ak

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17akf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    bid : float
        The value of the derivative of the Airy function Bi(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

            `x` is too large and positive. The function returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= *<value>*.

            `x` is too large and negative. The function returns zero.

    Notes
    -----
    ``airy_bi_deriv`` calculates an approximate value for the derivative
    of the Airy function Bi(x).
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Bi'(x) = fourthroot(-x)[-a(t)sin(z)+b(t)/zetacos(z)],

    where z = pi/4+zeta, zeta = 2/3sqrt(-x^3) and a(t) and b(t) are
    expansions in the variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Bi'(x) = sqrt(3)(x^2f(t)+g(t)),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Bi'(x) = e^(3x/2)y(t),

    where y(t) is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Bi'(x) = e^(21x/8)u(t),

    where u(t) is an expansion in t = 4x/9-3.

    For x >= 9,

        Bi'(x) = fourthroot(x)e^zv(t),

    where z = 2/3sqrt(x^3) and v(t) is an expansion in t = 2*(18/z)-1.

    For |x| < the square of the machine precision, the result is set
    directly to Bi'(0).
    This saves time and avoids possible underflows in calculation.

    For large negative arguments, it becomes impossible to calculate a
    result for the oscillating function with any accuracy so the
    function must fail.
    This occurs for x < -((sqrt(pi))/epsilon)^(4/7), where epsilon is
    the machine precision.

    For large positive arguments, where Bi' grows in an essentially
    exponential manner, there is a danger of overflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_bi_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, bid.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _bid = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _bid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and positive. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large and negative. The function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _bid_py = utils._EngineFloat64ScalarType.to_py(_bid)
        return_data = _bid_py
    _err_class.manage()
    return return_data

def bessel_zeros(a, n, mode, rel, x):
    u"""
    Zeros of Bessel functions J_alpha(x), J'_alpha(x), Y_alpha(x) or
    Y'_alpha(x).

    ``bessel_zeros`` determines the leading `n` zeros of one of the
    Bessel functions J_alpha(x), Y_alpha(x), J'_alpha(x) or Y'_alpha(x)
    for real x and non-negative alpha.

    For full information please refer to the NAG Library document for
    s17al

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17alf.html

    Parameters
    ----------
    a : float
        The order alpha of the function.

    n : int
        The number N of zeros required.

    mode : int
        Specifies the form of the function whose zeros are required.

        `mode` = 1

            The zeros of J_alpha(x) are required.

        `mode` = 2

            The zeros of Y_alpha(x) are required;

        `mode` = 3

            The zeros of J'_alpha(x) are required;

        `mode` = 4

            The zeros of Y'_alpha(x) are required.

    rel : float
        The relative accuracy to which the zeros are required.

        Suggested value: the square root of the machine precision.

    x : float, shape (`n`), modified in place
        `On exit`: the N required zeros of the function specified by
        `mode`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rel` = *<value>*.

            Constraint: `rel` > 0.0.

        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` <= 4.

        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` <= 100000.0.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

    Notes
    -----
    ``bessel_zeros`` attempts to find the leading N zeros of one of the
    Bessel functions J_alpha(x), Y_alpha(x), J'_alpha(x) or Y'_alpha(x),
    where x is real.
    When alpha is real, these functions each have an infinite number of
    real zeros, all of which are simple with the possible exception of x
    = 0.
    If alpha >= 0, the nth positive zero is denoted by
    j_{alpha,n},j'_alpha,n,y_{alpha,n} and y'_alpha,n, respectively, for
    n = 1,2,...,N, except that x = 0 is counted as the first zero of
    J'_alpha(x) when alpha = 0.
    Since J'_0(x) = -J_1(x), it, therefore, follows that j'_0,1 = 0 and
    j'_0,n = -j_{1,n-1} for n = 2,3,...,N-1.
    Further details can be found in Section 9.5 of Abramowitz and Stegun
    (1972).

    ``bessel_zeros`` is based on Algol 60 procedures given by Temme
    (1979).
    Initial approximations to the zeros are computed from asymptotic
    expansions.
    These are then improved by higher-order Newton iteration making use
    of the differential equation for the Bessel functions.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Temme, N M, 1976, `On the numerical evaluation of the ordinary
    Bessel function of the second kind`, J. Comput. Phys. (21), 343--350

    Temme, N M, 1979, `An algorithm with Algol 60 program for the
    computation of the zeros of ordinary Bessel functions and those of
    their derivatives`, J. Comput. Phys. (32), 270--279
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_zeros'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17alft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_FLOAT64, # Input, rel.
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _rel = utils._EngineFloat64ScalarType(
        rel, {'fun_name': fun_name, 'entity': 'rel'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _n,
        _mode,
        _rel,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rel > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode <= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a <= 100000.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def bessel_y0_real_vector(n, x, f, ivalid):
    u"""
    Bessel function vectorized Y_0(x).

    ``bessel_y0_real_vector`` returns an array of values of the Bessel
    function Y_0(x).

    For full information please refer to the NAG Library document for
    s17aq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17aqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Y_0(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i is too large. `f`[i-1] contains the amplitude
            of the Y_0 oscillation, sqrt(2/pi x_i).

        `ivalid`[i-1] = 2

            On entry, x_i <= 0.0, Y_0 is undefined. `f`[i-1] contains
            0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_y0_real_vector`` evaluates an approximation to the Bessel
    function of the second kind Y_0(x_i) for an array of arguments x_i,
    for i = 1,2,...,n.

    **Note:**  Y_0(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on four Chebyshev expansions:

    For 0 < x <= 8,

        Y_0(x) = 2/piln(x)sum'_{r = 0}a_r T_r(t)+sum'_{r = 0}b_r T_r(t),
        with t = 2(x/8)^2-1.

    For x > 8,

        Y_0(x) = sqrt(2/pi x){P_0(x)sin(x-pi/4)+Q_0(x)cos(x-pi/4)}

    where P_0(x) = sum'_{r = 0}c_r T_r(t),

    and Q_0(x) = 8/xsum'_{r = 0}d_r T_r(t),with t = 2(8/x)^2-1.

    For x near zero, Y_0(x) ~= 2/pi(ln(x/2)+gamma), where gamma denotes
    Euler's constant.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of Y_0(x); only the amplitude, sqrt(2/pi n),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if x≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_y0_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17aqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_y1_real_vector(n, x, f, ivalid):
    u"""
    Bessel function vectorized Y_1(x).

    ``bessel_y1_real_vector`` returns an array of values of the Bessel
    function Y_1(x).

    For full information please refer to the NAG Library document for
    s17ar

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17arf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Y_1(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i is too large. `f`[i-1] contains the amplitude
            of the Y_1 oscillation, sqrt(2/pi x_i).

        `ivalid`[i-1] = 2

            On entry, x_i <= 0.0, Y_1 is undefined. `f`[i-1] contains
            0.0.

        `ivalid`[i-1] = 3

            x_i is too close to zero, there is a danger of overflow. On
            failure, `f`[i-1] contains the value of Y_1(x) at the
            smallest valid argument.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_y1_real_vector`` evaluates an approximation to the Bessel
    function of the second kind Y_1(x_i) for an array of arguments x_i,
    for i = 1,2,...,n.

    **Note:**  Y_1(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on four Chebyshev expansions:

    For 0 < x <= 8,

        Y_1(x) = 2/piln(x)x/8sum'_{r = 0}a_r T_r(t)-2/pi x+x/8sum'_{r =
        0}b_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        Y_1(x) = sqrt(2/pi x){P_1(x)sin(x-3pi/4)+Q_1(x)cos(x-3pi/4)}

    where P_1(x) = sum'_{r = 0}c_r T_r(t),

    and Q_1(x) = 8/xsum'_{r = 0}d_r T_r(t), with t = 2(8/x)^2-1.

    For x near zero, Y_1(x) ~= -2/pi x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For extremely small x, there is a danger of overflow in calculating
    -2/pi x and for such arguments the function will fail.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of Y_1(x); only the amplitude, sqrt(2/pi x),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if x≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_y1_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17arft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_j0_real_vector(n, x, f, ivalid):
    u"""
    Bessel function vectorized J_0(x).

    ``bessel_j0_real_vector`` returns an array of values of the Bessel
    function J_0(x).

    For full information please refer to the NAG Library document for
    s17as

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17asf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: J_0(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i is too large. `f`[i-1] contains the amplitude
            of the J_0 oscillation, sqrt(2/pi|x_i|).

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_j0_real_vector`` evaluates an approximation to the Bessel
    function of the first kind J_0(x_i) for an array of arguments x_i,
    for i = 1,2,...,n.

    **Note:**  J_0(-x) = J_0(x), so the approximation need only consider
    x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 8,

        J_0(x) = sum'_{r = 0}a_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        J_0(x) = sqrt(2/pi x){P_0(x)cos(x-pi/4)-Q_0(x)sin(x-pi/4)},

    where P_0(x) = sum'_{r = 0}b_r T_r(t),

    and Q_0(x) = 8/xsum'_{r = 0}c_r T_r(t),

    with t = 2(8/x)^2-1.

    For x near zero, J_0(x) ~= 1.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of J_0(x); only the amplitude, sqrt(2/pi|x|),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if |x|≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j0_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17asft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_j1_real_vector(n, x, f, ivalid):
    u"""
    Bessel function vectorized J_1(x).

    ``bessel_j1_real_vector`` returns an array of values of the Bessel
    function J_1(x).

    For full information please refer to the NAG Library document for
    s17at

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17atf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: J_1(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i is too large. `f`[i-1] contains the amplitude
            of the J_1 oscillation, sqrt(2/pi|x_i|).

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_j1_real_vector`` evaluates an approximation to the Bessel
    function of the first kind J_1(x_i) for an array of arguments x_i,
    for i = 1,2,...,n.

    **Note:**  J_1(-x) = -J_1(x), so the approximation need only
    consider x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 8,

        J_1(x) = x/8sum'_{r = 0}a_r T_r(t), with t = 2(x/8)^2-1.

    For x > 8,

        J_1(x) = sqrt(2/pi x){P_1(x)cos(x-3 pi/4)-Q_1(x)sin(x-3 pi/4)}

    where P_1(x) = sum'_{r = 0}b_r T_r(t),

    and Q_1(x) = 8/xsum'_{r = 0}c_r T_r(t),

    with t = 2(8/x)^2-1.

    For x near zero, J_1(x) ~= x/2.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For very large x, it becomes impossible to provide results with any
    reasonable accuracy (see [Accuracy]), hence the function fails.
    Such arguments contain insufficient information to determine the
    phase of oscillation of J_1(x); only the amplitude, sqrt(2/pi|x|),
    can be determined and this is returned on failure.
    The range for which this occurs is roughly related to machine
    precision; the function will fail if |x|≳1/machine precision.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j1_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17atft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def airy_ai_real_vector(n, x, f, ivalid):
    u"""
    Airy function vectorized Ai(x).

    ``airy_ai_real_vector`` returns an array of values for the Airy
    function, Ai(x).

    For full information please refer to the NAG Library document for
    s17au

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17auf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Ai(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 1 in
            ``airy_ai_real``.

        `ivalid`[i-1] = 2

            x_i is too large and negative. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 2 in
            ``airy_ai_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``airy_ai_real_vector`` evaluates an approximation to the Airy
    function, Ai(x_i) for an array of arguments x_i, for i = 1,2,...,n.
    It is based on a number of Chebyshev expansions:

    For x < -5,

        Ai(x) = (a(t)sin(z)-b(t)cos(z))/((-x)^(1/4))

    where z = pi/4+2/3sqrt(-x^3), and a(t) and b(t) are expansions in
    the variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Ai(x) = f(t)-xg(t),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Ai(x) = e^(-3x/2)y(t),

    where y is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Ai(x) = e^(-5x/2)u(t),

    where u is an expansion in t = 4x/9-3.

    For x >= 9,

        Ai(x) = e^(-z)v(t)/(x^(1/4)),

    where z = 2/3sqrt(x^3) and v is an expansion in t = 2*(18/z)-1.

    For |x| < machine precision, the result is set directly to Ai(0).
    This both saves time and guards against underflow in intermediate
    calculations.

    For large negative arguments, it becomes impossible to calculate the
    phase of the oscillatory function with any precision and so the
    function must fail.
    This occurs if x < -(3/2 epsilon)^(2/3), where epsilon is the
    machine precision.

    For large positive arguments, where Ai decays in an essentially
    exponential manner, there is a danger of underflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_ai_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17auft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def airy_bi_real_vector(n, x, f, ivalid):
    u"""
    Airy function vectorized Bi(x).

    ``airy_bi_real_vector`` returns an array of values of the Airy
    function, Bi(x).

    For full information please refer to the NAG Library document for
    s17av

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17avf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Bi(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 1 in
            ``airy_bi_real``.

        `ivalid`[i-1] = 2

            x_i is too large and negative. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 2 in
            ``airy_bi_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``airy_bi_real_vector`` evaluates an approximation to the Airy
    function Bi(x_i) for an array of arguments x_i, for i = 1,2,...,n.
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Bi(x) = (a(t)cos(z)+b(t)sin(z))/((-x)^(1/4)),

    where z = pi/4+2/3sqrt(-x^3) and a(t) and b(t) are expansions in the
    variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Bi(x) = sqrt(3)(f(t)+xg(t)),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Bi(x) = e^(11x/8)y(t),

    where y is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Bi(x) = e^(5x/2)v(t),

    where v is an expansion in t = 4x/9-3.

    For x >= 9,

        Bi(x) = e^zu(t)/(x^(1/4)),

    where z = 2/3sqrt(x^3) and u is an expansion in t = 2*(18/z)-1.

    For |x| < machine precision, the result is set directly to Bi(0).
    This both saves time and avoids possible intermediate underflows.

    For large negative arguments, it becomes impossible to calculate the
    phase of the oscillating function with any accuracy so the function
    must fail.
    This occurs if x < -(3/2 epsilon)^(2/3), where epsilon is the
    machine precision.

    For large positive arguments, there is a danger of causing overflow
    since Bi grows in an essentially exponential manner, so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_bi_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17avft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def airy_ai_deriv_vector(n, x, f, ivalid):
    u"""
    Derivatives of the Airy function, vectorized Ai'(x).

    ``airy_ai_deriv_vector`` returns an array of values of the
    derivative of the Airy function Ai(x).

    For full information please refer to the NAG Library document for
    s17aw

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17awf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Ai'(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 1 in
            ``airy_ai_deriv``.

        `ivalid`[i-1] = 2

            x_i is too large and negative. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 2 in
            ``airy_ai_deriv``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``airy_ai_deriv_vector`` evaluates an approximation to the
    derivative of the Airy function Ai(x_i) for an array of arguments
    x_i, for i = 1,2,...,n.
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Ai'(x) = fourthroot(-x)[a(t)cos(z)+b(t)/zetasin(z)],

    where z = pi/4+zeta, zeta = 2/3sqrt(-x^3) and a(t) and b(t) are
    expansions in variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Ai'(x) = x^2f(t)-g(t),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Ai'(x) = e^(-11x/8)y(t),

    where y(t) is an expansion in t = 4*(x/9)-1.

    For 4.5 <= x < 9,

        Ai'(x) = e^(-5x/2)v(t),

    where v(t) is an expansion in t = 4*(x/9)-3.

    For x >= 9,

        Ai'(x) = fourthroot(x)e^(-z)u(t),

    where z = 2/3sqrt(x^3) and u(t) is an expansion in t = 2*(18/z)-1.

    For |x| < the square of the machine precision, the result is set
    directly to Ai'(0).
    This both saves time and avoids possible intermediate underflows.

    For large negative arguments, it becomes impossible to calculate a
    result for the oscillating function with any accuracy and so the
    function must fail.
    This occurs for x < -((sqrt(pi))/epsilon)^(4/7), where epsilon is
    the machine precision.

    For large positive arguments, where Ai' decays in an essentially
    exponential manner, there is a danger of underflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_ai_deriv_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17awft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def airy_bi_deriv_vector(n, x, f, ivalid):
    u"""
    Derivatives of the Airy function, vectorized Bi'(x).

    ``airy_bi_deriv_vector`` returns an array of values for the
    derivative of the Airy function Bi(x).

    For full information please refer to the NAG Library document for
    s17ax

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17axf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: Bi'(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large and positive. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 1 in
            ``airy_bi_deriv``.

        `ivalid`[i-1] = 2

            x_i is too large and negative. `f`[i-1] contains zero. The
            threshold value is the same as for `errno` = 2 in
            ``airy_bi_deriv``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``airy_bi_deriv_vector`` calculates an approximate value for the
    derivative of the Airy function Bi(x_i) for an array of arguments
    x_i, for i = 1,2,...,n.
    It is based on a number of Chebyshev expansions.

    For x < -5,

        Bi'(x) = fourthroot(-x)[-a(t)sin(z)+b(t)/zetacos(z)],

    where z = pi/4+zeta, zeta = 2/3sqrt(-x^3) and a(t) and b(t) are
    expansions in the variable t = -2(5/x)^3-1.

    For -5 <= x <= 0,

        Bi'(x) = sqrt(3)(x^2f(t)+g(t)),

    where f and g are expansions in t = -2(x/5)^3-1.

    For 0 < x < 4.5,

        Bi'(x) = e^(3x/2)y(t),

    where y(t) is an expansion in t = 4x/9-1.

    For 4.5 <= x < 9,

        Bi'(x) = e^(21x/8)u(t),

    where u(t) is an expansion in t = 4x/9-3.

    For x >= 9,

        Bi'(x) = fourthroot(x)e^zv(t),

    where z = 2/3sqrt(x^3) and v(t) is an expansion in t = 2*(18/z)-1.

    For |x| < the square of the machine precision, the result is set
    directly to Bi'(0).
    This saves time and avoids possible underflows in calculation.

    For large negative arguments, it becomes impossible to calculate a
    result for the oscillating function with any accuracy so the
    function must fail.
    This occurs for x < -((sqrt(pi))/epsilon)^(4/7), where epsilon is
    the machine precision.

    For large positive arguments, where Bi' grows in an essentially
    exponential manner, there is a danger of overflow so the function
    must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_bi_deriv_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17axft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_y_complex(fnu, z, n, scal, cy):
    u"""
    Bessel functions Y_{nu+a}(z), real a >= 0, complex z, nu =
    0,1,2,....

    ``bessel_y_complex`` returns a sequence of values for the Bessel
    functions Y_{nu+n}(z) for complex z, non-negative nu and n =
    0,1,...,N-1, with an option for exponential scaling.

    For full information please refer to the NAG Library document for
    s17dc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17dcf.html

    Parameters
    ----------
    fnu : float
        nu, the order of the first member of the sequence of functions.

    z : complex
        z, the argument of the functions.

    n : int
        N, the number of members required in the sequence
        Y_nu(z),Y_{nu+1}(z),...,Y_{nu+N-1}(z).

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The results are returned unscaled.

        `scal` = 'S'

            The results are returned scaled by the factor e^(-|Im(z)|).

    cy : complex, shape (`n`), modified in place
        `On exit`: the N required function values: `cy`[i-1] contains
        Y_{nu+i-1}(z), for i = 1,2,...,N.

    Returns
    -------
    nz : int
        The number of components of `cy` that are set to zero due to
        underflow. The positions of such components in the array `cy`
        are arbitrary.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `z` = (0.0,0.0).

        (`errno` 1)
            On entry, `fnu` = *<value>*.

            Constraint: `fnu` >= 0.0.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            No computation because abs(`z`) = *<value>* < *<value>*.

        (`errno` 2)
            No computation because Re(`z`) = *<value>* > *<value>*,
            `scal` = 'U'.

        (`errno` 3)
            No computation because `fnu`+`n`-1 = *<value>* is too large.

        (`errno` 5)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 5)
            No computation because `fnu`+`n`-1 = *<value>* > *<value>*.

        (`errno` 6)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

        (`errno` 4)
            Results lack precision because `fnu`+`n`-1 = *<value>* >
            *<value>*.

    Notes
    -----
    ``bessel_y_complex`` evaluates a sequence of values for the Bessel
    function Y_nu(z), where z is complex, -pi < argz <= pi, and nu is
    the real, non-negative order.
    The N-member sequence is generated for orders nu, nu+1,...,nu+N-1.
    Optionally, the sequence is scaled by the factor e^(-|Im(z)|).

    **Note:** although the function may not be called with nu less than
    zero, for negative orders the formula Y_{-nu}(z) =
    Y_nu(z)cos(pinu)+J_nu(z)sin(pinu) may be used (for the Bessel
    function J_nu(z), see ``bessel_j_complex``).

    The function is derived from the function CBESY in Amos (1986).
    It is based on the relation Y_nu(z) = (H_nu^(1)(z)-H_nu^(2)(z))/2i,
    where H_nu^(1)(z) and H_nu^(2)(z) are the Hankel functions of the
    first and second kinds respectively (see ``hankel_complex``).

    When N is greater than 1, extra values of Y_nu(z) are computed using
    recurrence relations.

    For very large |z| or (nu+N-1), argument reduction will cause total
    loss of accuracy, and so no computation is performed.
    For slightly smaller |z| or (nu+N-1), the computation is performed
    but results are accurate to less than half of machine precision.
    If |z| is very small, near the machine underflow threshold, or
    (nu+N-1) is too large, there is a risk of overflow and so no
    computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_y_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, fnu.
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._EngineComplex128ArrayType.c_type, # Output, cy[n].
        utils._BYREF_INT, # Output, nz.
        utils._EngineComplex128ArrayType.c_type, # Workspace, cwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _fnu = utils._EngineFloat64ScalarType(
        fnu, {'fun_name': fun_name, 'entity': 'fnu'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _cy = _cy_cls.to_ctypes(can_reference_contents=False)
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _cwrk = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _fnu,
        _z,
        _n,
        _scal,
        _cy,
        _nz,
        _cwrk,
        _errbuf,
        _ierr,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, z = (0.0,0.0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fnu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fnu >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} < {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'Re(z) = {:s} > {:s}, scal = \'U\'.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because fnu+n-1 = {:s} is too large.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1411:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cy_cls.refresh_py(_cy)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _nz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nz'],
            )(
                _nz_py,
            )
    _err_class.manage()
    return return_data

def bessel_j_complex(fnu, z, n, scal, cy):
    u"""
    Bessel functions J_{nu+a}(z), real a >= 0, complex z, nu =
    0,1,2,....

    ``bessel_j_complex`` returns a sequence of values for the Bessel
    functions J_{nu+n}(z) for complex z, non-negative nu and n =
    0,1,...,N-1, with an option for exponential scaling.

    For full information please refer to the NAG Library document for
    s17de

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17def.html

    Parameters
    ----------
    fnu : float
        nu, the order of the first member of the sequence of functions.

    z : complex
        The argument z of the functions.

    n : int
        N, the number of members required in the sequence
        J_nu(z),J_{nu+1}(z),...,J_{nu+N-1}(z).

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The results are returned unscaled.

        `scal` = 'S'

            The results are returned scaled by the factor e^(-|Im(z)|).

    cy : complex, shape (`n`), modified in place
        `On exit`: the N required function values: `cy`[i-1] contains
        J_{nu+i-1}(z), for i = 1,2,...,N.

    Returns
    -------
    nz : int
        The number of components of `cy` that are set to zero due to
        underflow. If `nz` > 0, then elements `cy`[`n`-`nz`],
        `cy`[`n`-`nz`+1],...,`cy`[`n`-1] are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `fnu` = *<value>*.

            Constraint: `fnu` >= 0.0.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            No computation because Im(`z`) = *<value>* > *<value>*,
            `scal` = 'U'.

        (`errno` 4)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 4)
            No computation because `fnu`+`n`-1 = *<value>* > *<value>*.

        (`errno` 5)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

        (`errno` 3)
            Results lack precision because `fnu`+`n`-1 = *<value>* >
            *<value>*.

    Notes
    -----
    ``bessel_j_complex`` evaluates a sequence of values for the Bessel
    function J_nu(z), where z is complex, -pi < argz <= pi, and nu is
    the real, non-negative order.
    The N-member sequence is generated for orders nu, nu+1,...,nu+N-1.
    Optionally, the sequence is scaled by the factor e^(-|Im(z)|).

    **Note:** although the function may not be called with nu less than
    zero, for negative orders the formula J_{-nu}(z) =
    J_nu(z)cos(pinu)-Y_nu(z)sin(pinu) may be used (for the Bessel
    function Y_nu(z), see ``bessel_y_complex``).

    The function is derived from the function CBESJ in Amos (1986).
    It is based on the relations J_nu(z) = e^(nupi i/2)I_nu(-iz), Im(z)
    >= 0.0, and J_nu(z) = e^(-nupi i/2)I_nu(iz), Im(z) < 0.0.

    The Bessel function I_nu(z) is computed using a variety of
    techniques depending on the region under consideration.

    When N is greater than 1, extra values of J_nu(z) are computed using
    recurrence relations.

    For very large |z| or (nu+N-1), argument reduction will cause total
    loss of accuracy, and so no computation is performed.
    For slightly smaller |z| or (nu+N-1), the computation is performed
    but results are accurate to less than half of machine precision.
    If Im(z) is large, there is a risk of overflow and so no computation
    is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, fnu.
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._EngineComplex128ArrayType.c_type, # Output, cy[n].
        utils._BYREF_INT, # Output, nz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _fnu = utils._EngineFloat64ScalarType(
        fnu, {'fun_name': fun_name, 'entity': 'fnu'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _cy = _cy_cls.to_ctypes(can_reference_contents=False)
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _fnu,
        _z,
        _n,
        _scal,
        _cy,
        _nz,
        _errbuf,
        _ierr,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fnu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fnu >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'Im(z) = {:s} > {:s}, scal = \'U\'.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cy_cls.refresh_py(_cy)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _nz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nz'],
            )(
                _nz_py,
            )
    _err_class.manage()
    return return_data

def airy_ai_complex(deriv, z, scal):
    u"""
    Airy functions Ai(z) and Ai'(z), complex z.

    ``airy_ai_complex`` returns the value of the Airy function Ai(z) or
    its derivative Ai'(z) for complex z, with an option for exponential
    scaling.

    For full information please refer to the NAG Library document for
    s17dg

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17dgf.html

    Parameters
    ----------
    deriv : str, length 1
        Specifies whether the function or its derivative is required.

        `deriv` = 'F'

            Ai(z) is returned.

        `deriv` = 'D'

            Ai'(z) is returned.

    z : complex
        The argument z of the function.

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The result is returned unscaled.

        `scal` = 'S'

            The result is returned scaled by the factor
            e^(2z*sqrt(z)/3).

    Returns
    -------
    ai : complex
        The required function or derivative value.

    nz : int
        Indicates whether or not `ai` is set to zero due to underflow.
        This can only occur when `scal` = 'U'.

        `nz` = 0

            `ai` is not set to zero.

        `nz` = 1

            `ai` is set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `deriv` has an illegal value: `deriv` = *<value>*.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 2)
            No computation because Re(omega) too large, where omega =
            (2/3)*`z`^(3/2).

        (`errno` 4)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 5)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

    Notes
    -----
    ``airy_ai_complex`` returns a value for the Airy function Ai(z) or
    its derivative Ai'(z), where z is complex, -pi < argz <= pi.
    Optionally, the value is scaled by the factor e^(2z*sqrt(z)/3).

    The function is derived from the function CAIRY in Amos (1986).
    It is based on the relations Ai(z) = sqrt(z)K_{1/3}(w)/(pi*sqrt(3)),
    and Ai'(z) = (-zK_{2/3}(w))/(pi*sqrt(3)), where K_nu is the modified
    Bessel function and w = 2z*sqrt(z)/3.

    For very large |z|, argument reduction will cause total loss of
    accuracy, and so no computation is performed.
    For slightly smaller |z|, the computation is performed but results
    are accurate to less than half of machine precision.
    If Re(w) is too large, and the unscaled function is required, there
    is a risk of overflow and so no computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_ai_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17dgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, deriv(len=1).
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._BYREF_COMPLEX128, # Output, ai.
        utils._BYREF_INT, # Output, nz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of deriv.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _deriv = utils._EngineCharScalarType(
        deriv, {'fun_name': fun_name, 'entity': 'deriv'}, exp_chrlen=1,
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _ai = utils._EngineComplex128ScalarType.empty_ctypes()
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _deriv,
        _z,
        _scal,
        _ai,
        _nz,
        _errbuf,
        _ierr,
        _deriv._length_,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, deriv has an illegal value: deriv = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because Re(omega) too large, where omega = (2/3)*z^(3/2).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because Re(omega) too large, where omega = (2/3)*z^(3/2).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ai_py = utils._EngineComplex128ScalarType.to_py(_ai)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ai', 'nz'],
        )(
            _ai_py,
            _nz_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def airy_bi_complex(deriv, z, scal):
    u"""
    Airy functions Bi(z) and Bi'(z), complex z.

    ``airy_bi_complex`` returns the value of the Airy function Bi(z) or
    its derivative Bi'(z) for complex z, with an option for exponential
    scaling.

    For full information please refer to the NAG Library document for
    s17dh

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17dhf.html

    Parameters
    ----------
    deriv : str, length 1
        Specifies whether the function or its derivative is required.

        `deriv` = 'F'

            Bi(z) is returned.

        `deriv` = 'D'

            Bi'(z) is returned.

    z : complex
        The argument z of the function.

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The result is returned unscaled.

        `scal` = 'S'

            The result is returned scaled by the factor
            e^(|Re(2z*sqrt(z)/3)|).

    Returns
    -------
    bi : complex
        The required function or derivative value.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `deriv` has an illegal value: `deriv` = *<value>*.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 2)
            No computation because Re(`z`) = *<value>* is too large when
            `scal` = 'U'.

        (`errno` 4)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 5)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

    Notes
    -----
    ``airy_bi_complex`` returns a value for the Airy function Bi(z) or
    its derivative Bi'(z), where z is complex, -pi < argz <= pi.
    Optionally, the value is scaled by the factor
    e^(|Re(2z*sqrt(z)/3)|).

    The function is derived from the function CBIRY in Amos (1986).
    It is based on the relations Bi(z) =
    sqrt(z)/(sqrt(3))(I_{-1/3}(w)+I_{1/3}(w)), and Bi'(z) =
    z/(sqrt(3))(I_{-2/3}(w)+I_{2/3}(w)), where I_nu is the modified
    Bessel function and w = 2z*sqrt(z)/3.

    For very large |z|, argument reduction will cause total loss of
    accuracy, and so no computation is performed.
    For slightly smaller |z|, the computation is performed but results
    are accurate to less than half of machine precision.
    If Re(z) is too large, and the unscaled function is required, there
    is a risk of overflow and so no computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.airy_bi_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17dhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, deriv(len=1).
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._BYREF_COMPLEX128, # Output, bi.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of deriv.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _deriv = utils._EngineCharScalarType(
        deriv, {'fun_name': fun_name, 'entity': 'deriv'}, exp_chrlen=1,
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _bi = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _deriv,
        _z,
        _scal,
        _bi,
        _errbuf,
        _ierr,
        _deriv._length_,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, deriv has an illegal value: deriv = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because Re(z) = {:s} is too large when'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'scal = \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _bi_py = utils._EngineComplex128ScalarType.to_py(_bi)
        return_data = _bi_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['bi'],
            )(
                _bi_py,
            )
    _err_class.manage()
    return return_data

def hankel_complex(m, fnu, z, n, scal, cy):
    u"""
    Hankel functions H_nu+a^(j)(z), j = 1,2, real a >= 0, complex z, nu
    = 0,1,2,....

    ``hankel_complex`` returns a sequence of values for the Hankel
    functions H_nu+n^(1)(z) or H_nu+n^(2)(z) for complex z, non-negative
    nu and n = 0,1,...,N-1, with an option for exponential scaling.

    For full information please refer to the NAG Library document for
    s17dl

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17dlf.html

    Parameters
    ----------
    m : int
        The kind of functions required.

        `m` = 1

            The functions are H_nu^(1)(z).

        `m` = 2

            The functions are H_nu^(2)(z).

    fnu : float
        nu, the order of the first member of the sequence of functions.

    z : complex
        The argument z of the functions.

    n : int
        N, the number of members required in the sequence
        H_nu^(`m`)(z),H_nu+1^(`m`)(z),...,H_nu+N-1^(`m`)(z).

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The results are returned unscaled.

        `scal` = 'S'

            The results are returned scaled by the factor e^(-iz) when
            `m` = 1, or by the factor e^(iz) when `m` = 2.

    cy : complex, shape (`n`), modified in place
        `On exit`: the N required function values: `cy`[i-1] contains
        H_nu+i-1^(`m`)(z), for i = 1,2,...,N.

    Returns
    -------
    nz : int
        The number of components of `cy` that are set to zero due to
        underflow. If `nz` > 0, then if Im(z) > 0.0 and `m` = 1, or
        Im(z) < 0.0 and `m` = 2, elements
        `cy`[0],`cy`[1],...,`cy`[`nz`-1] are set to zero. In the
        complementary half-planes, `nz` simply states the number of
        underflows, and not which elements they are.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `z` = (0.0,0.0).

        (`errno` 1)
            On entry, `fnu` = *<value>*.

            Constraint: `fnu` >= 0.0.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` has illegal value: `m` = *<value>*.

        (`errno` 2)
            No computation because abs(`z`) = *<value>* < *<value>*.

        (`errno` 3)
            No computation because `fnu`+`n`-1 = *<value>* is too large.

        (`errno` 5)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 5)
            No computation because `fnu`+`n`-1 = *<value>* > *<value>*.

        (`errno` 6)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

        (`errno` 4)
            Results lack precision, `fnu`+`n`-1 = *<value>* > *<value>*.

    Notes
    -----
    ``hankel_complex`` evaluates a sequence of values for the Hankel
    function H_nu^(1)(z) or H_nu^(2)(z), where z is complex, -pi < argz
    <= pi, and nu is the real, non-negative order.
    The N-member sequence is generated for orders nu, nu+1,...,nu+N-1.
    Optionally, the sequence is scaled by the factor e^(-iz) if the
    function is H_nu^(1)(z) or by the factor e^(iz) if the function is
    H_nu^(2)(z).

    **Note:** although the function may not be called with nu less than
    zero, for negative orders the formulae H_-nu^(1)(z) = e^(nupi
    i)H_nu^(1)(z), and H_-nu^(2)(z) = e^(-nupi i)H_nu^(2)(z) may be
    used.

    The function is derived from the function CBESH in Amos (1986).
    It is based on the relation

        H_nu^(m)(z) = 1/pe^(-p nu)K_nu(ze^(-p)),

    where p = i pi/2 if m = 1 and p = -i pi/2 if m = 2, and the Bessel
    function K_nu(z) is computed in the right half-plane only.
    Continuation of K_nu(z) to the left half-plane is computed in terms
    of the Bessel function I_nu(z).
    These functions are evaluated using a variety of different
    techniques, depending on the region under consideration.

    When N is greater than 1, extra values of H_nu^(m)(z) are computed
    using recurrence relations.

    For very large |z| or (nu+N-1), argument reduction will cause total
    loss of accuracy, and so no computation is performed.
    For slightly smaller |z| or (nu+N-1), the computation is performed
    but results are accurate to less than half of machine precision.
    If |z| is very small, near the machine underflow threshold, or
    (nu+N-1) is too large, there is a risk of overflow and so no
    computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.hankel_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17dlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, fnu.
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._EngineComplex128ArrayType.c_type, # Output, cy[n].
        utils._BYREF_INT, # Output, nz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _fnu = utils._EngineFloat64ScalarType(
        fnu, {'fun_name': fun_name, 'entity': 'fnu'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _cy = _cy_cls.to_ctypes(can_reference_contents=False)
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _fnu,
        _z,
        _n,
        _scal,
        _cy,
        _nz,
        _errbuf,
        _ierr,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, z = (0.0,0.0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fnu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fnu >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m has illegal value: m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} < {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because fnu+n-1 = {:s} is too large.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision,',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1411:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1412:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cy_cls.refresh_py(_cy)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _nz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nz'],
            )(
                _nz_py,
            )
    _err_class.manage()
    return return_data

def struve_h0(x):
    u"""
    Struve function of order 0, H_0(x).

    ``struve_h0`` returns the value of the Struve function of order 0,
    H_0(x).

    For full information please refer to the NAG Library document for
    s17ga

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17gaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    h0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `x` is too large and the function returns zero.

    Notes
    -----
    ``struve_h0`` evaluates an approximation to the Struve function of
    order zero, H_0(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_h0'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, h0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _h0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _h0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'x is too large and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _h0_py = utils._EngineFloat64ScalarType.to_py(_h0)
        return_data = _h0_py
    _err_class.manage()
    return return_data

def struve_h1(x):
    u"""
    Struve function of order 1, H_1(x).

    ``struve_h1`` returns the value of the Struve function of order 1,
    H_1(x).

    For full information please refer to the NAG Library document for
    s17gb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s17gbf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    h1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `x` is too large and the function returns zero.

    Notes
    -----
    ``struve_h1`` evaluates an approximation to the Struve function of
    order 1, H_1(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_h1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s17gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, h1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _h1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _h1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'x is too large and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _h1_py = utils._EngineFloat64ScalarType.to_py(_h1)
        return_data = _h1_py
    _err_class.manage()
    return return_data

def bessel_k0_real(x):
    u"""
    Modified Bessel function K_0(x).

    ``bessel_k0_real`` returns the value of the modified Bessel function
    K_0(x).

    For full information please refer to the NAG Library document for
    s18ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    k0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            K_0 is undefined and the function returns zero.

    Notes
    -----
    ``bessel_k0_real`` evaluates an approximation to the modified Bessel
    function of the second kind K_0(x).

    **Note:**  K_0(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on five Chebyshev expansions:

    For 0 < x <= 1,

        K_0(x) = -ln(x)sum'_{r = 0}a_r T_r(t)+sum'_{r = 0}b_r T_r(t),
        where t = 2x^2-1.

    For 1 < x <= 2,

        K_0(x) = e^(-x)sum'_{r = 0}c_r T_r(t), where t = 2x-3.

    For 2 < x <= 4,

        K_0(x) = e^(-x)sum'_{r = 0}d_r T_r(t), where t = x-3.

    For x > 4,

        K_0(x) = (e^(-x))/sqrt(x)sum'_{r = 0}e_r T_r(t),where t =
        (9-x)/(1+x).

    For x near zero, K_0(x) ~= -gamma-ln(x/2), where gamma denotes
    Euler's constant.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, where there is a danger of underflow due to the
    smallness of K_0, the result is set exactly to zero.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k0_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, k0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _k0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _k0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'K_0 is undefined and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _k0_py = utils._EngineFloat64ScalarType.to_py(_k0)
        return_data = _k0_py
    _err_class.manage()
    return return_data

def bessel_k1_real(x):
    u"""
    Modified Bessel function K_1(x).

    ``bessel_k1_real`` returns the value of the modified Bessel function
    K_1(x).

    For full information please refer to the NAG Library document for
    s18ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    k1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            K_0 is undefined and the function returns zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` > *<value>*.

            `x` is too small, there is a danger of overflow and the
            function returns approximately the largest representable
            value.

    Notes
    -----
    ``bessel_k1_real`` evaluates an approximation to the modified Bessel
    function of the second kind K_1(x).

    **Note:**  K_1(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on five Chebyshev expansions:

    For 0 < x <= 1,

        K_1(x) = 1/x+xln(x)sum'_{r = 0}a_r T_r(t)-xsum'_{r = 0}b_r
        T_r(t), where t = 2x^2-1.

    For 1 < x <= 2,

        K_1(x) = e^(-x)sum'_{r = 0}c_r T_r(t), where t = 2x-3.

    For 2 < x <= 4,

        K_1(x) = e^(-x)sum'_{r = 0}d_r T_r(t), where t = x-3.

    For x > 4,

        K_1(x) = (e^(-x))/sqrt(x)sum'_{r = 0}e_r T_r(t), where t =
        (9-x)/(1+x).

    For x near zero, K_1(x) ~= 1/x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For very small x on some machines, it is impossible to calculate 1/x
    without overflow and the function must fail.

    For large x, where there is a danger of underflow due to the
    smallness of K_1, the result is set exactly to zero.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k1_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, k1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _k1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _k1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'K_0 is undefined and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too small, there is a danger of overflow and the function returns',
                    'approximately the largest representable value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _k1_py = utils._EngineFloat64ScalarType.to_py(_k1)
        return_data = _k1_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['k1'],
            )(
                _k1_py,
            )
    _err_class.manage()
    return return_data

def bessel_i0_real(x):
    u"""
    Modified Bessel function I_0(x).

    ``bessel_i0_real`` returns the value of the modified Bessel function
    I_0(x).

    For full information please refer to the NAG Library document for
    s18ae

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18aef.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    i0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large and the function returns the
            approximate value of I_0(x) at the nearest valid argument.

    Notes
    -----
    ``bessel_i0_real`` evaluates an approximation to the modified Bessel
    function of the first kind I_0(x).

    **Note:**  I_0(-x) = I_0(x), so the approximation need only consider
    x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 4,

        I_0(x) = e^xsum'_{r = 0}a_r T_r(t), where t = 2*(x/4)-1.

    For 4 < x <= 12,

        I_0(x) = e^xsum'_{r = 0}b_r T_r(t), where t = (x-8)/4.

    For x > 12,

        I_0(x) = (e^x)/sqrt(x)sum'_{r = 0}c_r T_r(t), where t =
        2*(12/x)-1.

    For small x, I_0(x) ~= 1.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, the function must fail because of the danger of
    overflow in calculating e^x.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i0_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, i0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _i0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _i0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large and the function returns the approximate value of',
                    'I_0(x) at the nearest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _i0_py = utils._EngineFloat64ScalarType.to_py(_i0)
        return_data = _i0_py
    _err_class.manage()
    return return_data

def bessel_i1_real(x):
    u"""
    Modified Bessel function I_1(x).

    ``bessel_i1_real`` returns a value for the modified Bessel function
    I_1(x).

    For full information please refer to the NAG Library document for
    s18af

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18aff.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    i1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large and the function returns the
            approximate value of I_1(x) at the nearest valid argument.

    Notes
    -----
    ``bessel_i1_real`` evaluates an approximation to the modified Bessel
    function of the first kind I_1(x).

    **Note:**  I_1(-x) = -I_1(x), so the approximation need only
    consider x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 4,

        I_1(x) = xsum'_{r = 0}a_r T_r(t), where t = 2(x/4)^2-1;

    For 4 < x <= 12,

        I_1(x) = e^xsum'_{r = 0}b_r T_r(t), where t = (x-8)/4;

    For x > 12,

        I_1(x) = (e^x)/sqrt(x)sum'_{r = 0}c_r T_r(t), where t =
        2*(12/x)-1.

    For small x, I_1(x) ~= x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, the function must fail because I_1(x) cannot be
    represented without overflow.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i1_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, i1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _i1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _i1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large and the function returns the approximate value of',
                    'I_1(x) at the nearest valid argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _i1_py = utils._EngineFloat64ScalarType.to_py(_i1)
        return_data = _i1_py
    _err_class.manage()
    return return_data

def bessel_k0_real_vector(n, x, f, ivalid):
    u"""
    Modified Bessel function vectorized K_0(x).

    ``bessel_k0_real_vector`` returns an array of values of the modified
    Bessel function K_0(x).

    For full information please refer to the NAG Library document for
    s18aq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18aqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: K_0(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i <= 0.0, K_0(x_i) is undefined. `f`[i-1] contains 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_k0_real_vector`` evaluates an approximation to the modified
    Bessel function of the second kind K_0(x_i) for an array of
    arguments x_i, for i = 1,2,...,n.

    **Note:**  K_0(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on five Chebyshev expansions:

    For 0 < x <= 1,

        K_0(x) = -ln(x)sum'_{r = 0}a_r T_r(t)+sum'_{r = 0}b_r T_r(t),
        where t = 2x^2-1.

    For 1 < x <= 2,

        K_0(x) = e^(-x)sum'_{r = 0}c_r T_r(t), where t = 2x-3.

    For 2 < x <= 4,

        K_0(x) = e^(-x)sum'_{r = 0}d_r T_r(t), where t = x-3.

    For x > 4,

        K_0(x) = (e^(-x))/sqrt(x)sum'_{r = 0}e_r T_r(t),where t =
        (9-x)/(1+x).

    For x near zero, K_0(x) ~= -gamma-ln(x/2), where gamma denotes
    Euler's constant.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, where there is a danger of underflow due to the
    smallness of K_0, the result is set exactly to zero.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k0_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18aqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_k1_real_vector(n, x, f, ivalid):
    u"""
    Modified Bessel function vectorized K_1(x).

    ``bessel_k1_real_vector`` returns an array of values of the modified
    Bessel function K_1(x).

    For full information please refer to the NAG Library document for
    s18ar

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18arf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: K_1(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i <= 0.0, K_1(x_i) is undefined. `f`[i-1] contains 0.0.

        `ivalid`[i-1] = 2

            x_i is too small, there is a danger of overflow. `f`[i-1]
            contains zero. The threshold value is the same as for
            `errno` = 2 in ``bessel_k1_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_k1_real_vector`` evaluates an approximation to the modified
    Bessel function of the second kind K_1(x_i) for an array of
    arguments x_i, for i = 1,2,...,n.

    **Note:**  K_1(x) is undefined for x <= 0 and the function will fail
    for such arguments.

    The function is based on five Chebyshev expansions:

    For 0 < x <= 1,

        K_1(x) = 1/x+xln(x)sum'_{r = 0}a_r T_r(t)-xsum'_{r = 0}b_r
        T_r(t), where t = 2x^2-1.

    For 1 < x <= 2,

        K_1(x) = e^(-x)sum'_{r = 0}c_r T_r(t), where t = 2x-3.

    For 2 < x <= 4,

        K_1(x) = e^(-x)sum'_{r = 0}d_r T_r(t), where t = x-3.

    For x > 4,

        K_1(x) = (e^(-x))/sqrt(x)sum'_{r = 0}e_r T_r(t), where t =
        (9-x)/(1+x).

    For x near zero, K_1(x) ~= 1/x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For very small x it is impossible to calculate 1/x without overflow
    and the function must fail.

    For large x, where there is a danger of underflow due to the
    smallness of K_1, the result is set exactly to zero.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k1_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18arft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_i0_real_vector(n, x, f, ivalid):
    u"""
    Modified Bessel function vectorized I_0(x).

    ``bessel_i0_real_vector`` returns an array of values of the modified
    Bessel function I_0(x).

    For full information please refer to the NAG Library document for
    s18as

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18asf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: I_0(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large. `f`[i-1] contains the approximate value of
            I_0(x_i) at the nearest valid argument. The threshold value
            is the same as for `errno` = 1 in ``bessel_i0_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_i0_real_vector`` evaluates an approximation to the modified
    Bessel function of the first kind I_0(x_i) for an array of arguments
    x_i, for i = 1,2,...,n.

    **Note:**  I_0(-x) = I_0(x), so the approximation need only consider
    x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 4,

        I_0(x) = e^xsum'_{r = 0}a_r T_r(t), where t = 2*(x/4)-1.

    For 4 < x <= 12,

        I_0(x) = e^xsum'_{r = 0}b_r T_r(t), where t = (x-8)/4.

    For x > 12,

        I_0(x) = (e^x)/sqrt(x)sum'_{r = 0}c_r T_r(t), where t =
        2*(12/x)-1.

    For small x, I_0(x) ~= 1.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, the function must fail because of the danger of
    overflow in calculating e^x.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i0_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18asft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_i1_real_vector(n, x, f, ivalid):
    u"""
    Modified Bessel function vectorized I_1(x).

    ``bessel_i1_real_vector`` returns an array of values for the
    modified Bessel function I_1(x).

    For full information please refer to the NAG Library document for
    s18at

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18atf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: I_1(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large. `f`[i-1] contains the approximate value of
            I_1(x_i) at the nearest valid argument. The threshold value
            is the same as for `errno` = 1 in ``bessel_i1_real``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_i1_real_vector`` evaluates an approximation to the modified
    Bessel function of the first kind I_1(x_i) for an array of arguments
    x_i, for i = 1,2,...,n.

    **Note:**  I_1(-x) = -I_1(x), so the approximation need only
    consider x >= 0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 4,

        I_1(x) = xsum'_{r = 0}a_r T_r(t), where t = 2(x/4)^2-1;

    For 4 < x <= 12,

        I_1(x) = e^xsum'_{r = 0}b_r T_r(t), where t = (x-8)/4;

    For x > 12,

        I_1(x) = (e^x)/sqrt(x)sum'_{r = 0}c_r T_r(t), where t =
        2*(12/x)-1.

    For small x, I_1(x) ~= x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, the function must fail because I_1(x) cannot be
    represented without overflow.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i1_real_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18atft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_k0_scaled(x):
    u"""
    Scaled modified Bessel function e^xK_0(x).

    ``bessel_k0_scaled`` returns a value of the scaled modified Bessel
    function e^xK_0(x).

    For full information please refer to the NAG Library document for
    s18cc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18ccf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    sk0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

    Notes
    -----
    ``bessel_k0_scaled`` evaluates an approximation to e^xK_0(x), where
    K_0 is a modified Bessel function of the second kind.
    The scaling factor e^x removes most of the variation in K_0(x).

    The function uses the same Chebyshev expansions as
    ``bessel_k0_real``, which returns the unscaled value of K_0(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k0_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, sk0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _sk0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _sk0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sk0_py = utils._EngineFloat64ScalarType.to_py(_sk0)
        return_data = _sk0_py
    _err_class.manage()
    return return_data

def bessel_k1_scaled(x):
    u"""
    Scaled modified Bessel function e^xK_1(x).

    ``bessel_k1_scaled`` returns a value of the scaled modified Bessel
    function e^xK_1(x).

    For full information please refer to the NAG Library document for
    s18cd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18cdf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    sk1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            K_1 is undefined and the function returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` > *<value>*.

            The function returns the value of the function at the
            smallest permitted value of the argument.

    Notes
    -----
    ``bessel_k1_scaled`` evaluates an approximation to e^xK_1(x), where
    K_1 is a modified Bessel function of the second kind.
    The scaling factor e^x removes most of the variation in K_1(x).

    The function uses the same Chebyshev expansions as
    ``bessel_k1_real``, which returns the unscaled value of K_1(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k1_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, sk1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _sk1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _sk1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'K_1 is undefined and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The function returns the value of the function at the smallest',
                    'permitted value of the argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sk1_py = utils._EngineFloat64ScalarType.to_py(_sk1)
        return_data = _sk1_py
    _err_class.manage()
    return return_data

def bessel_i0_scaled(x):
    u"""
    Scaled modified Bessel function e^(-|x|)I_0(x).

    ``bessel_i0_scaled`` returns a value of the scaled modified Bessel
    function e^(-|x|)I_0(x).

    For full information please refer to the NAG Library document for
    s18ce

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18cef.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    si0 : float
        The value of the function at x.

    Notes
    -----
    ``bessel_i0_scaled`` evaluates an approximation to e^(-|x|)I_0(x),
    where I_0 is a modified Bessel function of the first kind.
    The scaling factor e^(-|x|) removes most of the variation in I_0(x).

    The function uses the same Chebyshev expansions as
    ``bessel_i0_real``, which returns the unscaled value of I_0(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i0_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, si0.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _si0 = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _si0,
    )
    utils._handle_cb_exception(excs)
    _si0_py = utils._EngineFloat64ScalarType.to_py(_si0)
    return_data = _si0_py
    return return_data

def bessel_i1_scaled(x):
    u"""
    Scaled modified Bessel function e^(-|x|)I_1(x).

    ``bessel_i1_scaled`` returns a value of the scaled modified Bessel
    function e^(-|x|)I_1(x).

    For full information please refer to the NAG Library document for
    s18cf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18cff.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    si1 : float
        The value of the function at x.

    Notes
    -----
    ``bessel_i1_scaled`` evaluates an approximation to e^(-|x|)I_1(x),
    where I_1 is a modified Bessel function of the first kind.
    The scaling factor e^(-|x|) removes most of the variation in I_1(x).

    The function uses the same Chebyshev expansions as
    ``bessel_i1_real``, which returns the unscaled value of I_1(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i1_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18cfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, si1.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _si1 = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _si1,
    )
    utils._handle_cb_exception(excs)
    _si1_py = utils._EngineFloat64ScalarType.to_py(_si1)
    return_data = _si1_py
    return return_data

def bessel_k0_scaled_vector(n, x, f, ivalid):
    u"""
    Scaled modified Bessel function vectorized e^xK_0(x).

    ``bessel_k0_scaled_vector`` returns an array of values of the scaled
    modified Bessel function e^xK_0(x).

    For full information please refer to the NAG Library document for
    s18cq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18cqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: e^(x_i)K_0(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i <= 0.0, K_0(x_i) is undefined. `f`[i-1]
            contains 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_k0_scaled_vector`` evaluates an approximation to
    e^(x_i)K_0(x_i), where K_0 is a modified Bessel function of the
    second kind for an array of arguments x_i, for i = 1,2,...,n.
    The scaling factor e^x removes most of the variation in K_0(x).

    The function uses the same Chebyshev expansions as
    ``bessel_k0_real_vector``, which returns an array of the unscaled
    values of K_0(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k0_scaled_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18cqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_k1_scaled_vector(n, x, f, ivalid):
    u"""
    Scaled modified Bessel function vectorized e^xK_1(x).

    ``bessel_k1_scaled_vector`` returns an array of values of the scaled
    modified Bessel function e^xK_1(x).

    For full information please refer to the NAG Library document for
    s18cr

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18crf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: e^(x_i)K_1(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, x_i <= 0.0, K_1(x_i) is undefined. `f`[i-1]
            contains 0.0.

        `ivalid`[i-1] = 2

            x_i is too close to zero, as determined by the value of the
            safe-range parameter ``machine.real_safe``: there is a
            danger of causing overflow. `f`[i-1] contains the reciprocal
            of the safe-range parameter.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``bessel_k1_scaled_vector`` evaluates an approximation to
    e^(x_i)K_1(x_i), where K_1 is a modified Bessel function of the
    second kind for an array of arguments x_i, for i = 1,2,...,n.
    The scaling factor e^x removes most of the variation in K_1(x).

    The function uses the same Chebyshev expansions as
    ``bessel_k1_real_vector``, which returns an array of the unscaled
    values of K_1(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k1_scaled_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18crft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_i0_scaled_vector(n, x, f):
    u"""
    Scaled modified Bessel function vectorized e^(-|x|)I_0(x).

    ``bessel_i0_scaled_vector`` returns an array of values of the scaled
    modified Bessel function e^(-|x|)I_0(x).

    For full information please refer to the NAG Library document for
    s18cs

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18csf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: e^(-|x_i|)I_0(x_i), the function values.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``bessel_i0_scaled_vector`` evaluates an approximation to
    e^(-|x_i|)I_0(x_i), where I_0 is a modified Bessel function of the
    first kind for an array of arguments x_i, for i = 1,2,...,n.
    The scaling factor e^(-|x|) removes most of the variation in I_0(x).

    The function uses the same Chebyshev expansions as
    ``bessel_i0_real_vector``, which returns an array of the unscaled
    values of I_0(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i0_scaled_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18csft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_i1_scaled_vector(n, x, f):
    u"""
    Scaled modified Bessel function vectorized e^(-|x|)I_1(x).

    ``bessel_i1_scaled_vector`` returns an array of values of the scaled
    modified Bessel function e^(-|x|)I_1(x).

    For full information please refer to the NAG Library document for
    s18ct

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18ctf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: e^(-|x_i|)I_1(x_i), the function values.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``bessel_i1_scaled_vector`` evaluates an approximation to
    e^(-|x_i|)I_1(x_i), where I_1 is a modified Bessel function of the
    first kind for an array of arguments x_i, for i = 1,2,...,n.
    The scaling factor e^(-|x|) removes most of the variation in I_1(x).

    The function uses the same Chebyshev expansions as
    ``bessel_i1_real_vector``, which returns an array of the unscaled
    values of I_1(x).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i1_scaled_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18ctft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def bessel_k_complex(fnu, z, n, scal, cy):
    u"""
    Modified Bessel functions K_{nu+a}(z), real a >= 0, complex z, nu =
    0,1,2,....

    ``bessel_k_complex`` returns a sequence of values for the modified
    Bessel functions K_{nu+n}(z) for complex z, non-negative nu and n =
    0,1,...,N-1, with an option for exponential scaling.

    For full information please refer to the NAG Library document for
    s18dc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18dcf.html

    Parameters
    ----------
    fnu : float
        nu, the order of the first member of the sequence of functions.

    z : complex
        The argument z of the functions.

    n : int
        N, the number of members required in the sequence
        K_nu(z),K_{nu+1}(z),...,K_{nu+N-1}(z).

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The results are returned unscaled.

        `scal` = 'S'

            The results are returned scaled by the factor e^z.

    cy : complex, shape (`n`), modified in place
        `On exit`: the N required function values: `cy`[i-1] contains
        K_{nu+i-1}(z), for i = 1,2,...,N.

    Returns
    -------
    nz : int
        The number of components of `cy` that are set to zero due to
        underflow. If `nz` > 0 and Re(z) >= 0.0, elements
        `cy`[0],`cy`[1],...,`cy`[`nz`-1] are set to zero. If Re(z) <
        0.0, `nz` simply states the number of underflows, and not which
        elements they are.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `z` = (0.0,0.0).

        (`errno` 1)
            On entry, `fnu` = *<value>*.

            Constraint: `fnu` >= 0.0.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            No computation because abs(`z`) = *<value>* < *<value>*.

        (`errno` 3)
            No computation because `fnu`+`n`-1 = *<value>* is too large.

        (`errno` 5)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 5)
            No computation because `fnu`+`n`-1 = *<value>* > *<value>*.

        (`errno` 6)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

        (`errno` 4)
            Results lack precision because `fnu`+`n`-1 = *<value>* >
            *<value>*.

    Notes
    -----
    ``bessel_k_complex`` evaluates a sequence of values for the modified
    Bessel function K_nu(z), where z is complex, -pi < argz <= pi, and
    nu is the real, non-negative order.
    The N-member sequence is generated for orders nu, nu+1,...,nu+N-1.
    Optionally, the sequence is scaled by the factor e^z.

    The function is derived from the function CBESK in Amos (1986).

    **Note:** although the function may not be called with nu less than
    zero, for negative orders the formula K_{-nu}(z) = K_nu(z) may be
    used.

    When N is greater than 1, extra values of K_nu(z) are computed using
    recurrence relations.

    For very large |z| or (nu+N-1), argument reduction will cause total
    loss of accuracy, and so no computation is performed.
    For slightly smaller |z| or (nu+N-1), the computation is performed
    but results are accurate to less than half of machine precision.
    If |z| is very small, near the machine underflow threshold, or
    (nu+N-1) is too large, there is a risk of overflow and so no
    computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_k_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, fnu.
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._EngineComplex128ArrayType.c_type, # Output, cy[n].
        utils._BYREF_INT, # Output, nz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _fnu = utils._EngineFloat64ScalarType(
        fnu, {'fun_name': fun_name, 'entity': 'fnu'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _cy = _cy_cls.to_ctypes(can_reference_contents=False)
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _fnu,
        _z,
        _n,
        _scal,
        _cy,
        _nz,
        _errbuf,
        _ierr,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, z = (0.0,0.0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fnu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fnu >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} < {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because fnu+n-1 = {:s} is too large.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1411:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1412:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cy_cls.refresh_py(_cy)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _nz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nz'],
            )(
                _nz_py,
            )
    _err_class.manage()
    return return_data

def bessel_i_complex(fnu, z, n, scal, cy):
    u"""
    Modified Bessel functions I_{nu+a}(z), real a >= 0, complex z, nu =
    0,1,2,....

    ``bessel_i_complex`` returns a sequence of values for the modified
    Bessel functions I_{nu+n}(z) for complex z, non-negative nu and n =
    0,1,...,N-1, with an option for exponential scaling.

    For full information please refer to the NAG Library document for
    s18de

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18def.html

    Parameters
    ----------
    fnu : float
        nu, the order of the first member of the sequence of functions.

    z : complex
        The argument z of the functions.

    n : int
        N, the number of members required in the sequence
        I_nu(z),I_{nu+1}(z),...,I_{nu+N-1}(z).

    scal : str, length 1
        The scaling option.

        `scal` = 'U'

            The results are returned unscaled.

        `scal` = 'S'

            The results are returned scaled by the factor e^(-|Re(z)|).

    cy : complex, shape (`n`), modified in place
        `On exit`: the N required function values: `cy`[i-1] contains
        I_{nu+i-1}(z), for i = 1,2,...,N.

    Returns
    -------
    nz : int
        The number of components of `cy` that are set to zero due to
        underflow.

        If `nz` > 0, then elements
        `cy`[`n`-`nz`],`cy`[`n`-`nz`+1],...,`cy`[`n`-1] are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `fnu` = *<value>*.

            Constraint: `fnu` >= 0.0.

        (`errno` 1)
            On entry, `scal` has an illegal value: `scal` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            No computation because Re(`z`) = *<value>* > *<value>*,
            `scal` = 'U'.

        (`errno` 4)
            No computation because abs(`z`) = *<value>* > *<value>*.

        (`errno` 4)
            No computation because `fnu`+`n`-1 = *<value>* > *<value>*.

        (`errno` 5)
            No computation -- algorithm termination condition not met.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Results lack precision because abs(`z`) = *<value>* >
            *<value>*.

        (`errno` 3)
            Results lack precision because `fnu`+`n`-1 = *<value>* >
            *<value>*.

    Notes
    -----
    ``bessel_i_complex`` evaluates a sequence of values for the modified
    Bessel function I_nu(z), where z is complex, -pi < argz <= pi, and
    nu is the real, non-negative order.
    The N-member sequence is generated for orders nu,nu+1,...,nu+N-1.
    Optionally, the sequence is scaled by the factor e^(-|Re(z)|).

    The function is derived from the function CBESI in Amos (1986).

    **Note:** although the function may not be called with nu less than
    zero, for negative orders the formula I_{-nu}(z) =
    I_nu(z)+2/pisin(pinu)K_nu(z) may be used (for the Bessel function
    K_nu(z), see ``bessel_k_complex``).

    When N is greater than 1, extra values of I_nu(z) are computed using
    recurrence relations.

    For very large |z| or (nu+N-1), argument reduction will cause total
    loss of accuracy, and so no computation is performed.
    For slightly smaller |z| or (nu+N-1), the computation is performed
    but results are accurate to less than half of machine precision.
    If Re(z) is too large and the unscaled function is required, there
    is a risk of overflow and so no computation is performed.
    In all the above cases, a warning is given by the function.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Amos, D E, 1986, `Algorithm 644: A portable package for Bessel
    functions of a complex argument and non-negative order`, ACM Trans.
    Math. Software (12), 265--273
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_i_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, fnu.
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, scal(len=1).
        utils._EngineComplex128ArrayType.c_type, # Output, cy[n].
        utils._BYREF_INT, # Output, nz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of scal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _fnu = utils._EngineFloat64ScalarType(
        fnu, {'fun_name': fun_name, 'entity': 'fnu'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _scal = utils._EngineCharScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _cy = _cy_cls.to_ctypes(can_reference_contents=False)
    _nz = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _fnu,
        _z,
        _n,
        _scal,
        _cy,
        _nz,
        _errbuf,
        _ierr,
        _scal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fnu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fnu >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, scal has an illegal value: scal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'Re(z) = {:s} > {:s}, scal = \'U\'.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Results lack precision because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1311:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'fnu+n-1 = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1312:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation because',
                    'abs(z) = {:s} > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No computation -- algorithm termination condition not met.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cy_cls.refresh_py(_cy)
        _nz_py = utils._EngineIntScalarType.to_py(_nz)
        return_data = _nz_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nz'],
            )(
                _nz_py,
            )
    _err_class.manage()
    return return_data

def struve_l0(x):
    u"""
    Modified Struve function of order 0, L_0(x).

    ``struve_l0`` returns the value of the modified Struve function of
    order 0, L_0(x).

    For full information please refer to the NAG Library document for
    s18ga

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18gaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    l0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The value of L_0(x) is larger than the largest positive
            floating-point number, which is returned.

    Notes
    -----
    ``struve_l0`` evaluates an approximation to the modified Struve
    function of order zero, L_0(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_l0'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, l0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _l0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _l0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of L_0(x) is larger than the largest positive floating-point',
                    'number, which is returned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _l0_py = utils._EngineFloat64ScalarType.to_py(_l0)
        return_data = _l0_py
    _err_class.manage()
    return return_data

def struve_l1(x):
    u"""
    Modified Struve function of order 1, L_1(x).

    ``struve_l1`` returns the value of the modified Struve function of
    order 1, L_1(x).

    For full information please refer to the NAG Library document for
    s18gb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18gbf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    l1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The value of L_1(x) is larger than the largest positive
            floating-point number, which is returned.

    Notes
    -----
    ``struve_l1`` evaluates an approximation to the modified Struve
    function of order 1, L_1(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_l1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, l1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _l1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _l1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of L_1(x) is larger than the largest positive floating-point',
                    'number, which is returned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _l1_py = utils._EngineFloat64ScalarType.to_py(_l1)
        return_data = _l1_py
    _err_class.manage()
    return return_data

def struve_i0ml0(x):
    u"""
    The function I_0(x)-L_0(x), where I_0(x) is a modified Bessel
    function and L_0(x) is a Struve function.

    ``struve_i0ml0`` returns the value of I_0(x)-L_0(x) where I_0(x) is
    the modified Bessel function of the first kind of order zero, and
    L_0(x) is the modified Struve function of order 0.

    For full information please refer to the NAG Library document for
    s18gc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18gcf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    iml0 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

    Notes
    -----
    ``struve_i0ml0`` evaluates an approximation to I_0(x)-L_0(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_i0ml0'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18gcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, iml0.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _iml0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _iml0,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iml0_py = utils._EngineFloat64ScalarType.to_py(_iml0)
        return_data = _iml0_py
    _err_class.manage()
    return return_data

def struve_i1ml1(x):
    u"""
    The function I_1(x)-L_1(x), where I_1(x) is a modified Bessel
    function and L_1(x) is a Struve function.

    ``struve_i1ml1`` returns the value of I_1(x)-L_1(x) where I_1(x) is
    the modified Bessel function of the first kind of order 1, and
    L_1(x) is the modified Struve function of order 1.

    For full information please refer to the NAG Library document for
    s18gd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18gdf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    iml1 : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

    Notes
    -----
    ``struve_i1ml1`` evaluates an approximation to I_1(x)-L_1(x).

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Struve function including special
    cases, transformations, relations and asymptotic approximations.

    The approximation method used by this function is based on Chebyshev
    expansions.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    MacLeod, A J, 1996, `MISCFUN, a software package to compute uncommon
    special functions`, ACM Trans. Math. Software (TOMS) (22(3)),
    288--301
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.struve_i1ml1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18gdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, iml1.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _iml1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _iml1,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iml1_py = utils._EngineFloat64ScalarType.to_py(_iml1)
        return_data = _iml1_py
    _err_class.manage()
    return return_data

def bessel_j_seq_complex(z, a, nl, b):
    u"""
    Bessel function of the 1st kind J_{alpha+/-n}(z).

    ``bessel_j_seq_complex`` returns a sequence of values for the Bessel
    functions J_{alpha+n-1}(z) or J_{alpha-n+1}(z) for complex z,
    non-negative alpha < 1 and n = 1,2,...,|N|+1.

    For full information please refer to the NAG Library document for
    s18gk

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s18gkf.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    a : float
        The order alpha of the first member in the required sequence of
        function values.

    nl : int
        The value of N.

    b : complex, shape (|`nl`|+1), modified in place
        `On exit`: with the function exits successfully or `errno` = 3,
        the required sequence of function values: `b`[n-1] contains
        J_{alpha+n-1}(z) if `nl` >= 0 and J_{alpha-n+1}(z) otherwise,
        for n = 1,2,...,abs(`nl`)+1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, abs(`nl`) = *<value>*.

            Constraint: abs(`nl`) <= 101.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` < 1.0.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` >= 0.0.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: when `nl` < 0, `z` != (0.0,0.0).

        (`errno` 2)
            Computation abandoned due to the likelihood of overflow.

        (`errno` 4)
            Computation abandoned due to total loss of precision.

        (`errno` 5)
            Computation abandoned due to failure to satisfy the
            termination condition.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Computation completed but some precision has been lost.

    Notes
    -----
    ``bessel_j_seq_complex`` evaluates a sequence of values for the
    Bessel function of the first kind J_alpha(z), where z is complex and
    nonzero and alpha is the order with 0 <= alpha < 1.
    The (|N|+1)-member sequence is generated for orders
    alpha,alpha+1,...,alpha+|N| when N >= 0.
    Note that + is replaced by - when N < 0.
    For positive orders the function may also be called with z = 0,
    since J_q(0) = 0 when q > 0.
    For negative orders the formula

        J_{-q}(z) = cos(pi q)J_q(z)-sin(pi q)Y_q(z)

    is used to generate the required sequence.
    The appropriate values of J_q(z) and Y_q(z) are obtained by calls to
    ``bessel_j_complex`` and ``bessel_y_complex``.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.bessel_j_seq_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s18gkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_INT, # Input, nl.
        utils._EngineComplex128ArrayType.c_type, # Output, b[abs(nl)+1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(abs(_nl_py)+1,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _a,
        _nl,
        _b,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(nl) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(nl) <= 101.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: when nl < 0, z != (0.0,0.0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation abandoned due to the likelihood of overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation completed but some precision has been lost.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation abandoned due to total loss of precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Computation abandoned due to failure to satisfy the termination condition.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kelvin_ber(x):
    u"""
    Kelvin function ber(x).

    ``kelvin_ber`` returns a value for the Kelvin function ber(x).

    For full information please refer to the NAG Library document for
    s19aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19aaf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    ber : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large for an accurate result to be returned
            and the function returns zero.

    Notes
    -----
    ``kelvin_ber`` evaluates an approximation to the Kelvin function
    ber(x).

    **Note:**  ber(-x) = ber(x), so the approximation need only consider
    x >= 0.0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 5,

        ber(x) = sum'_{r = 0}a_r T_r(t), with t = 2(x/5)^4-1.

    For x > 5,

        [table omitted]

    where alpha = x/(sqrt(2))-pi/8, beta = x/(sqrt(2))+pi/8,

    and a(t), b(t), c(t), and d(t) are expansions in the variable t =
    10/x-1.

    When x is sufficiently close to zero, the result is set directly to
    ber(0) = 1.0.

    For large x, there is a danger of the result being totally
    inaccurate, as the error amplification factor grows in an
    essentially exponential manner;, therefore, the function must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_ber'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ber.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ber = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ber,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large for an accurate result to be returned and the function',
                    'returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ber_py = utils._EngineFloat64ScalarType.to_py(_ber)
        return_data = _ber_py
    _err_class.manage()
    return return_data

def kelvin_bei(x):
    u"""
    Kelvin function bei(x).

    ``kelvin_bei`` returns a value for the Kelvin function bei(x).

    For full information please refer to the NAG Library document for
    s19ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19abf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    bei : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= *<value>*.

            abs(`x`) is too large for an accurate result to be returned
            and the function returns zero.

    Notes
    -----
    ``kelvin_bei`` evaluates an approximation to the Kelvin function
    bei(x).

    **Note:**  bei(-x) = bei(x), so the approximation need only consider
    x >= 0.0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 5,

        bei(x) = (x^2)/4sum'_r = 0a_r T_r(t), with t = 2(x/5)^4-1;

    For x > 5,

        bei(x) = (e^(x/sqrt(2)))/sqrt(2 pi
        x)[(1+1/xa(t))sin(alpha)-1/xb(t)cos(alpha)]

        +(e^(x/sqrt(2)))/sqrt(2 pi
        x)[(1+1/xc(t))cos(beta)-1/xd(t)sin(beta)]

    where alpha = x/(sqrt(2))-pi/8, beta = x/(sqrt(2))+pi/8,

    and a(t), b(t), c(t), and d(t) are expansions in the variable t =
    10/x-1.

    When x is sufficiently close to zero, the result is computed as
    bei(x) = (x^2)/4.
    If this result would underflow, the result returned is bei(x) = 0.0.

    For large x, there is a danger of the result being totally
    inaccurate, as the error amplification factor grows in an
    essentially exponential manner;, therefore, the function must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_bei'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, bei.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _bei = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _bei,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'abs(x) is too large for an accurate result to be returned and the function',
                    'returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _bei_py = utils._EngineFloat64ScalarType.to_py(_bei)
        return_data = _bei_py
    _err_class.manage()
    return return_data

def kelvin_ker(x):
    u"""
    Kelvin function ker(x).

    ``kelvin_ker`` returns a value for the Kelvin function ker(x).

    For full information please refer to the NAG Library document for
    s19ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    ker : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*. The function returns zero.

            Constraint: `x` <= *<value>*.

            `x` is too large, the result underflows and the function
            returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

            The function is undefined and returns zero.

    Notes
    -----
    ``kelvin_ker`` evaluates an approximation to the Kelvin function
    ker(x).

    **Note:** for x < 0 the function is undefined and at x = 0 it is
    infinite so we need only consider x > 0.

    The function is based on several Chebyshev expansions:

    For 0 < x <= 1,

        ker(x) = -f(t)log(x)+pi/16x^2g(t)+y(t)

    where f(t), g(t) and y(t) are expansions in the variable t = 2x^4-1.

    For 1 < x <= 3,

        ker(x) = exp(-11/16x)q(t)

    where q(t) is an expansion in the variable t = x-2.

    For x > 3,

        ker(x) =
        sqrt(pi/2x)e^(-x/sqrt(2))[(1+1/xc(t))cos(beta)-1/xd(t)sin(beta)]

    where beta = x/(sqrt(2))+pi/8, and c(t) and d(t) are expansions in
    the variable t = 6/x-1.

    When x is sufficiently close to zero, the result is computed as

        ker(x) = -gamma-log(x/2)+(pi-3/8x^2)(x^2)/16

    and when x is even closer to zero, simply as ker(x) =
    -gamma-log(x/2).

    For large x, ker(x) is asymptotically given by
    sqrt(pi/2x)e^(-x/sqrt(2)) and this becomes so small that it cannot
    be computed without underflow and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_ker'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, ker.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ker = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ker,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'The function returns zero.',
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large, the result underflows and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x > 0.0.',
                    'The function is undefined and returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ker_py = utils._EngineFloat64ScalarType.to_py(_ker)
        return_data = _ker_py
    _err_class.manage()
    return return_data

def kelvin_kei(x):
    u"""
    Kelvin function kei(x).

    ``kelvin_kei`` returns a value for the Kelvin function kei(x).

    For full information please refer to the NAG Library document for
    s19ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    kei : float
        The value of the function at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*. The function returns zero.

            Constraint: `x` <= *<value>*.

            `x` is too large, the result underflows and the function
            returns zero.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

            The function is undefined and returns zero.

    Notes
    -----
    ``kelvin_kei`` evaluates an approximation to the Kelvin function
    kei(x).

    **Note:** for x < 0 the function is undefined, so we need only
    consider x >= 0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 1,

        kei(x) = -pi/4f(t)+(x^2)/4[-g(t)log(x)+v(t)]

    where f(t), g(t) and v(t) are expansions in the variable t = 2x^4-1;

    For 1 < x <= 3,

        kei(x) = exp(-9/8x)u(t)

    where u(t) is an expansion in the variable t = x-2;

    For x > 3,

        kei(x) =
        sqrt(pi/2x)e^(-x/sqrt(2))[(1+1/x)c(t)sin(beta)+1/xd(t)cos(beta)]

    where beta = x/(sqrt(2))+pi/8, and c(t) and d(t) are expansions in
    the variable t = 6/x-1.

    For x < 0, the function is undefined, and hence the function fails
    and returns zero.

    When x is sufficiently close to zero, the result is computed as

        kei(x) = -pi/4+(1-gamma-log(x/2))(x^2)/4

    and when x is even closer to zero simply as

        kei(x) = -pi/4.

    For large x, kei(x) is asymptotically given by
    sqrt(pi/2x)e^(-x/sqrt(2)) and this becomes so small that it cannot
    be computed without underflow and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_kei'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, kei.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _kei = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _kei,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'The function returns zero.',
                    'Constraint: x <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x is too large, the result underflows and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 0.0.',
                    'The function is undefined and returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _kei_py = utils._EngineFloat64ScalarType.to_py(_kei)
        return_data = _kei_py
    _err_class.manage()
    return return_data

def kelvin_ber_vector(n, x, f, ivalid):
    u"""
    Kelvin function vectorized ber(x).

    ``kelvin_ber_vector`` returns an array of values for the Kelvin
    function ber(x).

    For full information please refer to the NAG Library document for
    s19an

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19anf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: ber(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            abs(x_i) is too large for an accurate result to be returned.
            `f`[i-1] contains zero. The threshold value is the same as
            for `errno` = 1 in ``kelvin_ber``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``kelvin_ber_vector`` evaluates an approximation to the Kelvin
    function ber(x_i) for an array of arguments x_i, for i = 1,2,...,n.

    **Note:**  ber(-x) = ber(x), so the approximation need only consider
    x >= 0.0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 5,

        ber(x) = sum'_{r = 0}a_r T_r(t), with t = 2(x/5)^4-1.

    For x > 5,

        [table omitted]

    where alpha = x/(sqrt(2))-pi/8, beta = x/(sqrt(2))+pi/8,

    and a(t), b(t), c(t), and d(t) are expansions in the variable t =
    10/x-1.

    When x is sufficiently close to zero, the result is set directly to
    ber(0) = 1.0.

    For large x, there is a danger of the result being totally
    inaccurate, as the error amplification factor grows in an
    essentially exponential manner;, therefore, the function must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_ber_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19anft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kelvin_bei_vector(n, x, f, ivalid):
    u"""
    Kelvin function vectorized bei(x).

    ``kelvin_bei_vector`` returns an array of values for the Kelvin
    function bei(x).

    For full information please refer to the NAG Library document for
    s19ap

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19apf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: bei(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            abs(x_i) is too large for an accurate result to be returned.
            `f`[i-1] contains zero. The threshold value is the same as
            for `errno` = 1 in ``kelvin_bei``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``kelvin_bei_vector`` evaluates an approximation to the Kelvin
    function bei(x_i) for an array of arguments x_i, for i = 1,2,...,n.

    **Note:**  bei(-x) = bei(x), so the approximation need only consider
    x >= 0.0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 5,

        bei(x) = (x^2)/4sum'_{r = 0}a_r T_r(t), with t = 2(x/5)^4-1;

    For x > 5,

        bei(x) = (e^(x/sqrt(2)))/sqrt(2 pi
        x)[(1+1/xa(t))sin(alpha)-1/xb(t)cos(alpha)]

        +(e^(x/sqrt(2)))/sqrt(2 pi
        x)[(1+1/xc(t))cos(beta)-1/xd(t)sin(beta)]

    where alpha = x/(sqrt(2))-pi/8, beta = x/(sqrt(2))+pi/8,

    and a(t), b(t), c(t), and d(t) are expansions in the variable t =
    10/x-1.

    When x is sufficiently close to zero, the result is computed as
    bei(x) = (x^2)/4.
    If this result would underflow, the result returned is bei(x) = 0.0.

    For large x, there is a danger of the result being totally
    inaccurate, as the error amplification factor grows in an
    essentially exponential manner;, therefore, the function must fail.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_bei_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19apft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kelvin_ker_vector(n, x, f, ivalid):
    u"""
    Kelvin function vectorized ker(x).

    ``kelvin_ker_vector`` returns an array of values for the Kelvin
    function ker(x).

    For full information please refer to the NAG Library document for
    s19aq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19aqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: ker(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large, the result underflows. `f`[i-1] contains
            zero. The threshold value is the same as for `errno` = 1 in
            ``kelvin_ker``.

        `ivalid`[i-1] = 2

            x_i <= 0.0, the function is undefined. `f`[i-1] contains
            0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``kelvin_ker_vector`` evaluates an approximation to the Kelvin
    function ker(x_i) for an array of arguments x_i, for i = 1,2,...,n.

    **Note:** for x < 0 the function is undefined and at x = 0 it is
    infinite so we need only consider x > 0.

    The function is based on several Chebyshev expansions:

    For 0 < x <= 1,

        ker(x) = -f(t)log(x)+pi/16x^2g(t)+y(t)

    where f(t), g(t) and y(t) are expansions in the variable t = 2x^4-1.

    For 1 < x <= 3,

        ker(x) = exp(-11/16x)q(t)

    where q(t) is an expansion in the variable t = x-2.

    For x > 3,

        ker(x) =
        sqrt(pi/2x)e^(-x/sqrt(2))[(1+1/xc(t))cos(beta)-1/xd(t)sin(beta)]

    where beta = x/(sqrt(2))+pi/8, and c(t) and d(t) are expansions in
    the variable t = 6/x-1.

    When x is sufficiently close to zero, the result is computed as

        ker(x) = -gamma-log(x/2)+(pi-3/8x^2)(x^2)/16

    and when x is even closer to zero, simply as ker(x) =
    -gamma-log(x/2).

    For large x, ker(x) is asymptotically given by
    sqrt(pi/2x)e^(-x/sqrt(2)) and this becomes so small that it cannot
    be computed without underflow and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_ker_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19aqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kelvin_kei_vector(n, x, f, ivalid):
    u"""
    Kelvin function vectorized kei(x).

    ``kelvin_kei_vector`` returns an array of values for the Kelvin
    function kei(x).

    For full information please refer to the NAG Library document for
    s19ar

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s19arf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: kei(x_i), the function values.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] contains the error code for x_i, for i
        = 1,2,...,`n`.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            x_i is too large, the result underflows. `f`[i-1] contains
            zero. The threshold value is the same as for `errno` = 1 in
            ``kelvin_kei``.

        `ivalid`[i-1] = 2

            x_i < 0.0, the function is undefined. `f`[i-1] contains 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    ``kelvin_kei_vector`` evaluates an approximation to the Kelvin
    function kei(x_i) for an array of arguments x_i, for i = 1,2,...,n.

    **Note:** for x < 0 the function is undefined, so we need only
    consider x >= 0.

    The function is based on several Chebyshev expansions:

    For 0 <= x <= 1,

        kei(x) = -pi/4f(t)+(x^2)/4[-g(t)log(x)+v(t)]

    where f(t), g(t) and v(t) are expansions in the variable t = 2x^4-1;

    For 1 < x <= 3,

        kei(x) = exp(-9/8x)u(t)

    where u(t) is an expansion in the variable t = x-2;

    For x > 3,

        kei(x) =
        sqrt(pi/2x)e^(-x/sqrt(2))[(1+1/x)c(t)sin(beta)+1/xd(t)cos(beta)]

    where beta = x/(sqrt(2))+pi/8, and c(t) and d(t) are expansions in
    the variable t = 6/x-1.

    For x < 0, the function is undefined, and hence the function fails
    and returns zero.

    When x is sufficiently close to zero, the result is computed as

        kei(x) = -pi/4+(1-gamma-log(x/2))(x^2)/4

    and when x is even closer to zero simply as

        kei(x) = -pi/4.

    For large x, kei(x) is asymptotically given by
    sqrt(pi/2x)e^(-x/sqrt(2)) and this becomes so small that it cannot
    be computed without underflow and the function fails.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.kelvin_kei_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s19arft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of x was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def fresnel_s(x):
    u"""
    Fresnel integral S(x).

    ``fresnel_s`` returns a value for the Fresnel integral S(x).

    For full information please refer to the NAG Library document for
    s20ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s20acf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    sx : float
        The value of the function at x.

    Notes
    -----
    ``fresnel_s`` evaluates an approximation to the Fresnel integral

        S(x) = int_0^xsin(pi/2t^2)dt.

    **Note:**  S(x) = -S(-x), so the approximation need only consider x
    >= 0.0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 3,

        S(x) = x^3sum'_{r = 0}a_r T_r(t), with t = 2(x/3)^4-1.

    For x > 3,

        S(x) = 1/2-f(x)/xcos(pi/2x^2)-g(x)/(x^3)sin(pi/2x^2),

    where f(x) = sum'_{r = 0}b_r T_r(t),

    and g(x) = sum'_{r = 0}c_r T_r(t),

    with t = 2(3/x)^4-1.

    For small x, S(x) ~= pi/6x^3.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For very small x, this approximation would underflow; the result is
    then set exactly to zero.

    For large x, f(x) ~= 1/pi and g(x) ~= 1/pi^2.
    Therefore, for moderately large x, when 1/pi^2x^3 is negligible
    compared with 1/2, the second term in the approximation for x > 3
    may be dropped.
    For very large x, when 1/pi x becomes negligible, S(x) ~= 1/2.
    However, there will be considerable difficulties in calculating
    cos(pi/2x^2) accurately before this final limiting value can be
    used.
    Since cos(pi/2x^2) is periodic, its value is essentially determined
    by the fractional part of x^2.
    If x^2 = N+theta where N is an integer and 0 <= theta < 1, then
    cos(pi/2x^2) depends on theta and on N modulo 4.
    By exploiting this fact, it is possible to retain significance in
    the calculation of cos(pi/2x^2) either all the way to the very large
    x limit, or at least until the integer part of x/2 is equal to the
    maximum integer allowed on the machine.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.fresnel_s'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s20acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, sx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _sx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _sx,
    )
    utils._handle_cb_exception(excs)
    _sx_py = utils._EngineFloat64ScalarType.to_py(_sx)
    return_data = _sx_py
    return return_data

def fresnel_c(x):
    u"""
    Fresnel integral C(x).

    ``fresnel_c`` returns a value for the Fresnel integral C(x).

    For full information please refer to the NAG Library document for
    s20ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s20adf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    cx : float
        The value of the function at x.

    Notes
    -----
    ``fresnel_c`` evaluates an approximation to the Fresnel integral

        C(x) = int_0^xcos(pi/2t^2)dt.

    **Note:**  C(x) = -C(-x), so the approximation need only consider x
    >= 0.0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 3,

        C(x) = xsum'_{r = 0}a_r T_r(t), with t = 2(x/3)^4-1.

    For x > 3,

        C(x) = 1/2+f(x)/xsin(pi/2x^2)-g(x)/(x^3)cos(pi/2x^2),

    where f(x) = sum'_{r = 0}b_r T_r(t),

    and g(x) = sum'_{r = 0}c_r T_r(t),

    with t = 2(3/x)^4-1.

    For small x, C(x) ~= x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, f(x) ~= 1/pi and g(x) ~= 1/pi^2.
    Therefore, for moderately large x, when 1/pi^2x^3 is negligible
    compared with 1/2, the second term in the approximation for x > 3
    may be dropped.
    For very large x, when 1/pi x becomes negligible, C(x) ~= 1/2.
    However, there will be considerable difficulties in calculating
    sin(pi/2x^2) accurately before this final limiting value can be
    used.
    Since sin(pi/2x^2) is periodic, its value is essentially determined
    by the fractional part of x^2.
    If x^2 = N+theta, where N is an integer and 0 <= theta < 1, then
    sin(pi/2x^2) depends on theta and on N modulo 4.
    By exploiting this fact, it is possible to retain some significance
    in the calculation of sin(pi/2x^2) either all the way to the very
    large x limit, or at least until the integer part of x/2 is equal to
    the maximum integer allowed on the machine.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.fresnel_c'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s20adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, cx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _cx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _cx,
    )
    utils._handle_cb_exception(excs)
    _cx_py = utils._EngineFloat64ScalarType.to_py(_cx)
    return_data = _cx_py
    return return_data

def fresnel_s_vector(n, x, f):
    u"""
    Fresnel integral vectorized S(x).

    ``fresnel_s_vector`` returns an array of values for the Fresnel
    integral S(x).

    For full information please refer to the NAG Library document for
    s20aq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s20aqf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: S(x_i), the function values.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``fresnel_s_vector`` evaluates an approximation to the Fresnel
    integral

        S(x_i) = int_0^x_isin(pi/2t^2)dt

    for an array of arguments x_i, for i = 1,2,...,n.

    **Note:**  S(x) = -S(-x), so the approximation need only consider x
    >= 0.0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 3,

        S(x) = x^3sum'_{r = 0}a_r T_r(t), with t = 2(x/3)^4-1.

    For x > 3,

        S(x) = 1/2-f(x)/xcos(pi/2x^2)-g(x)/(x^3)sin(pi/2x^2),

    where f(x) = sum'_{r = 0}b_r T_r(t),

    and g(x) = sum'_{r = 0}c_r T_r(t),

    with t = 2(3/x)^4-1.

    For small x, S(x) ~= pi/6x^3.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.
    For very small x, this approximation would underflow; the result is
    then set exactly to zero.

    For large x, f(x) ~= 1/pi and g(x) ~= 1/pi^2.
    Therefore, for moderately large x, when 1/pi^2x^3 is negligible
    compared with 1/2, the second term in the approximation for x > 3
    may be dropped.
    For very large x, when 1/pi x becomes negligible, S(x) ~= 1/2.
    However, there will be considerable difficulties in calculating
    cos(pi/2x^2) accurately before this final limiting value can be
    used.
    Since cos(pi/2x^2) is periodic, its value is essentially determined
    by the fractional part of x^2.
    If x^2 = N+theta where N is an integer and 0 <= theta < 1, then
    cos(pi/2x^2) depends on theta and on N modulo 4.
    By exploiting this fact, it is possible to retain significance in
    the calculation of cos(pi/2x^2) either all the way to the very large
    x limit, or at least until the integer part of x/2 is equal to the
    maximum integer allowed on the machine.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.fresnel_s_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s20aqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def fresnel_c_vector(n, x, f):
    u"""
    Fresnel integral vectorized C(x).

    ``fresnel_c_vector`` returns an array of values for the Fresnel
    integral C(x).

    For full information please refer to the NAG Library document for
    s20ar

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s20arf.html

    Parameters
    ----------
    n : int
        n, the number of points.

    x : float, shape (`n`)
        The argument x_i of the function, for i = 1,2,...,`n`.

    f : float, shape (`n`), modified in place
        `On exit`: C(x_i), the function values.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``fresnel_c_vector`` evaluates an approximation to the Fresnel
    integral

        C(x_i) = int_0^x_icos(pi/2t^2)dt

    for an array of arguments x_i, for i = 1,2,...,n.

    **Note:**  C(x) = -C(-x), so the approximation need only consider x
    >= 0.0.

    The function is based on three Chebyshev expansions:

    For 0 < x <= 3,

        C(x) = xsum'_{r = 0}a_r T_r(t), with t = 2(x/3)^4-1.

    For x > 3,

        C(x) = 1/2+f(x)/xsin(pi/2x^2)-g(x)/(x^3)cos(pi/2x^2),

    where f(x) = sum'_{r = 0}b_r T_r(t),

    and g(x) = sum'_{r = 0}c_r T_r(t),

    with t = 2(3/x)^4-1.

    For small x, C(x) ~= x.
    This approximation is used when x is sufficiently small for the
    result to be correct to machine precision.

    For large x, f(x) ~= 1/pi and g(x) ~= 1/pi^2.
    Therefore, for moderately large x, when 1/pi^2x^3 is negligible
    compared with 1/2, the second term in the approximation for x > 3
    may be dropped.
    For very large x, when 1/pi x becomes negligible, C(x) ~= 1/2.
    However, there will be considerable difficulties in calculating
    sin(pi/2x^2) accurately before this final limiting value can be
    used.
    Since sin(pi/2x^2) is periodic, its value is essentially determined
    by the fractional part of x^2.
    If x^2 = N+theta, where N is an integer and 0 <= theta < 1, then
    sin(pi/2x^2) depends on theta and on N modulo 4.
    By exploiting this fact, it is possible to retain some significance
    in the calculation of sin(pi/2x^2) either all the way to the very
    large x limit, or at least until the integer part of x/2 is equal to
    the maximum integer allowed on the machine.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.fresnel_c_vector'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s20arft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_cls.refresh_py(_f)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ellipint_symm_1_degen(x, y):
    u"""
    Degenerate symmetrised elliptic integral of 1st kind R_C(x,y).

    ``ellipint_symm_1_degen`` returns a value of an elementary integral,
    which occurs as a degenerate case of an elliptic integral of the
    first kind.

    For full information please refer to the NAG Library document for
    s21ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21baf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    y : float
        The argument y of the function.

    Returns
    -------
    rc : float
        The value of an elementary integral, which occurs as a
        degenerate case of an elliptic integral of the first kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

            The function is undefined.

        (`errno` 2)
            On entry, `y` = 0.0.

            Constraint: `y` != 0.0.

            The function is undefined and returns zero.

    Notes
    -----
    ``ellipint_symm_1_degen`` calculates an approximate value for the
    integral

        R_C(x,y) = 1/2int_0^infinity dt/((t+y).sqrt(t+x))

    where x >= 0 and y != 0.

    This function, which is related to the logarithm or inverse
    hyperbolic functions for y < x and to inverse circular functions if
    x < y, arises as a degenerate form of the elliptic integral of the
    first kind.
    If y < 0, the result computed is the Cauchy principal value of the
    integral.

    The basic algorithm, which is due to Carlson (1979) and Carlson
    (1988), is to reduce the arguments recursively towards their mean by
    the system:

        [table omitted]

    The quantity |S_n| for n = 0,1,2,3,... decreases with increasing n,
    eventually |S_n|~1/4^n.
    For small enough S_n the required function value can be approximated
    by the first few terms of the Taylor series about the mean.
    That is

        R_C(x,y) =
        (1+3S_n^2/10+(S_n^3)/7+3S_n^4/8+9S_n^5/22)/sqrt(mu_n).

    The truncation error involved in using this approximation is bounded
    by 16|S_n|^6/(1-2|S_n|) and the recursive process is stopped when
    S_n is small enough for this truncation error to be negligible
    compared to the machine precision.

    Within the domain of definition, the function value is itself
    representable for all representable values of its arguments.
    However, for values of the arguments near the extremes the above
    algorithm must be modified so as to avoid causing underflows or
    overflows in intermediate steps.
    In extreme regions arguments are prescaled away from the extremes
    and compensating scaling of the result is done before returning to
    the calling program.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_symm_1_degen'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Output, rc.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _rc = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _rc,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: x >= 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y = 0.0.',
                    'Constraint: y != 0.0.',
                    'The function is undefined and returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rc_py = utils._EngineFloat64ScalarType.to_py(_rc)
        return_data = _rc_py
    _err_class.manage()
    return return_data

def ellipint_symm_1(x, y, z):
    u"""
    Symmetrised elliptic integral of 1st kind R_F(x,y,z).

    ``ellipint_symm_1`` returns a value of the symmetrised elliptic
    integral of the first kind.

    For full information please refer to the NAG Library document for
    s21bb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bbf.html

    Parameters
    ----------
    x : float
        The arguments x, y and z of the function.

    y : float
        The arguments x, y and z of the function.

    z : float
        The arguments x, y and z of the function.

    Returns
    -------
    rf : float
        The value of the symmetrised elliptic integral of the first
        kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*, `y` = *<value>* and `z` =
            *<value>*.

            Constraint: `x` >= 0.0 and `y` >= 0.0 and `z` >= 0.0.

            The function is undefined.

        (`errno` 2)
            On entry, `x` = *<value>*, `y` = *<value>* and `z` =
            *<value>*.

            Constraint: at most one of `x`, `y` and `z` is 0.0.

            The function is undefined and returns zero.

    Notes
    -----
    ``ellipint_symm_1`` calculates an approximation to the integral

        R_F(x,y,z) = 1/2int_0^infinity dt/sqrt((t+x)(t+y)(t+z))

    where x, y, z >= 0 and at most one is zero.

    The basic algorithm, which is due to Carlson (1979) and Carlson
    (1988), is to reduce the arguments recursively towards their mean by
    the rule:

        x_0 = min(x, y, z),  z_0 = max(x,y, z),

        y_0 = remaining third intermediate value argument.

    (This ordering, which is possible because of the symmetry of the
    function, is done for technical reasons related to the avoidance of
    overflow and underflow.)

        [table omitted]

    epsilon_n = max(|X_n|,|Y_n|, |Z_n|) and the function may be
    approximated adequately by a fifth order power series:

        R_F(x,y,z) = 1/sqrt(mu_n)(1-(E_2)/10+(E_2^2)/24-3E_2
        E_3/44+(E_3)/14)

    where E_2 = X_n Y_n+Y_n Z_n+Z_n X_n, E_3 = X_n Y_n Z_n.

    The truncation error involved in using this approximation is bounded
    by epsilon_n^6/4*(1-epsilon_n) and the recursive process is stopped
    when this truncation error is negligible compared with the machine
    precision.

    Within the domain of definition, the function value is itself
    representable for all representable values of its arguments.
    However, for values of the arguments near the extremes the above
    algorithm must be modified so as to avoid causing underflows or
    overflows in intermediate steps.
    In extreme regions arguments are prescaled away from the extremes
    and compensating scaling of the result is done before returning to
    the calling program.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_symm_1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Input, z.
        utils._BYREF_FLOAT64, # Output, rf.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineFloat64ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _rf = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _z,
        _rf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x >= 0.0 and y >= 0.0 and z >= 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: at most one of x, y and z is 0.0.',
                    'The function is undefined and returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rf_py = utils._EngineFloat64ScalarType.to_py(_rf)
        return_data = _rf_py
    _err_class.manage()
    return return_data

def ellipint_symm_2(x, y, z):
    u"""
    Symmetrised elliptic integral of 2nd kind R_D(x,y,z).

    ``ellipint_symm_2`` returns a value of the symmetrised elliptic
    integral of the second kind.

    For full information please refer to the NAG Library document for
    s21bc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bcf.html

    Parameters
    ----------
    x : float
        The arguments x, y and z of the function.

    y : float
        The arguments x, y and z of the function.

    z : float
        The arguments x, y and z of the function.

    Returns
    -------
    rd : float
        The value of the symmetrised elliptic integral of the second
        kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` and `y` are both 0.0.

            Constraint: at most one of `x` and `y` is 0.0.

            The function is undefined.

        (`errno` 1)
            On entry, `x` = *<value>* and `y` = *<value>*.

            Constraint: `x` >= 0.0 and `y` >= 0.0.

            The function is undefined.

        (`errno` 2)
            On entry, `z` = *<value>*.

            Constraint: `z` > 0.0.

            The function is undefined.

        (`errno` 3)
            On entry, L = *<value>*, `x` = *<value>*, `y` = *<value>*
            and `z` = *<value>*.

            Constraint: `z` >= L and (`x` >= L or `y` >= L).

            The function is undefined.

        (`errno` 4)
            On entry, U = *<value>*, `x` = *<value>*, `y` = *<value>*
            and `z` = *<value>*.

            Constraint: `x` < U and `y` < U and `z` < U.

            There is a danger of setting underflow and the function
            returns zero.

    Notes
    -----
    ``ellipint_symm_2`` calculates an approximate value for the integral

        R_D(x,y,z) = 3/2int_0^infinity dt/sqrt((t+x)(t+y)(t+z)^3)

    where x, y >= 0, at most one of x and y is zero, and z > 0.

    The basic algorithm, which is due to Carlson (1979) and Carlson
    (1988), is to reduce the arguments recursively towards their mean by
    the rule:

        [table omitted]

    For n sufficiently large,

        epsilon_n = max(|X_n|,|Y_n|, |Z_n|)~(1/4)^n

    and the function may be approximated adequately by a fifth order
    power series

        [table omitted]

    where S_n^(m) = (X_n^m+Y_n^m+3Z_n^m)/2m. The truncation error in
    this expansion is bounded by 3 epsilon_n^6/sqrt((1-epsilon_n)^3) and
    the recursive process is terminated when this quantity is negligible
    compared with the machine precision.

    The function may fail either because it has been called with
    arguments outside the domain of definition, or with arguments so
    extreme that there is an unavoidable danger of setting underflow or
    overflow.

    **Note:**  R_D(x,x,x) = x^(-3/2), so there exists a region of
    extreme arguments for which the function value is not representable.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_symm_2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Input, z.
        utils._BYREF_FLOAT64, # Output, rd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineFloat64ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _rd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _z,
        _rd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x and y are both 0.0.',
                    'Constraint: at most one of x and y is 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s} and y = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x >= 0.0 and y >= 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: z > 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, L = {:s}, x = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: z >= L and (x >= L or y >= L).',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, U = {:s}, x = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'y = {:s} and z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x < U and y < U and z < U.',
                    'There is a danger of setting underflow and the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rd_py = utils._EngineFloat64ScalarType.to_py(_rd)
        return_data = _rd_py
    _err_class.manage()
    return return_data

def ellipint_symm_3(x, y, z, r):
    u"""
    Symmetrised elliptic integral of 3rd kind R_J(x,y,z,r).

    ``ellipint_symm_3`` returns a value of the symmetrised elliptic
    integral of the third kind.

    For full information please refer to the NAG Library document for
    s21bd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bdf.html

    Parameters
    ----------
    x : float
        The arguments x, y, z and rho of the function.

    y : float
        The arguments x, y, z and rho of the function.

    z : float
        The arguments x, y, z and rho of the function.

    r : float
        The arguments x, y, z and rho of the function.

    Returns
    -------
    rj : float
        The value of the symmetrised elliptic integral of the third
        kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*, `y` = *<value>* and `z` =
            *<value>*.

            Constraint: at most one of `x`, `y` and `z` is 0.0.

            The function is undefined.

        (`errno` 1)
            On entry, `x` = *<value>*, `y` = *<value>* and `z` =
            *<value>*.

            Constraint: `x` >= 0.0 and `y` >= 0.0 and `z` >= 0.0.

        (`errno` 2)
            On entry, `r` = 0.0.

            Constraint: `r` != 0.0.

        (`errno` 3)
            On entry, L = *<value>*, `r` = *<value>*, `x` = *<value>*,
            `y` = *<value>* and `z` = *<value>*.

            Constraint: abs(`r`) >= L and at most one of `x`, `y` and
            `z` is less than L.

        (`errno` 4)
            On entry, U = *<value>*, `r` = *<value>*, `x` = *<value>*,
            `y` = *<value>* and `z` = *<value>*.

            Constraint: abs(`r`) <= U and `x` <= U and `y` <= U and `z`
            <= U.

    Notes
    -----
    ``ellipint_symm_3`` calculates an approximation to the integral

        R_J(x,y,z,rho) = 3/2int_0^infinity
        dt/(t+rho)sqrt((t+x)(t+y)(t+z))

    where x, y, z >= 0, rho != 0 and at most one of x, y and z is zero.

    If rho < 0, the result computed is the Cauchy principal value of the
    integral.

    The basic algorithm, which is due to Carlson (1979) and Carlson
    (1988), is to reduce the arguments recursively towards their mean by
    the rule:

        [table omitted]

    For n sufficiently large,

        epsilon_n = max(|X_n|,|Y_n|,|Z_n|, |P_n|)~1/(4^n)

    and the function may be approximated by a fifth order power series

        [table omitted]

    where S_n^(m) = (X_n^m+Y_n^m+Z_n^m+2P_n^m)/2m.

    The truncation error in this expansion is bounded by 3
    epsilon_n^6/sqrt((1-epsilon_n)^3) and the recursion process is
    terminated when this quantity is negligible compared with the
    machine precision.
    The function may fail either because it has been called with
    arguments outside the domain of definition or with arguments so
    extreme that there is an unavoidable danger of setting underflow or
    overflow.

    **Note:**  R_J(x,x,x,x) = x^(-3/2), so there exists a region of
    extreme arguments for which the function value is not representable.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_symm_3'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Input, z.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Output, rj.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineFloat64ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _rj = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _z,
        _r,
        _rj,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: at most one of x, y and z is 0.0.',
                    'The function is undefined.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x >= 0.0 and y >= 0.0 and z >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = 0.0.',
                    'Constraint: r != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, L = {:s}, r = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                    ),
                    'Constraint: abs(r) >= L and at most one of x, y and z is less than L.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, U = {:s}, r = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x = {:s}, y = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'z = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                    ),
                    'Constraint: abs(r) <= U and x <= U and y <= U and z <= U.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rj_py = utils._EngineFloat64ScalarType.to_py(_rj)
        return_data = _rj_py
    _err_class.manage()
    return return_data

def ellipint_legendre_1(phi, dm):
    u"""
    Elliptic integral of 1st kind, Legendre form, F(phi | m).

    ``ellipint_legendre_1`` returns a value of the classical (Legendre)
    form of the incomplete elliptic integral of the first kind.

    For full information please refer to the NAG Library document for
    s21be

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bef.html

    Parameters
    ----------
    phi : float
        The arguments phi and m of the function.

    dm : float
        The arguments phi and m of the function.

    Returns
    -------
    f : float
        The value of the classical (Legendre) form of the incomplete
        elliptic integral of the first kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `phi` = *<value>*.

            Constraint: 0 <= `phi` <= pi/2.

            On failure, the function returns zero.

        (`errno` 2)
            On entry, `phi` = *<value>* and `dm` = *<value>*; the
            integral is undefined.

            Constraint: `dm`*sin^2(`phi`) <= 1.0.

            On failure, the function returns zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, sin(`phi`) = 1 and `dm` = 1.0; the integral is
            infinite.

            On failure, the function returns the largest machine number
            (see ``machine.real_largest``).

    Notes
    -----
    ``ellipint_legendre_1`` calculates an approximation to the integral

        F(phi | m) = int_0^phi(1-msin^2(theta))^(-1/2)d theta,

    where 0 <= phi <= pi/2, msin^2(phi) <= 1 and m and sin(phi) may not
    both equal one.

    The integral is computed using the symmetrised elliptic integrals of
    Carlson (Carlson (1979) and Carlson (1988)).
    The relevant identity is

        F(phi | m) = R_F(q,r,1)sin(phi),

    where q = cos^2(phi), r = 1-msin^2(phi) and R_F is the Carlson
    symmetrised incomplete elliptic integral of the first kind (see
    ``ellipint_symm_1``).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_legendre_1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, phi.
        utils._BYREF_FLOAT64, # Input, dm.
        utils._BYREF_FLOAT64, # Output, f.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _phi = utils._EngineFloat64ScalarType(
        phi, {'fun_name': fun_name, 'entity': 'phi'},
    ).to_ctypes()
    _dm = utils._EngineFloat64ScalarType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
    ).to_ctypes()
    _f = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _phi,
        _dm,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= phi <= pi/2.',
                    'On failure, the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s} and dm = {:s};'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'the integral is undefined.',
                    'Constraint: dm*sin^2(phi) <= 1.0.',
                    'On failure, the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sin(phi) = 1 and dm = 1.0;',
                    'the integral is infinite.',
                    'On failure, the function returns the largest machine number',
                    '(see machine.real_largest).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f_py = utils._EngineFloat64ScalarType.to_py(_f)
        return_data = _f_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['f'],
            )(
                _f_py,
            )
    _err_class.manage()
    return return_data

def ellipint_legendre_2(phi, dm):
    u"""
    Elliptic integral of 2nd kind, Legendre form, E(phi | m).

    ``ellipint_legendre_2`` returns a value of the classical (Legendre)
    form of the incomplete elliptic integral of the second kind.

    For full information please refer to the NAG Library document for
    s21bf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bff.html

    Parameters
    ----------
    phi : float
        The arguments phi and m of the function.

    dm : float
        The arguments phi and m of the function.

    Returns
    -------
    e : float
        The value of the classical (Legendre) form of the incomplete
        elliptic integral of the second kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `phi` = *<value>*.

            Constraint: 0 <= `phi` <= pi/2.

        (`errno` 2)
            On entry, `phi` = *<value>* and `dm` = *<value>*; the
            integral is undefined.

            Constraint: `dm`*sin^2(`phi`) <= 1.0.

    Notes
    -----
    ``ellipint_legendre_2`` calculates an approximation to the integral

        E(phi | m) = int_0^phi(1-msin^2(theta))^(1/2)d theta,

    where 0 <= phi <= pi/2 and msin^2(phi) <= 1.

    The integral is computed using the symmetrised elliptic integrals of
    Carlson (Carlson (1979) and Carlson (1988)).
    The relevant identity is

        E(phi | m) = sin(phi)R_F(q,r,1)-1/3msin(phi)R_D(q,r,1),

    where q = cos^2(phi), r = 1-msin^2(phi), R_F is the Carlson
    symmetrised incomplete elliptic integral of the first kind (see
    ``ellipint_symm_1``) and R_D is the Carlson symmetrised incomplete
    elliptic integral of the second kind (see ``ellipint_symm_2``).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_legendre_2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, phi.
        utils._BYREF_FLOAT64, # Input, dm.
        utils._BYREF_FLOAT64, # Output, e.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _phi = utils._EngineFloat64ScalarType(
        phi, {'fun_name': fun_name, 'entity': 'phi'},
    ).to_ctypes()
    _dm = utils._EngineFloat64ScalarType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
    ).to_ctypes()
    _e = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _phi,
        _dm,
        _e,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= phi <= pi/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s} and dm = {:s};'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'the integral is undefined.',
                    'Constraint: dm*sin^2(phi) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _e_py = utils._EngineFloat64ScalarType.to_py(_e)
        return_data = _e_py
    _err_class.manage()
    return return_data

def ellipint_legendre_3(dn, phi, dm):
    u"""
    Elliptic integral of 3rd kind, Legendre form, Pi(n;phi | m).

    ``ellipint_legendre_3`` returns a value of the classical (Legendre)
    form of the incomplete elliptic integral of the third kind.

    For full information please refer to the NAG Library document for
    s21bg

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bgf.html

    Parameters
    ----------
    dn : float
        The arguments n, phi and m of the function.

    phi : float
        The arguments n, phi and m of the function.

    dm : float
        The arguments n, phi and m of the function.

    Returns
    -------
    p : float
        The value of the classical (Legendre) form of the incomplete
        elliptic integral of the third kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `phi` = *<value>*.

            Constraint: 0 <= `phi` <= (pi/2).

        (`errno` 2)
            On entry, `phi` = *<value>* and `dm` = *<value>*; the
            integral is undefined.

            Constraint: `dm`*sin^2(`phi`) <= 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, sin(`phi`) = 1 and `dm` = 1.0; the integral is
            infinite.

        (`errno` 4)
            On entry, `phi` = *<value>* and `dn` = *<value>*; the
            integral is infinite.

            Constraint: `dn`*sin^2(`phi`) != 1.0.

    Notes
    -----
    ``ellipint_legendre_3`` calculates an approximation to the integral

        Pi(n;phi | m) =
        int_0^phi(1-nsin^2(theta))^(-1)(1-msin^2(theta))^(-1/2)d theta,

    where 0 <= phi <= pi/2, msin^2(phi) <= 1, m and sin(phi) may not
    both equal one, and nsin^2(phi) != 1.

    The integral is computed using the symmetrised elliptic integrals of
    Carlson (Carlson (1979) and Carlson (1988)).
    The relevant identity is

        Pi(n;phi | m) = sin(phi)R_F(q,r,1)+1/3nsin(phi)R_J(q,r,1,s),

    where q = cos^2(phi), r = 1-msin^2(phi), s = 1-nsin^2(phi), R_F is
    the Carlson symmetrised incomplete elliptic integral of the first
    kind (see ``ellipint_symm_1``) and R_J is the Carlson symmetrised
    incomplete elliptic integral of the third kind (see
    ``ellipint_symm_3``).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_legendre_3'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, dn.
        utils._BYREF_FLOAT64, # Input, phi.
        utils._BYREF_FLOAT64, # Input, dm.
        utils._BYREF_FLOAT64, # Output, p.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dn = utils._EngineFloat64ScalarType(
        dn, {'fun_name': fun_name, 'entity': 'dn'},
    ).to_ctypes()
    _phi = utils._EngineFloat64ScalarType(
        phi, {'fun_name': fun_name, 'entity': 'phi'},
    ).to_ctypes()
    _dm = utils._EngineFloat64ScalarType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dn,
        _phi,
        _dm,
        _p,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= phi <= (pi/2).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s} and dm = {:s};'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'the integral is undefined.',
                    'Constraint: dm*sin^2(phi) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sin(phi) = 1 and dm = 1.0; the integral is infinite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1004:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, phi = {:s} and dn = {:s};'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'the integral is infinite.',
                    'Constraint: dn*sin^2(phi) != 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        return_data = _p_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['p'],
            )(
                _p_py,
            )
    _err_class.manage()
    return return_data

def ellipint_complete_1(dm):
    u"""
    Complete elliptic integral of 1st kind, Legendre form, K(m).

    ``ellipint_complete_1`` returns a value of the classical (Legendre)
    form of the complete elliptic integral of the first kind.

    For full information please refer to the NAG Library document for
    s21bh

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bhf.html

    Parameters
    ----------
    dm : float
        The argument m of the function.

    Returns
    -------
    k : float
        The value of the classical (Legendre) form of the complete
        elliptic integral of the first kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dm` = *<value>*; the integral is undefined.

            Constraint: `dm` < 1.0.

            On failure, the function returns zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `dm` = 1.0; the integral is infinite.

            On failure, the function returns the largest machine number
            (see ``machine.real_largest``).

    Notes
    -----
    ``ellipint_complete_1`` calculates an approximation to the integral

        K(m) = int_0^pi/2(1-msin^2(theta))^(-1/2)d theta,

    where m < 1.

    The integral is computed using the symmetrised elliptic integrals of
    Carlson (Carlson (1979) and Carlson (1988)).
    The relevant identity is

        K(m) = R_F(0,1-m,1),

    where R_F is the Carlson symmetrised incomplete elliptic integral of
    the first kind (see ``ellipint_symm_1``).

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_complete_1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, dm.
        utils._BYREF_FLOAT64, # Output, k.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dm = utils._EngineFloat64ScalarType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
    ).to_ctypes()
    _k = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dm,
        _k,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dm = {:s}; the integral is undefined.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dm < 1.0.',
                    'On failure, the function returns zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dm = 1.0; the integral is infinite.',
                    'On failure, the function returns the largest machine number',
                    '(see machine.real_largest).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _k_py = utils._EngineFloat64ScalarType.to_py(_k)
        return_data = _k_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['k'],
            )(
                _k_py,
            )
    _err_class.manage()
    return return_data

def ellipint_complete_2(dm):
    u"""
    Complete elliptic integral of 2nd kind, Legendre form, E(m).

    ``ellipint_complete_2`` returns a value of the classical (Legendre)
    form of the complete elliptic integral of the second kind.

    For full information please refer to the NAG Library document for
    s21bj

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21bjf.html

    Parameters
    ----------
    dm : float
        The argument m of the function.

    Returns
    -------
    e : float
        The value of the classical (Legendre) form of the complete
        elliptic integral of the second kind.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dm` = *<value>*; the integral is undefined.

            Constraint: `dm` <= 1.0.

    Notes
    -----
    ``ellipint_complete_2`` calculates an approximation to the integral

        E(m) = int_0^pi/2(1-msin^2(theta))^(1/2)d theta,

    where m <= 1.

    The integral is computed using the symmetrised elliptic integrals of
    Carlson (Carlson (1979) and Carlson (1988)).
    The relevant identity is

        E(m) = R_F(0,1-m,1)-1/3mR_D(0,1-m,1),

    where R_F is the Carlson symmetrised incomplete elliptic integral of
    the first kind (see ``ellipint_symm_1``) and R_D is the Carlson
    symmetrised incomplete elliptic integral of the second kind (see
    ``ellipint_symm_2``).

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Carlson, B C, 1979, `Computing elliptic integrals by duplication`,
    Numerische Mathematik (33), 1--16

    Carlson, B C, 1988, `A table of elliptic integrals of the third
    kind`, Math. Comput. (51), 267--280
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_complete_2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21bjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, dm.
        utils._BYREF_FLOAT64, # Output, e.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dm = utils._EngineFloat64ScalarType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
    ).to_ctypes()
    _e = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dm,
        _e,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dm = {:s}; the integral is undefined.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dm <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _e_py = utils._EngineFloat64ScalarType.to_py(_e)
        return_data = _e_py
    _err_class.manage()
    return return_data

def jacellip_real(u, m):
    u"""
    Jacobian elliptic functions sn, cn and dn of real argument.

    ``jacellip_real`` evaluates the Jacobian elliptic functions sn, cn
    and dn.

    For full information please refer to the NAG Library document for
    s21ca

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21caf.html

    Parameters
    ----------
    u : float
        The argument u.

    m : float
        The argument m.

    Returns
    -------
    sn : float
        The value of sn(u).

    cn : float
        The value of cn(u).

    dn : float
        The value of dn(u).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, abs(`u`) is too large: abs(`u`) = *<value>* it
            must be less than *<value>*.

        (`errno` 2)
            On entry, abs(`m`) is too large when used in conjunction
            with the supplied argument `u`: abs(`m`) = *<value>* it must
            be less than *<value>*.

    Notes
    -----
    ``jacellip_real`` evaluates the Jacobian elliptic functions of
    argument u and argument m,

        [table omitted]

    where phi, called the `amplitude` of u, is defined by the integral

        u = int_0^phi d theta/sqrt(1-msin^2(theta)).

    The elliptic functions are sometimes written simply as sn(u), cn(u)
    and dn(u), avoiding explicit reference to the argument m.

    Another nine elliptic functions may be computed via the formulae

        [table omitted]

    (see Abramowitz and Stegun (1972)).

    ``jacellip_real`` is based on a procedure given by Bulirsch (1960),
    and uses the process of the arithmetic-geometric mean (16.9 in
    Abramowitz and Stegun (1972)).
    Constraints are placed on the values of u and m in order to avoid
    the possibility of machine overflow.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Bulirsch, R, 1960, `Numerical calculation of elliptic integrals and
    elliptic functions`, Numer. Math. (7), 76--90
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.jacellip_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, u.
        utils._BYREF_FLOAT64, # Input, m.
        utils._BYREF_FLOAT64, # Output, sn.
        utils._BYREF_FLOAT64, # Output, cn.
        utils._BYREF_FLOAT64, # Output, dn.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _u = utils._EngineFloat64ScalarType(
        u, {'fun_name': fun_name, 'entity': 'u'},
    ).to_ctypes()
    _m = utils._EngineFloat64ScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _sn = utils._EngineFloat64ScalarType.empty_ctypes()
    _cn = utils._EngineFloat64ScalarType.empty_ctypes()
    _dn = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _u,
        _m,
        _sn,
        _cn,
        _dn,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(u) is too large: abs(u) = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'it must be less than {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(m) is too large when used in conjunction with the supplied',
                    'argument u: abs(m) = {:s} it must be less than'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sn_py = utils._EngineFloat64ScalarType.to_py(_sn)
        _cn_py = utils._EngineFloat64ScalarType.to_py(_cn)
        _dn_py = utils._EngineFloat64ScalarType.to_py(_dn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['sn', 'cn', 'dn'],
        )(
            _sn_py,
            _cn_py,
            _dn_py,
        )
    _err_class.manage()
    return return_data

def jacellip_complex(z, ak2):
    u"""
    Jacobian elliptic functions sn, cn and dn of complex argument.

    ``jacellip_complex`` evaluates the Jacobian elliptic functions
    sn(z), cn(z) and dn(z) for a complex argument z.

    For full information please refer to the NAG Library document for
    s21cb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21cbf.html

    Parameters
    ----------
    z : complex
        The argument z of the functions.

    ak2 : float
        The value of k^2.

    Returns
    -------
    sn : complex
        The value of the function sn(z).

    cn : complex
        The value of the function cn(z).

    dn : complex
        The value of the function dn(z).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, abs(Im(`z`)) is too large: abs(Im(`z`)) =
            *<value>*. It must be less than *<value>*.

        (`errno` 1)
            On entry, abs(Re(`z`)) is too large: abs(Re(`z`)) =
            *<value>*. It must be less than *<value>*.

        (`errno` 1)
            On entry, `ak2` = *<value>*.

            Constraint: `ak2` <= 1.0.

        (`errno` 1)
            On entry, `ak2` = *<value>*.

            Constraint: `ak2` >= 0.0.

    Notes
    -----
    ``jacellip_complex`` evaluates the Jacobian elliptic functions sn(z
    | k), cn(z | k) and dn(z | k) given by

        [table omitted]

    where z is a complex argument, k is a real argument (the `modulus`)
    with k^2 <= 1 and phi (the `amplitude` of z) is defined by the
    integral

        z = int_0^phi d theta/sqrt(1-k^2sin^2(theta)).

    The above definitions can be extended for values of k^2 > 1 (see
    Salzer (1962)) by means of the formulae

        [table omitted]

    where k_1 = 1/k.

    Special values include

        [table omitted]

    These functions are often simply written as sn(z), cn(z) and dn(z),
    thereby avoiding explicit reference to the argument k.
    They can also be expressed in terms of Jacobian theta functions (see
    ``jactheta_real``).

    Another nine elliptic functions may be computed via the formulae

        [table omitted]

    (see Abramowitz and Stegun (1972)).

    The values of sn(z), cn(z) and dn(z) are obtained by calls to
    ``jacellip_real``.
    Further details can be found in [Further Comments].

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Salzer, H E, 1962, `Quick calculation of Jacobian elliptic
    functions`, Comm. ACM (5), 399
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.jacellip_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_FLOAT64, # Input, ak2.
        utils._BYREF_COMPLEX128, # Output, sn.
        utils._BYREF_COMPLEX128, # Output, cn.
        utils._BYREF_COMPLEX128, # Output, dn.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _ak2 = utils._EngineFloat64ScalarType(
        ak2, {'fun_name': fun_name, 'entity': 'ak2'},
    ).to_ctypes()
    _sn = utils._EngineComplex128ScalarType.empty_ctypes()
    _cn = utils._EngineComplex128ScalarType.empty_ctypes()
    _dn = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _ak2,
        _sn,
        _cn,
        _dn,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(Im(z)) is too large:',
                    'abs(Im(z)) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'It must be less than {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(Re(z)) is too large:',
                    'abs(Re(z)) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'It must be less than {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ak2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ak2 <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ak2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ak2 >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sn_py = utils._EngineComplex128ScalarType.to_py(_sn)
        _cn_py = utils._EngineComplex128ScalarType.to_py(_cn)
        _dn_py = utils._EngineComplex128ScalarType.to_py(_dn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['sn', 'cn', 'dn'],
        )(
            _sn_py,
            _cn_py,
            _dn_py,
        )
    _err_class.manage()
    return return_data

def jactheta_real(k, x, q):
    u"""
    Jacobian theta functions theta_k(x,q) of real argument.

    ``jactheta_real`` returns the value of one of the Jacobian theta
    functions theta_0(x,q), theta_1(x,q), theta_2(x,q), theta_3(x,q) or
    theta_4(x,q) for a real argument x and non-negative q < 1.

    For full information please refer to the NAG Library document for
    s21cc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21ccf.html

    Parameters
    ----------
    k : int
        Denotes the function theta_k(x,q) to be evaluated. Note that `k`
        = 4 is equivalent to `k` = 0.

    x : float
        The argument x of the function.

    q : float
        The argument q of the function.

    Returns
    -------
    jt : float
        The value of one of the Jacobian theta functions theta_0(x,q),
        theta_1(x,q), theta_2(x,q), theta_3(x,q) or theta_4(x,q) for a
        real argument x and non-negative q < 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `q` = *<value>*.

            Constraint: `q` < 1.0.

        (`errno` 1)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` <= 4.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

    Notes
    -----
    ``jactheta_real`` evaluates an approximation to the Jacobian theta
    functions theta_0(x,q), theta_1(x,q), theta_2(x,q), theta_3(x,q) and
    theta_4(x,q) given by

        [table omitted]

    where x and q (the `nome`) are real with 0 <= q < 1.

    These functions are important in practice because every one of the
    Jacobian elliptic functions (see ``jacellip_complex``) can be
    expressed as the ratio of two Jacobian theta functions (see
    Whittaker and Watson (1990)).
    There is also a bewildering variety of notations used in the
    literature to define them.
    Some authors (e.g., Section 16.27 of Abramowitz and Stegun (1972))
    define the argument in the trigonometric terms to be x instead of pi
    x.
    This can often lead to confusion, so great care must, therefore, be
    exercised when consulting the literature.
    Further details (including various relations and identities) can be
    found in the references.

    ``jactheta_real`` is based on a truncated series approach.
    If t differs from x or -x by an integer when 0 <= t <= 1/2, it
    follows from the periodicity and symmetry properties of the
    functions that theta_1(x,q) = +/-theta_1(t,q) and theta_3(x,q) =
    +/-theta_3(t,q).
    In a region for which the approximation is sufficiently accurate,
    theta_1 is set equal to the first term (n = 0) of the transformed
    series

        theta_1(t,q) = 2sqrt(lambda/pi)e^(-lambda t^2)sum_n = 0^infinity
        (-1)^ne^(-lambda(n+1/2)^2)sinh{(2n+1)lambda t}

    and theta_3 is set equal to the first two terms (i.e., n <= 1) of

        theta_3(t,q) = sqrt(lambda/pi)e^(-lambda t^2){1+2sum_n =
        1^infinity e^(-lambda n^2)cosh(2n lambda t)},

    where lambda = pi^2/|loge(q)|.
    Otherwise, the trigonometric series for theta_1(t,q) and
    theta_3(t,q) are used.
    For all values of x, theta_0 and theta_2 are computed from the
    relations theta_0(x,q) = theta_3(1/2-|x|,q) and theta_2(x,q) =
    theta_1(1/2-|x|,q).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Byrd, P F and Friedman, M D, 1971, `Handbook of Elliptic Integrals
    for Engineers and Scientists`, pp. 315--320, (2nd Edition),
    Springer--Verlag

    Magnus, W, Oberhettinger, F and Soni, R P, 1966, `Formulas and
    Theorems for the Special Functions of Mathematical Physics`,
    371--377, Springer--Verlag

    Tølke, F, 1966, `Praktische Funktionenlehre (Bd. II)`, 1--38,
    Springer--Verlag

    Whittaker, E T and Watson, G N, 1990, `A Course in Modern Analysis`,
    (4th Edition), Cambridge University Press
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.jactheta_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, q.
        utils._BYREF_FLOAT64, # Output, jt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _jt = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _x,
        _q,
        _jt,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k <= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _jt_py = utils._EngineFloat64ScalarType.to_py(_jt)
        return_data = _jt_py
    _err_class.manage()
    return return_data

def ellipint_general_2(z, akp, a, b):
    u"""
    General elliptic integral of 2nd kind F(z,k',a,b) of complex
    argument.

    ``ellipint_general_2`` returns the value of the general elliptic
    integral of the second kind F(z,k',a,b) for a complex argument z.

    For full information please refer to the NAG Library document for
    s21da

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s21daf.html

    Parameters
    ----------
    z : complex
        The argument z of the function.

    akp : float
        The argument k' of the function.

    a : float
        The argument a of the function.

    b : float
        The argument b of the function.

    Returns
    -------
    f : complex
        The value of the general elliptic integral of the second kind
        F(z,k',a,b) for a complex argument z.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, abs(Im(`z`)) is too large: abs(Im(`z`)) =
            *<value>*. It must not exceed *<value>*.

        (`errno` 1)
            On entry, Re(`z`) is too large: Re(`z`) = *<value>*. It must
            not exceed *<value>*.

        (`errno` 1)
            On entry, Re(`z`) < 0.0: Re(`z`) = *<value>*.

        (`errno` 1)
            On entry, abs(`akp`) is too large: abs(`akp`) = *<value>*.
            It must not exceed *<value>*.

        (`errno` 2)
            The iterative procedure used to evaluate the integral has
            failed to converge.

    Notes
    -----
    ``ellipint_general_2`` evaluates an approximation to the general
    elliptic integral of the second kind F(z,k',a,b) given by

        F(z,k',a,b) = int_0^z(a+b
        zeta^2)/(1+zeta^2)sqrt((1+zeta^2)(1+k'^2 zeta^2))d zeta,

    where a and b are real arguments, z is a complex argument whose real
    part is non-negative and k' is a real argument (the `complementary
    modulus`).
    The evaluation of F is based on the Gauss transformation.
    Further details, in particular for the conformal mapping provided by
    F, can be found in Bulirsch (1960).

    Special values include

        F(z,k',1,1) = int_0^zd zeta/(sqrt((1+zeta^2)(1+k'2 zeta^2))),

    or F_1(z,k') (the `elliptic integral of the first kind`) and

        F(z,k',1,k'^2) = int_0^zsqrt(1+k'^2
        zeta^2)/(1+zeta^2)sqrt(1+zeta^2)d zeta,

    or F_2(z,k') (the `elliptic integral of the second kind`).
    Note that the values of F_1(z,k') and F_2(z,k') are equal to
    tan^(-1)(z) in the trivial case k' = 1.

    ``ellipint_general_2`` is derived from an Algol 60 procedure given
    by Bulirsch (1960).
    Constraints are placed on the values of z and k' in order to avoid
    the possibility of machine overflow.

    References
    ----------
    Bulirsch, R, 1960, `Numerical calculation of elliptic integrals and
    elliptic functions`, Numer. Math. (7), 76--90
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.ellipint_general_2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s21daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, z.
        utils._BYREF_FLOAT64, # Input, akp.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_COMPLEX128, # Output, f.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _akp = utils._EngineFloat64ScalarType(
        akp, {'fun_name': fun_name, 'entity': 'akp'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _f = utils._EngineComplex128ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _z,
        _akp,
        _a,
        _b,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(Im(z)) is too large:',
                    'abs(Im(z)) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'It must not exceed {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, Re(z) is too large: Re(z) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'It must not exceed {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, Re(z) < 0.0: Re(z) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(akp) is too large: abs(akp) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'It must not exceed {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to evaluate the integral has',
                    'failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_py = utils._EngineComplex128ScalarType.to_py(_f)
        return_data = _f_py
    _err_class.manage()
    return return_data

def legendre_p(mode, x, m, nl, p):
    u"""
    Legendre functions of 1st kind P_n^m(x) or -P_n^m - (x).

    ``legendre_p`` returns a sequence of values for either the
    unnormalized or normalized Legendre functions of the first kind
    P_n^m(x) or -P_n^m - (x) for real x of a given order m and degree n
    = 0,1,...,N.

    For full information please refer to the NAG Library document for
    s22aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22aaf.html

    Parameters
    ----------
    mode : int
        Indicates whether the sequence of function values is to be
        returned unnormalized or normalized.

        `mode` = 1

            The sequence of function values is returned unnormalized.

        `mode` = 2

            The sequence of function values is returned normalized.

    x : float
        The argument x of the function.

    m : int
        The order m of the function.

    nl : int
        The degree N of the last function required in the sequence.

    p : float, shape (`nl`+1), modified in place
        `On exit`: the required sequence of function values as follows:

            if `mode` = 1, `p`[n] contains P_n^m(x), for n = 0,1,...,N;

            if `mode` = 2, `p`[n] contains - P_n^m - (x), for n =
            0,1,...,N.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nl` = *<value>* and abs(`m`) = *<value>*.

            Constraint: `nl`+abs(`m`) <= 55.

        (`errno` 1)
            On entry, abs(`m`) = *<value>*.

            Constraint: abs(`m`) <= 27.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: when abs(`m`) != 0, `nl` >= 0.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` <= 100 when `m` = 0.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` >= 0.

        (`errno` 1)
            On entry, abs(`x`) = *<value>*.

            Constraint: abs(`x`) <= 1.0.

        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` <= 2.

        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` >= 1.

    Notes
    -----
    ``legendre_p`` evaluates a sequence of values for either the
    unnormalized or normalized Legendre (m = 0) or associated Legendre
    (m != 0) functions of the first kind P_n^m(x) or -P_n^m - (x), where
    x is real with -1 <= x <= 1, of order m and degree n = 0,1,...,N
    defined by

        [table omitted]

    respectively; P_n(x) is the (unassociated) Legendre polynomial of
    degree n given by

        P_n(x) = P_n^0(x) = 1/(2^nn!)(d^n)/dx^n(x^2-1)^n

    (the `Rodrigues formula`).
    Note that some authors (e.g., Abramowitz and Stegun (1972)) include
    an additional factor of (-1)^m (the `Condon--Shortley Phase`) in the
    definitions of P_n^m(x) and -P_n^m - (x).
    They use the notation P_{mn}(x) = (-1)^mP_n^m(x) in order to
    distinguish between the two cases.

    ``legendre_p`` is based on a standard recurrence relation described
    in Section 8.5.3 of Abramowitz and Stegun (1972).
    Constraints are placed on the values of m and n in order to avoid
    the possibility of machine overflow.
    It also sets the appropriate elements of the array `p` (see
    [Parameters]) to zero whenever the required function is not defined
    for certain values of m and n (e.g., m = -5 and n = 3).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.legendre_p'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, nl.
        utils._EngineFloat64ArrayType.c_type, # Output, p[nl+1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_nl_py+1,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _x,
        _m,
        _nl,
        _p,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s} and abs(m) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nl+abs(m) <= 55.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(m) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(m) <= 27.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: when abs(m) != 0, nl >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nl <= 100 when m = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nl >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, abs(x) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def hyperg_confl_real(a, b, x):
    u"""
    Real confluent hypergeometric function 1_F_1(a;b;x).

    ``hyperg_confl_real`` returns a value for the confluent
    hypergeometric function 1_F_1(a;b;x) with real parameters a and b,
    and real argument x.
    This function is sometimes also known as Kummer's function M(a,b,x).

    For full information please refer to the NAG Library document for
    s22ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22baf.html

    Parameters
    ----------
    a : float
        The parameter a of the function.

    b : float
        The parameter b of the function.

    x : float
        The argument x of the function.

    Returns
    -------
    m : float
        The solution M(a,b,x).

        Note: if overflow occurs upon completion, as indicated by
        `errno` = 2, |M(a,b,x)| may be assumed to be too large to be
        representable. `m` will be returned as +/-R_max, where R_max is
        the largest representable real number (see
        ``machine.real_largest``). The sign of `m` should match the sign
        of M(a,b,x). If overflow occurs during a subcalculation, as
        indicated by `errno` = 5, the sign may be incorrect, and the
        true value of M(a,b,x) may or may not be greater than R_max. In
        either case it is advisable to subsequently use
        ``hyperg_confl_real_scaled``.

    Raises
    ------
    NagValueError
        (`errno` 4)
            All approximations have completed, and the final residual
            estimate indicates no accuracy can be guaranteed.

            Relative residual = *<value>*.

        (`errno` 5)
            Overflow occurred in a subcalculation of M(a,b,x).

            The answer may be completely incorrect.

        (`errno` 11)
            On entry, `a` = *<value>*.

            Constraint: abs(`a`) <= arbnd = *<value>*.

        (`errno` 31)
            On entry, `b` = *<value>*.

            Constraint: abs(`b`) <= arbnd = *<value>*.

        (`errno` 32)
            On entry, `b` = *<value>*.

            M(a,b,x) is undefined when b is zero or a negative integer.

        (`errno` 51)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= arbnd = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Underflow occurred during the evaluation of M(a,b,x).

            The returned value may be inaccurate.

        (`errno` 2)
            On completion, overflow occurred in the evaluation of
            M(a,b,x).

        (`errno` 3)
            All approximations have completed, and the final residual
            estimate indicates some precision may have been lost.

            Relative residual = *<value>*.

    Notes
    -----
    ``hyperg_confl_real`` returns a value for the confluent
    hypergeometric function 1_F_1(a;b;x) with real parameters a and b,
    and real argument x.
    This function is unbounded or not uniquely defined for b equal to
    zero or a negative integer.

    The associated function ``hyperg_confl_real_scaled`` performs the
    same operations, but returns M in the scaled form M = m_f*2^(m_s) to
    allow calculations to be performed when M is not representable as a
    single working precision number.
    It also accepts the parameters a and b as summations of an integer
    and a decimal fraction, giving higher accuracy when a or b are close
    to an integer.
    In such cases, ``hyperg_confl_real_scaled`` should be used when high
    accuracy is required.

    The confluent hypergeometric function is defined by the confluent
    series

        1_F_1(a;b;x) = M(a,b,x) = sum((a)_s x^s/(b)_s s!) =
        1+a/bx+a(a+1)/b(b+1)2!x^2+⋯

    where (a)_s = 1*(a)*(a+1)*(a+2)...(a+s-1) is the rising factorial of
    a. M(a,b,x) is a solution to the second order ODE (Kummer's
    Equation):

        xd^2M/dx^2+(b-x)dM/dx-aM = 0.

    Given the parameters and argument (a,b,x), this function determines
    a set of safe values {(alpha_i,beta_i,zeta_i) | i <= 2} and selects
    an appropriate algorithm to accurately evaluate the functions
    M_i(alpha_i,beta_i,zeta_i).
    The result is then used to construct the solution to the original
    problem M(a,b,x) using, where necessary, recurrence relations and/or
    continuation.

    Additionally, an artificial bound, arbnd is placed on the magnitudes
    of a, b and x to minimize the occurrence of overflow in internal
    calculations. arbnd = 0.0001*I_max, where I_max =
    machine.integer_max.
    It should, however, not be assumed that this function will produce
    an accurate result for all values of a, b and x satisfying this
    criterion.

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the confluent hypergeometric function
    including special cases, transformations, relations and asymptotic
    approximations.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Pearson, J, 2009, `Computation of hypergeometric functions`, MSc
    Dissertation, Mathematical Institute, University of Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.hyperg_confl_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, m.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _m = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _x,
        _m,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_s_consts.S22BA_IERR_UFLOW and _ifmt == _nag_s_consts.S22BA_IFMT_UFLOW:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Underflow occurred during the evaluation of M(a,b,x).',
                    'The returned value may be inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_OFLOW and _ifmt == _nag_s_consts.S22BA_IFMT_OFLOW_FULLCALC:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On completion, overflow occurred in the evaluation of M(a,b,x).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_WARN_RES and _ifmt == _nag_s_consts.S22BA_IFMT_RES_TOL1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate',
                    'indicates some precision may have been lost.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_RES and _ifmt == _nag_s_consts.S22BA_IFMT_RES_NOPREC:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate',
                    'indicates no accuracy can be guaranteed.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_OFLOW_SUB and _ifmt == _nag_s_consts.S22BA_IFMT_OFLOW_SUBCALC:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow occurred in a subcalculation of M(a,b,x).',
                    'The answer may be completely incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_A_ARBND and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(a) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_B_ARBND and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(b) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_B_NEGI and _ifmt == _nag_s_consts.S22BA_IFMT_B_NEGI:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'M(a,b,x) is undefined when b is zero or a negative integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_X and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _m_py = utils._EngineFloat64ScalarType.to_py(_m)
        return_data = _m_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['m'],
            )(
                _m_py,
            )
    _err_class.manage()
    return return_data

def hyperg_confl_real_scaled(ani, adr, bni, bdr, x):
    u"""
    Real confluent hypergeometric function 1_F_1(a;b;x) in scaled form.

    ``hyperg_confl_real_scaled`` returns a value for the confluent
    hypergeometric function 1_F_1(a;b;x), with real parameters a and b
    and real argument x.
    The solution is returned in the scaled form 1_F_1(a;b;x) =
    m_f*2^(m_s).
    This function is sometimes also known as Kummer's function M(a,b,x).

    For full information please refer to the NAG Library document for
    s22bb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22bbf.html

    Parameters
    ----------
    ani : float
        a_i, the nearest integer to a, satisfying a_i = a-a_r.

    adr : float
        a_r, the signed decimal remainder satisfying a_r = a-a_i and
        |a_r| <= 0.5.

        Note: if |`adr`| < 100.0 epsilon, a_r = 0.0 will be used, where
        epsilon is the machine precision as returned by
        ``machine.precision``.

    bni : float
        b_i, the nearest integer to b, satisfying b_i = b-b_r.

    bdr : float
        b_r, the signed decimal remainder satisfying b_r = b-b_i and
        |b_r| <= 0.5.

        Note: if |`bdr`-`adr`| < 100.0 epsilon, a_r = b_r will be used,
        where epsilon is the machine precision as returned by
        ``machine.precision``.

    x : float
        The argument x of the function.

    Returns
    -------
    frm : float
        m_f, the scaled real component of the solution satisfying m_f =
        M(a,b,x)*2^(-m_s).

        Note: if overflow occurs upon completion, as indicated by
        `errno` = 2, the value of m_f returned may still be correct. If
        overflow occurs in a subcalculation, as indicated by `errno` =
        5, this should not be assumed.

    scm : int
        m_s, the scaling power of two, satisfying m_s =
        log2(M(a,b,x)/(m_f)).

        Note: if overflow occurs upon completion, as indicated by
        `errno` = 2, then m_s >= I_max, where I_max is the largest
        representable integer (see ``machine.integer_max``). If overflow
        occurs during a subcalculation, as indicated by `errno` = 5, m_s
        may or may not be greater than I_max. In either case, `scm` =
        machine.integer_max will have been returned.

    Raises
    ------
    NagValueError
        (`errno` 4)
            All approximations have completed, and the final residual
            estimate indicates no accuracy can be guaranteed.

            Relative residual = *<value>*.

        (`errno` 5)
            Overflow occurred in a subcalculation of M(a,b,x).

            The answer may be completely incorrect.

        (`errno` 11)
            On entry, `ani` = *<value>*.

            Constraint: abs(`ani`) <= arbnd = *<value>*.

        (`errno` 13)
            `ani` is non-integral.

            On entry, `ani` = *<value>*.

            Constraint: `ani` = floor(`ani`).

        (`errno` 21)
            On entry, `adr` = *<value>*.

            Constraint: abs(`adr`) <= 0.5.

        (`errno` 31)
            On entry, `bni` = *<value>*.

            Constraint: abs(`bni`) <= arbnd = *<value>*.

        (`errno` 32)
            On entry, b = `bni`+`bdr` = *<value>*.

            M(a,b,x) is undefined when b is zero or a negative integer.

        (`errno` 33)
            `bni` is non-integral.

            On entry, `bni` = *<value>*.

            Constraint: `bni` = floor(`bni`).

        (`errno` 41)
            On entry, `bdr` = *<value>*.

            Constraint: abs(`bdr`) <= 0.5.

        (`errno` 51)
            On entry, `x` = *<value>*.

            Constraint: abs(`x`) <= arbnd = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Underflow occurred during the evaluation of M(a,b,x).

            The returned value may be inaccurate.

        (`errno` 2)
            On completion, overflow occurred in the evaluation of
            M(a,b,x).

        (`errno` 3)
            All approximations have completed, and the final residual
            estimate indicates some precision may have been lost.

            Relative residual = *<value>*.

    Notes
    -----
    ``hyperg_confl_real_scaled`` returns a value for the confluent
    hypergeometric function 1_F_1(a;b;x), with real parameters a and b
    and real argument x, in the scaled form 1_F_1(a;b;x) = m_f*2^(m_s),
    where m_f is the real scaled component and m_s is the integer power
    of two scaling.
    This function is unbounded or not uniquely defined for b equal to
    zero or a negative integer.

    The confluent hypergeometric function is defined by the confluent
    series,

        1_F_1(a;b;x) = M(a,b,x) = sum((a)_s x^s/(b)_s s!) =
        1+a/bx+a(a+1)/b(b+1)2!x^2+⋯

    where (a)_s = 1*(a)*(a+1)*(a+2)...(a+s-1) is the rising factorial of
    a. M(a,b,x) is a solution to the second order ODE (Kummer's
    Equation):

        xd^2M/dx^2+(b-x)dM/dx-aM = 0.

    Given the parameters and argument (a,b,x), this function determines
    a set of safe values {(alpha_i,beta_i,zeta_i) | i <= 2} and selects
    an appropriate algorithm to accurately evaluate the functions
    M_i(alpha_i,beta_i,zeta_i).
    The result is then used to construct the solution to the original
    problem M(a,b,x) using, where necessary, recurrence relations and/or
    continuation.

    For improved precision in the final result, this function accepts a
    and b split into an integral and a decimal fractional component.
    Specifically a = a_i+a_r, where |a_r| <= 0.5 and a_i = a-a_r is
    integral. b is similarly deconstructed.

    Additionally, an artificial bound, arbnd is placed on the magnitudes
    of a_i, b_i and x to minimize the occurrence of overflow in internal
    calculations. arbnd = 0.0001*I_max, where I_max =
    machine.integer_max.
    It should, however, not be assumed that this function will produce
    an accurate result for all values of a_i, b_i and x satisfying this
    criterion.

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the confluent hypergeometric function
    including special cases, transformations, relations and asymptotic
    approximations.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Pearson, J, 2009, `Computation of hypergeometric functions`, MSc
    Dissertation, Mathematical Institute, University of Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.hyperg_confl_real_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, ani.
        utils._BYREF_FLOAT64, # Input, adr.
        utils._BYREF_FLOAT64, # Input, bni.
        utils._BYREF_FLOAT64, # Input, bdr.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, frm.
        utils._BYREF_INT, # Output, scm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ani = utils._EngineFloat64ScalarType(
        ani, {'fun_name': fun_name, 'entity': 'ani'},
    ).to_ctypes()
    _adr = utils._EngineFloat64ScalarType(
        adr, {'fun_name': fun_name, 'entity': 'adr'},
    ).to_ctypes()
    _bni = utils._EngineFloat64ScalarType(
        bni, {'fun_name': fun_name, 'entity': 'bni'},
    ).to_ctypes()
    _bdr = utils._EngineFloat64ScalarType(
        bdr, {'fun_name': fun_name, 'entity': 'bdr'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _frm = utils._EngineFloat64ScalarType.empty_ctypes()
    _scm = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ani,
        _adr,
        _bni,
        _bdr,
        _x,
        _frm,
        _scm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_s_consts.S22BA_IERR_UFLOW and _ifmt == _nag_s_consts.S22BA_IFMT_UFLOW:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Underflow occurred during the evaluation of M(a,b,x).',
                    'The returned value may be inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_OFLOW and _ifmt == _nag_s_consts.S22BA_IFMT_OFLOW_FULLCALC:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On completion, overflow occurred in the evaluation of M(a,b,x).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_WARN_RES and _ifmt == _nag_s_consts.S22BA_IFMT_RES_TOL1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate',
                    'indicates some precision may have been lost.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_RES and _ifmt == _nag_s_consts.S22BA_IFMT_RES_NOPREC:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate',
                    'indicates no accuracy can be guaranteed.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_OFLOW_SUB and _ifmt == _nag_s_consts.S22BA_IFMT_OFLOW_SUBCALC:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow occurred in a subcalculation of M(a,b,x).',
                    'The answer may be completely incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_AI_ARBND and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ani = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(ani) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_AI_INT and _ifmt == _nag_s_consts.S22BB_IFMT_INT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ani is non-integral.',
                    'On entry, ani = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ani = floor(ani).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_AR and _ifmt == _nag_s_consts.S22BB_IFMT_MODR:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, adr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(adr) <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_BI_ARBND and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bni = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(bni) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_BI_NEGI and _ifmt == _nag_s_consts.S22BA_IFMT_B_NEGI:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = bni+bdr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'M(a,b,x) is undefined when b is zero or a negative integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_BI_INT and _ifmt == _nag_s_consts.S22BB_IFMT_INT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'bni is non-integral.',
                    'On entry, bni = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bni = floor(bni).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BB_IERR_BR and _ifmt == _nag_s_consts.S22BB_IFMT_MODR:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bdr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(bdr) <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BA_IERR_X and _ifmt == _nag_s_consts.S22BA_IFMT_ARBND:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(x) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _frm_py = utils._EngineFloat64ScalarType.to_py(_frm)
        _scm_py = utils._EngineIntScalarType.to_py(_scm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['frm', 'scm'],
        )(
            _frm_py,
            _scm_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def hyperg_gauss_real(a, b, c, x):
    u"""
    Real Gauss hypergeometric function 2_F_1(a,b;c;x).

    ``hyperg_gauss_real`` returns a value for the Gauss hypergeometric
    function 2_F_1(a,b;c;x) for real parameters a,b and c, and real
    argument x.

    For full information please refer to the NAG Library document for
    s22be

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22bef.html

    Parameters
    ----------
    a : float
        The parameter a.

    b : float
        The parameter b.

    c : float
        The parameter c.

    x : float
        The argument x.

    Returns
    -------
    f21 : float
        The value returned for the Gauss hypergeometric function
        2_F_1(a,b;c;x) for real parameters a,b and c, and real argument
        x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            All approximations have completed, and the final residual
            estimate indicates no accuracy can be guaranteed.

            Relative residual = *<value>*.

        (`errno` 4)
            On entry, `x` = *<value>*, c = *<value>*, a+b = *<value>*.

            2_F_1(a,b;c;1) is infinite in the case c <= a+b.

        (`errno` 6)
            Overflow occurred in a subcalculation of 2_F_1(a,b;c;x). The
            result may or may not be infinite.

        (`errno` 9)
            An internal calculation has resulted in an undefined result.

        (`errno` 11)
            On entry, `a` does not satisfy abs(`a`) <= arbnd =
            *<value>*.

        (`errno` 21)
            On entry, `b` does not satisfy abs(`b`) <= arbnd =
            *<value>*.

        (`errno` 31)
            On entry, `c` does not satisfy abs(`c`) <= arbnd =
            *<value>*.

        (`errno` 32)
            On entry, `c` = *<value>*.

            2_F_1(a,b;c;x) is undefined when c is zero or a negative
            integer.

        (`errno` 41)
            On entry, `x` does not satisfy abs(`x`) <= arbnd =
            *<value>*.

        (`errno` 42)
            On entry, `x` = *<value>*.

            In general, 2_F_1(a,b;c;x) is not real valued when x > 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Underflow occurred during the evaluation of 2_F_1(a,b;c;x).
            The returned value may be inaccurate.

        (`errno` 2)
            All approximations have completed, and the final residual
            estimate indicates some precision may have been lost.

            Relative residual = *<value>*.

        (`errno` 5)
            On completion, overflow occurred in the evaluation of
            2_F_1(a,b;c;x).

    Notes
    -----
    ``hyperg_gauss_real`` returns a value for the Gauss hypergeometric
    function 2_F_1(a,b;c;x) for real parameters a, b and c, and for real
    argument x.

    The associated function ``hyperg_gauss_real_scaled`` performs the
    same operations, but returns 2_F_1(a,b;c;x) in the scaled form
    2_F_1(a,b;c;x) = f_fr*2^(f_sc) to allow calculations to be performed
    when 2_F_1(a,b;c;x) is not representable as a single working
    precision number.
    It also accepts the parameters a, b and c as summations of an
    integer and a decimal fraction, giving higher accuracy when any are
    close to an integer.

    The Gauss hypergeometric function is a solution to the
    hypergeometric differential equation,

        x(1-x)d^2f/dx^2+(c-(a+b+1)x)df/dx-abf = 0.

    For |x| < 1, it may be defined by the Gauss series,

        2_F_1(a,b;c;x) = sum((a)_s(b)_s/(c)_s s!x^s) =
        1+ab/cx+a(a+1)b(b+1)/c(c+1)2!x^2+⋯,

    where (a)_s = 1*(a)*(a+1)*(a+2)...(a+s-1) is the rising factorial of
    a. 2_F_1(a,b;c;x) is undefined for c = 0 or c a negative integer.

    For |x| < 1, the series is absolutely convergent and 2_F_1(a,b;c;x)
    is finite.

    For x < 1, linear transformations of the form,

        2_F_1(a,b;c;x) =
        C_1(a_1,b_1,c_1,x_1)2_F_1(a_1,b_1;c_1;x_1)+C_2(a_2,b_2,c_2,x_2)2_F_1(a_2,b_2;c_2;x_2)

    exist, where x_1, x_2 in (0,1]. C_1 and C_2 are real valued
    functions of the parameters and argument, typically involving
    products of gamma functions.
    When these are degenerate, finite limiting cases exist.
    Hence for x < 0, 2_F_1(a,b;c;x) is defined by analytic continuation,
    and for x < 1, 2_F_1(a,b;c;x) is real and finite.

    For x = 1, the following apply:

        If c > a+b, 2_F_1(a,b;c;1) =
        Gamma(c)Gamma(c-a-b)/(Gamma(c-a)Gamma(c-b)), and hence is
        finite. Solutions also exist for the degenerate cases where c-a
        or c-b are negative integers or zero.

        If c <= a+b, 2_F_1(a,b;c;1) is infinite, and the sign of
        2_F_1(a,b;c;1) is determinable as x approaches 1 from below.

    In the complex plane, the principal branch of 2_F_1(a,b;c;z) is
    taken along the real axis from x = 1.0 increasing. 2_F_1(a,b;c;z) is
    multivalued along this branch, and for real parameters a,b and c is
    typically not real valued.
    As such, this function will not compute a solution when x > 1.

    The solution strategy used by this function is primarily dependent
    upon the value of the argument x.
    Once trivial cases and the case x = 1.0 are eliminated, this
    proceeds as follows.

    For 0 < x <= 0.5, sets of safe parameters
    {alpha_{i,j};beta_{i,j};zeta_{i,j};chi_j|1 <= j <= 2|;1 <= i <= 4}
    are determined, such that the values of 2_F_1(a_j,b_j;c_j;x_j)
    required for an appropriate transformation of the type (3) may be
    calculated either directly or using recurrence relations from the
    solutions of 2_F_1(alpha_{i,j},beta_{i,j};zeta_{i,j};chi_j).
    If c is positive, then only transformations with C_2 = 0.0 will be
    used, implying only 2_F_1(a_1,b_1;c_1;x_1) will be required, with
    the transformed argument x_1 = x.
    If c is negative, in some cases a transformation with C_2 != 0.0
    will be used, with the argument x_2 = 1.0-x.
    The function then cycles through these sets until acceptable
    solutions are generated.
    If no computation produces an accurate answer, the least inaccurate
    answer is selected to complete the computation.
    See [Accuracy].

    For 0.5 < x < 1.0, an identical approach is first used with the
    argument x.
    Should this fail, a linear transformation resulting in both
    transformed arguments satisfying x_j = 1.0-x is employed, and the
    above strategy for 0 < x <= 0.5 is utilized on both components.
    Further transformations in these sub-computations are however,
    limited to single terms with no argument transformation.

    For x < 0, a linear transformation mapping the argument x to the
    interval (0,0.5] is first employed.
    The strategy for 0 < x <= 0.5 is then used on each component,
    including possible further two term transforms.
    To avoid some degenerate cases, a transform mapping the argument x
    to [0.5,1) may also be used.

    In addition to the above restrictions on c and x, an artificial
    bound, arbnd, is placed on the magnitudes of a,b,c and x to minimize
    the occurrence of overflow in internal calculations, particularly
    those involving real to integer conversions. arbnd = 0.0001*I_max,
    where I_max is the largest machine integer (see
    ``machine.integer_max``).
    It should however, not be assumed that this function will produce
    accurate answers for all values of a,b,c and x satisfying this
    criterion.

    This function also tests for non-finite values of the parameters and
    argument on entry, and assigns non-finite values upon completion if
    appropriate.
    See [Further Comments] and submodule ``ieee``.

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Gauss hypergeometric function
    including special cases, transformations, relations and asymptotic
    approximations.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Pearson, J, 2009, `Computation of hypergeometric functions`, MSc
    Dissertation, Mathematical Institute, University of Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.hyperg_gauss_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, f21.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _f21 = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _x,
        _f21,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_s_consts.S22BF_IERR_UFLOW and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Underflow occurred during the evaluation of 2_F_1(a,b;c;x).',
                    'The returned value may be inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_WARN_RES and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate indicates',
                    'some precision may have been lost.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_RES and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate indicates',
                    'no accuracy can be guaranteed.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_X_EQ1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, c = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'a+b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    '2_F_1(a,b;c;1) is infinite in the case c <= a+b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_OFLOW_SAFE and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On completion, overflow occurred in the evaluation of',
                    '2_F_1(a,b;c;x).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_OFLOW_SUB and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow occurred in a subcalculation of 2_F_1(a,b;c;x).',
                    'The result may or may not be infinite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_NAN and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal calculation has resulted in an undefined result.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_A_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a does not satisfy abs(a) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_B_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b does not satisfy abs(b) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_C_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, c does not satisfy abs(c) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_C_NEGI and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, c = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    '2_F_1(a,b;c;x) is undefined when c is zero or a negative integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_X_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x does not satisfy abs(x) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BE_IERR_X_GT1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'In general, 2_F_1(a,b;c;x) is not real valued when x > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _f21_py = utils._EngineFloat64ScalarType.to_py(_f21)
        return_data = _f21_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['f21'],
            )(
                _f21_py,
            )
    _err_class.manage()
    return return_data

def hyperg_gauss_real_scaled(ani, adr, bni, bdr, cni, cdr, x):
    u"""
    Real Gauss hypergeometric function 2_F_1(a,b;c;x) in scaled form.

    ``hyperg_gauss_real_scaled`` returns a value for the Gauss
    hypergeometric function 2_F_1(a,b;c;x) for real parameters a,b and
    c, and real argument x.
    The result is returned in the scaled form 2_F_1(a,b;c;x) =
    f_fr*2^(f_sc).

    For full information please refer to the NAG Library document for
    s22bf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22bff.html

    Parameters
    ----------
    ani : float
        a_i, the nearest integer to a, satisfying a_i = a-a_r.

    adr : float
        a_r, the signed decimal remainder satisfying a_r = a-a_i and
        |a_r| <= 0.5.

    bni : float
        b_i, the nearest integer to b, satisfying b_i = b-b_r.

    bdr : float
        b_r, the signed decimal remainder satisfying b_r = b-b_i and
        |b_r| <= 0.5.

    cni : float
        c_i, the nearest integer to c, satisfying c_i = c-c_r.

    cdr : float
        c_r, the signed decimal remainder satisfying c_r = c-c_i and
        |c_r| <= 0.5.

    x : float
        The argument x.

    Returns
    -------
    frf : float
        f_fr, the scaled real component of the solution satisfying f_fr
        = 2_F_1(a,b;c;x)*2^(-f_sc), i.e., 2_F_1(a,b;c;x) =
        f_fr*2^(f_sc). See [Further Comments] for the behaviour of f_fr
        when a finite or non-finite answer is returned.

    scf : int
        f_sc, the scaling power of two, satisfying f_sc =
        log2(2_F_1(a,b;c;x)/f_fr), i.e., 2_F_1(a,b;c;x) = f_fr*2^(f_sc).
        See [Further Comments] for the behaviour of f_sc when a
        non-finite answer is returned.

    Raises
    ------
    NagValueError
        (`errno` 3)
            All approximations have completed, and the final residual
            estimate indicates no accuracy can be guaranteed.

            Relative residual = *<value>*.

        (`errno` 4)
            On entry, `x` = *<value>*, c = *<value>*, a+b = *<value>*.

            2_F_1(a,b;c;1) is infinite in the case c <= a+b.

        (`errno` 6)
            Overflow occurred in a subcalculation of 2_F_1(a,b;c;x). The
            answer may be completely incorrect.

        (`errno` 9)
            An internal calculation has resulted in an undefined result.

        (`errno` 11)
            On entry, `ani` does not satisfy abs(`ani`) <= arbnd =
            *<value>*.

        (`errno` 13)
            `ani` is non-integral.

            On entry, `ani` = *<value>*.

            Constraint: `ani` = floor(`ani`).

        (`errno` 21)
            On entry, `adr` does not satisfy abs(`adr`) <= 0.5.

        (`errno` 31)
            On entry, `bni` does not satisfy abs(`bni`) <= arbnd =
            *<value>*.

        (`errno` 33)
            `bni` is non-integral.

            On entry, `bni` = *<value>*.

            Constraint: `bni` = floor(`bni`).

        (`errno` 41)
            On entry, `bdr` does not satisfy abs(`bdr`) <= 0.5.

        (`errno` 51)
            On entry, `cni` does not satisfy abs(`cni`) <= arbnd =
            *<value>*.

        (`errno` 52)
            On entry, c = `cni`+`cdr` = *<value>*.

            2_F_1(a,b;c;x) is undefined when c is zero or a negative
            integer.

        (`errno` 53)
            `cni` is non-integral.

            On entry, `cni` = *<value>*.

            Constraint: `cni` = floor(`cni`).

        (`errno` 61)
            On entry, `cdr` does not satisfy abs(`cdr`) <= 0.5.

        (`errno` 71)
            On entry, `x` does not satisfy abs(`x`) <= arbnd =
            *<value>*.

        (`errno` 72)
            On entry, `x` = *<value>*.

            In general, 2_F_1(a,b;c;x) is not real valued when x > 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Underflow occurred during the evaluation of 2_F_1(a,b;c;x).
            The returned value may be inaccurate.

        (`errno` 2)
            All approximations have completed, and the final residual
            estimate indicates some precision may have been lost.

            Relative residual = *<value>*.

        (`errno` 5)
            On completion, overflow occurred in the evaluation of
            2_F_1(a,b;c;x).

    Notes
    -----
    ``hyperg_gauss_real_scaled`` returns a value for the Gauss
    hypergeometric function 2_F_1(a,b;c;x) for real parameters a, b and
    c, and for real argument x.

    The Gauss hypergeometric function is a solution to the
    hypergeometric differential equation,

        x(1-x)d^2f/dx^2+(c-(a+b+1)x)df/dx-abf = 0.

    For |x| < 1, it may be defined by the Gauss series,

        2_F_1(a,b;c;x) = sum((a)_s(b)_s/(c)_s s!x^s) =
        1+ab/cx+a(a+1)b(b+1)/c(c+1)2!x^2+⋯,

    where (a)_s = 1*(a)*(a+1)*(a+2)...(a+s-1) is the rising factorial of
    a. 2_F_1(a,b;c;x) is undefined for c = 0 or c a negative integer.

    For |x| < 1, the series is absolutely convergent and 2_F_1(a,b;c;x)
    is finite.

    For x < 1, linear transformations of the form,

        2_F_1(a,b;c;x) =
        C_1(a_1,b_1,c_1,x_1)2_F_1(a_1,b_1;c_1;x_1)+C_2(a_2,b_2,c_2,x_2)2_F_1(a_2,b_2;c_2;x_2)

    exist, where x_1, x_2 in (0,1]. C_1 and C_2 are real valued
    functions of the parameters and argument, typically involving
    products of gamma functions.
    When these are degenerate, finite limiting cases exist.
    Hence for x < 0, 2_F_1(a,b;c;x) is defined by analytic continuation,
    and for x < 1, 2_F_1(a,b;c;x) is real and finite.

    For x = 1, the following apply:

        If c > a+b, 2_F_1(a,b;c;1) =
        Gamma(c)Gamma(c-a-b)/(Gamma(c-a)Gamma(c-b)), and hence is
        finite. Solutions also exist for the degenerate cases where c-a
        or c-b are negative integers or zero.

        If c <= a+b, 2_F_1(a,b;c;1) is infinite, and the sign of
        2_F_1(a,b;c;1) is determinable as x approaches 1 from below.

    In the complex plane, the principal branch of 2_F_1(a,b;c;z) is
    taken along the real axis from x = 1.0 increasing. 2_F_1(a,b;c;z) is
    multivalued along this branch, and for real parameters a,b and c is
    typically not real valued.
    As such, this function will not compute a solution when x > 1.

    The solution strategy used by this function is primarily dependent
    upon the value of the argument x.
    Once trivial cases and the case x = 1.0 are eliminated, this
    proceeds as follows.

    For 0 < x <= 0.5, sets of safe parameters
    {alpha_{i,j},beta_{i,j},zeta_{i,j},chi_j|1 <= j <= 2|,1 <= i <= 4}
    are determined, such that the values of 2_F_1(a_j,b_j;c_j;x_j)
    required for an appropriate transformation of the type (3) may be
    calculated either directly or using recurrence relations from the
    solutions of 2_F_1(alpha_{i,j},beta_{i,j};zeta_{i,j};chi_j).
    If c is positive, then only transformations with C_2 = 0.0 will be
    used, implying only 2_F_1(a_1,b_1;c_1;x_1) will be required, with
    the transformed argument x_1 = x.
    If c is negative, in some cases a transformation with C_2 != 0.0
    will be used, with the argument x_2 = 1.0-x.
    The function then cycles through these sets until acceptable
    solutions are generated.
    If no computation produces an accurate answer, the least inaccurate
    answer is selected to complete the computation.
    See [Accuracy].

    For 0.5 < x < 1.0, an identical approach is first used with the
    argument x.
    Should this fail, a linear transformation resulting in both
    transformed arguments satisfying x_j = 1.0-x is employed, and the
    above strategy for 0 < x <= 0.5 is utilized on both components.
    Further transformations in these sub-computations are however,
    limited to single terms with no argument transformation.

    For x < 0, a linear transformation mapping the argument x to the
    interval (0,0.5] is first employed.
    The strategy for 0 < x <= 0.5 is then used on each component,
    including possible further two term transforms.
    To avoid some degenerate cases, a transform mapping the argument x
    to [0.5,1) may also be used.

    For improved precision in the final result, this function accepts
    a,b and c split into an integral and a decimal fractional component.
    Specifically, a = a_i+a_r, where |a_r| <= 0.5 and a_i = a-a_r is
    integral.
    The other parameters b and c are similarly deconstructed.

    In addition to the above restrictions on c and x, an artificial
    bound, arbnd, is placed on the magnitudes of a,b,c and x to minimize
    the occurrence of overflow in internal calculations, particularly
    those involving real to integer conversions. arbnd = 0.0001*I_max,
    where I_max is the largest machine integer (see
    ``machine.integer_max``).
    It should however, not be assumed that this function will produce
    accurate answers for all values of a,b,c and x satisfying this
    criterion.

    This function also tests for non-finite values of the parameters and
    argument on entry, and assigns non-finite values upon completion if
    appropriate.
    See [Further Comments] and submodule ``ieee``.

    Please consult the NIST Digital Library of Mathematical Functions
    for a detailed discussion of the Gauss hypergeometric function
    including special cases, transformations, relations and asymptotic
    approximations.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Pearson, J, 2009, `Computation of hypergeometric functions`, MSc
    Dissertation, Mathematical Institute, University of Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.hyperg_gauss_real_scaled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, ani.
        utils._BYREF_FLOAT64, # Input, adr.
        utils._BYREF_FLOAT64, # Input, bni.
        utils._BYREF_FLOAT64, # Input, bdr.
        utils._BYREF_FLOAT64, # Input, cni.
        utils._BYREF_FLOAT64, # Input, cdr.
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, frf.
        utils._BYREF_INT, # Output, scf.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ani = utils._EngineFloat64ScalarType(
        ani, {'fun_name': fun_name, 'entity': 'ani'},
    ).to_ctypes()
    _adr = utils._EngineFloat64ScalarType(
        adr, {'fun_name': fun_name, 'entity': 'adr'},
    ).to_ctypes()
    _bni = utils._EngineFloat64ScalarType(
        bni, {'fun_name': fun_name, 'entity': 'bni'},
    ).to_ctypes()
    _bdr = utils._EngineFloat64ScalarType(
        bdr, {'fun_name': fun_name, 'entity': 'bdr'},
    ).to_ctypes()
    _cni = utils._EngineFloat64ScalarType(
        cni, {'fun_name': fun_name, 'entity': 'cni'},
    ).to_ctypes()
    _cdr = utils._EngineFloat64ScalarType(
        cdr, {'fun_name': fun_name, 'entity': 'cdr'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _frf = utils._EngineFloat64ScalarType.empty_ctypes()
    _scf = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ani,
        _adr,
        _bni,
        _bdr,
        _cni,
        _cdr,
        _x,
        _frf,
        _scf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_s_consts.S22BF_IERR_UFLOW and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Underflow occurred during the evaluation of 2_F_1(a,b;c;x).',
                    'The returned value may be inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_WARN_RES and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate indicates',
                    'some precision may have been lost.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_RES and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All approximations have completed, and the final residual estimate indicates',
                    'no accuracy can be guaranteed.',
                    'Relative residual = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_X_EQ1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}, c = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'a+b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    '2_F_1(a,b;c;1) is infinite in the case c <= a+b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_OFLOW_SAFE and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On completion, overflow occurred in the evaluation of',
                    '2_F_1(a,b;c;x).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_OFLOW_SUB and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow occurred in a subcalculation of 2_F_1(a,b;c;x).',
                    'The answer may be completely incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_NAN and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal calculation has resulted in an undefined result.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_ANI_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ani does not satisfy abs(ani) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_ANI_NOTINT and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ani is non-integral.',
                    'On entry, ani = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ani = floor(ani).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_ADR and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, adr does not satisfy abs(adr) <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_BNI_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bni does not satisfy abs(bni) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_BNI_NOTINT and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'bni is non-integral.',
                    'On entry, bni = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bni = floor(bni).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_BDR and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bdr does not satisfy abs(bdr) <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_CNI_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cni does not satisfy abs(cni) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_C_NEGI and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, c = cni+cdr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    '2_F_1(a,b;c;x) is undefined when c is zero or a negative integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_CNI_NOTINT and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'cni is non-integral.',
                    'On entry, cni = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: cni = floor(cni).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_CDR and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cdr does not satisfy abs(cdr) <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_X_ARBND and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x does not satisfy abs(x) <= arbnd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_s_consts.S22BF_IERR_X_GT1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'In general, 2_F_1(a,b;c;x) is not real valued when x > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _frf_py = utils._EngineFloat64ScalarType.to_py(_frf)
        _scf_py = utils._EngineIntScalarType.to_py(_scf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['frf', 'scf'],
        )(
            _frf_py,
            _scf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def mathieu_ang_periodic_real(ordval, q, parity, mode, n, x, f, f_deriv):
    u"""
    Calculates values of real periodic angular Mathieu functions.

    ``mathieu_ang_periodic_real`` calculates real-valued periodic
    angular Mathieu functions (ce_m(x,q) or se_m(x,q)) and/or their
    first derivatives, where ce_m(x,q) and se_m(x,q) are solutions to
    the Mathieu differential equation d^2y/dx^2+[a-2qcos(2x)]y = 0.

    For full information please refer to the NAG Library document for
    s22ca

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s22caf.html

    Parameters
    ----------
    ordval : int
        m, the order number of the Mathieu function to be computed.

    q : float
        q, the Mathieu function parameter.

    parity : int
        Specifies whether to compute even or odd Mathieu function.

        `parity` = 0

            Compute even Mathieu function, ce_m(x,q).

        `parity` = 1

            Compute odd Mathieu function, se_m(x,q).

    mode : int
        Specifies whether the Mathieu function or its derivative is
        required.

        `mode` = 0

            Compute Mathieu function values.

        `mode` = 1

            Compute derivative values of Mathieu function.

        `mode` = 2

            Compute both Mathieu function and derivative values.

        `mode` = 3

            Compute neither Mathieu functions nor derivative values,
            returns only `a` (the characteristic value).

    n : int
        n, the number of values to compute.

    x : None or float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `mode`!=3: `n`; otherwise: 0.

        The values of x at which to compute Mathieu function or
        derivative values.

    f : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `mode` in (0, 2): `n`; otherwise: 0.

        `On exit`, if not None on entry: if `mode` = 0 or 2, the Mathieu
        function values ce_m(x,q) or se_m(x,q). If `mode` = 1 or 3, `f`
        is not used.

    f_deriv : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `mode` in (1, 2): `n`; otherwise: 0.

        `On exit`, if not None on entry: if `mode` = 1 or 2, the Mathieu
        function derivative values ce'_m(x,q) or se'_m(x,q). If `mode` =
        0 or 3, `f_deriv` is not used.

    Returns
    -------
    a : float
        a, the characteristic value for the Mathieu function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ordval` = *<value>* and `parity` = *<value>*.

            Constraint: if `parity` = 0, `ordval` >= 0 or if `parity` =
            1, `ordval` >= 1.

        (`errno` 2)
            On entry, `parity` = *<value>*.

            Constraint: `parity` = 0 or 1.

        (`errno` 3)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 4)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Notes
    -----
    ``mathieu_ang_periodic_real`` calculates an approximation to
    ce_m(x,q) and/or ce'_m(x,q), or se_m(x,q) and/or se'_m(x,q), where
    ce_m(x,q) and se_m(x,q) are respectively the even and odd parity
    real-valued periodic angular Mathieu functions, for an array of
    values of x, and for integer order value m, where m >= 0 for even
    parity, and m >= 1 for odd parity.
    The function also returns values of a for these periodic Mathieu
    functions, this is known as the characteristic value or eigenvalue.

    The solutions are computed by approximating Mathieu functions as
    Fourier series, where expansion coefficients are obtained by solving
    the eigenvalue problem generated from the relevant recurrence
    relation, see Module 28 in NIST Digital Library of Mathematical
    Functions.

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.mathieu_ang_periodic_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s22caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ordval.
        utils._BYREF_FLOAT64, # Input, q.
        utils._BYREF_INT, # Input, parity.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[*].
        utils._EngineFloat64ArrayType.c_type, # Output, f[*].
        utils._EngineFloat64ArrayType.c_type, # Output, f_deriv[*].
        utils._BYREF_FLOAT64, # Output, a.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ordval = utils._EngineIntScalarType(
        ordval, {'fun_name': fun_name, 'entity': 'ordval'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _parity = utils._EngineIntScalarType(
        parity, {'fun_name': fun_name, 'entity': 'parity'},
    ).to_ctypes()
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py != 3:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 3',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode_py != 3:
        _x_dim1 = _n_py
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        check_contents=_mode_py != 3,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_mode_py != 3)
    if _mode_py in (0, 2):
        if f is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'f'},
                'must not be None when mode in (0, 2)',
            )
    if _mode_py in (0, 2):
        _f_dim1 = _n_py
    else:
        _f_dim1 = 0
    exp_shape_f = (_f_dim1,)
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=exp_shape_f,
        check_contents=False,
        allow_none=True,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    if _mode_py in (1, 2):
        if f_deriv is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'f_deriv'},
                'must not be None when mode in (1, 2)',
            )
    if _mode_py in (1, 2):
        _f_deriv_dim1 = _n_py
    else:
        _f_deriv_dim1 = 0
    exp_shape_f_deriv = (_f_deriv_dim1,)
    _f_deriv_cls = utils._EngineFloat64ArrayType(
        f_deriv, {'fun_name': fun_name, 'entity': 'f_deriv'},
        exp_shape=exp_shape_f_deriv,
        check_contents=False,
        allow_none=True,
    )
    _f_deriv = _f_deriv_cls.to_ctypes(can_reference_contents=False)
    _a = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ordval,
        _q,
        _parity,
        _mode,
        _n,
        _x,
        _f,
        _f_deriv,
        _a,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ordval = {:s} and parity = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if parity = 0, ordval >= 0 or',
                    'if parity = 1, ordval >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, parity = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: parity = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if f is not None:
            _f_cls.refresh_py(_f)
        if f_deriv is not None:
            _f_deriv_cls.refresh_py(_f_deriv)
        _a_py = utils._EngineFloat64ScalarType.to_py(_a)
        return_data = _a_py
    _err_class.manage()
    return return_data

def opt_bsm_price(calput, m, n, x, s, t, sigma, r, q, p):
    u"""
    Black--Scholes--Merton option pricing formula.

    ``opt_bsm_price`` computes the European option price given by the
    Black--Scholes--Merton formula.

    For full information please refer to the NAG Library document for
    s30aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30aaf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_bsm_price`` computes the price of a European call (or put)
    option for constant volatility, sigma, and risk-free interest rate,
    r, with possible dividend yield, q, using the Black--Scholes--Merton
    formula (see Black and Scholes (1973) and Merton (1973)).
    For a given strike price, X, the price of a European call with
    underlying price, S, and time to expiry, T, is

        P_call = Se^(-qT)Phi(d_1)-Xe^(-rT)Phi(d_2)

    and the corresponding European put price is

        P_put = Xe^(-rT)Phi(-d_2)-Se^(-qT)Phi(-d_1)

    and where Phi denotes the cumulative Normal distribution function,

        Phi(x) = 1/(sqrt(2 pi))int_- infinity^xexp(-y^2/2)dy

    and

        d_1 = (ln(S/X)+(r-q+sigma^2/2)T)/(sigma*sqrt(T)), ; d_2 = d_1 -
        sigma*sqrt(T) .

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Black, F and Scholes, M, 1973, `The pricing of options and corporate
    liabilities`, Journal of Political Economy (81), 637--654

    Merton, R C, 1973, `Theory of rational option pricing`, Bell Journal
    of Economics and Management Science (4), 141--183

    See Also
    --------
    :meth:`naginterfaces.library.examples.roots.contfn_brent_interval_ex.main`
    :meth:`naginterfaces.library.examples.specfun.opt_bsm_price_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_bsm_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_bsm_greeks(calput, m, n, x, s, t, sigma, r, q, p, delta, gamma, vega, theta, rho, crho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Black--Scholes--Merton option pricing formula with Greeks.

    ``opt_bsm_greeks`` computes the European option price given by the
    Black--Scholes--Merton formula together with its sensitivities
    (Greeks).

    For full information please refer to the NAG Library document for
    s30ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30abf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    crho : float, shape (`m`, `n`), modified in place
        `On exit`: `crho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual cost of carry rate, i.e., -dP_{ij}/db, for i =
        1,2,...,`m` and j = 1,2,...,`n`, where b = r-q.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_bsm_greeks`` computes the price of a European call (or put)
    option together with the Greeks or sensitivities, which are the
    partial derivatives of the option price with respect to certain of
    the other input parameters, by the Black--Scholes--Merton formula
    (see Black and Scholes (1973) and Merton (1973)).
    The annual volatility, sigma, risk-free interest rate, r, and
    dividend yield, q, must be supplied as input.
    For a given strike price, X, the price of a European call with
    underlying price, S, and time to expiry, T, is

        P_call = Se^(-qT)Phi(d_1)-Xe^(-rT)Phi(d_2)

    and the corresponding European put price is

        P_put = Xe^(-rT)Phi(-d_2)-Se^(-qT)Phi(-d_1)

    and where Phi denotes the cumulative Normal distribution function,

        Phi(x) = 1/(sqrt(2 pi))int_- infinity^xexp(-y^2/2)dy

    and

        d_1 = (ln(S/X)+(r-q+sigma^2/2)T)/(sigma*sqrt(T)), ; d_2 = d_1 -
        sigma*sqrt(T) .

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Black, F and Scholes, M, 1973, `The pricing of options and corporate
    liabilities`, Journal of Political Economy (81), 637--654

    Merton, R C, 1973, `Theory of rational option pricing`, Bell Journal
    of Economics and Management Science (4), 141--183
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_bsm_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, crho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (crho, 'crho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = {:s}.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _crho_cls = utils._EngineFloat64ArrayType(
        crho, {'fun_name': fun_name, 'entity': 'crho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _crho = _crho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _crho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _crho_cls.refresh_py(_crho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data

def opt_imp_vol(calput, n, p, k, s0, t, r, sigma, mode, ivalid):
    u"""
    Black--Scholes--Merton implied volatility.

    ``opt_imp_vol`` computes the implied volatility of a European option
    contract based on the Black--Scholes--Merton formula.

    For full information please refer to the NAG Library document for
    s30ac

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30acf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    n : int
        n, the number of implied volatilities to be computed.

    p : float, shape (`n`)
        `p`[i-1] must contain P_i, the ith option price, for i =
        1,2,...,n.

    k : float, shape (`n`)
        `k`[i-1] must contain K_i, the ith strike price, for i =
        1,2,...,n.

    s0 : float, shape (`n`)
        `s0`[i-1] must contain S_{0i}, the ith spot price, for i =
        1,2,...,n.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to maturity,
        for i = 1,2,...,n.

    r : float, shape (`n`)
        `r`[i-1] must contain r_i, the ith interest rate, for i =
        1,2,...,n. Note that a rate of 5% should be entered as 0.05.

    sigma : float, shape (`n`), modified in place
        `On exit`: `sigma`[i-1] contains sigma_i, the ith implied
        volatility, for i = 1,2,...,n.

    mode : int
        Specifies which algorithm will be used to compute the implied
        volatilities. See [Accuracy] and [Parallelism and Performance]
        for further guidance on the choice of mode.

        `mode` = 0

            The Glau `et al.` (2018) algorithm will be used. The nodes
            used in the Chebyshev interpolation will be chosen to
            achieve relative accuracy to approximately seven decimal
            places;

        `mode` = 1

            The Glau `et al.` (2018) algorithm will be used. The nodes
            used in the Chebyshev interpolation will be chosen to
            achieve relative accuracy to approximately 15 decimal
            places, but limited by the machine precision;

        `mode` = 2

            The Jäckel (2015) algorithm will be used, aiming for
            accuracy to approximately 15--16 decimal places, but limited
            by machine precision.

    ivalid : int, shape (`n`), modified in place
        `On exit`: `ivalid`[i-1] indicates any errors with the input
        arguments that prevented sigma_i from being computed. If
        `ivalid`[i-1] != 0, `sigma`[i-1] contains 0.0.

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            P_i < 0.0.

        `ivalid`[i-1] = 2

            K_i <= 0.0.

        `ivalid`[i-1] = 3

            S_{0i} <= 0.0.

        `ivalid`[i-1] = 4

            T_i <= 0.0.

        `ivalid`[i-1] = 5

            The combination of P_i, K_i, S_{0i}, T_i and r_i is out of
            the domain in which sigma_i can be computed. See [Further
            Comments] for further details.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 4)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry at least one input argument was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The Black--Scholes formula for the price of a European option is

        P_call = S_0Phi((ln((S_0)/K)+[r+(sigma^2)/2]T)/sigma
        sqrt(T))-Ke^(-rT)Phi((ln((S_0)/K)+[r-(sigma^2)/2]T)/sigma
        sqrt(T)),

    for a call option, and

        P_put = Ke^(-rT)Phi((-ln((S_0)/K)-[r-(sigma^2)/2]T)/sigma
        sqrt(T))-S_0Phi((-ln((S_0)/K)-[r+(sigma^2)/2]T)/sigma sqrt(T)),

    for a put option, where Phi is the cumulative Normal distribution
    function, T is the time to maturity, S_0 is the spot price of the
    underlying asset, K is the strike price, r is the interest rate and
    sigma is the volatility.

    Given arrays of values P_i, K_i, S_{0i}, T_i and r_i, for i =
    1,2,...n, ``opt_imp_vol`` computes the implied volatilities sigma_i.

    ``opt_imp_vol`` offers the choice of two algorithms.
    The algorithm of Glau `et al.` (2018) uses Chebyshev interpolation
    to compute the implied volatilities, and performs best for long
    arrays of input data.
    The algorithm of Jäckel (2015) uses a third order Householder
    iteration and performs better for short arrays of input data.

    References
    ----------
    Glau, K, Herold, P, Madan, D B and Pötz, C, 2018, `The Chebyshev
    method for the implied volatility`, Accepted for publication in the
    Journal of Computational Finance

    Jäckel, P, 2015, `Let's be Rational`, Wilmott Magazine (2015(75)),
    40--53
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_imp_vol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, p[n].
        utils._EngineFloat64ArrayType.c_type, # Input, k[n].
        utils._EngineFloat64ArrayType.c_type, # Input, s0[n].
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._EngineFloat64ArrayType.c_type, # Input, r[n].
        utils._EngineFloat64ArrayType.c_type, # Output, sigma[n].
        utils._BYREF_INT, # Input, mode.
        utils._EngineIntArrayType.c_type, # Output, ivalid[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(4,),
        )
        _err_class.manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _k = utils._EngineFloat64ArrayType(
        k, {'fun_name': fun_name, 'entity': 'k'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _s0 = utils._EngineFloat64ArrayType(
        s0, {'fun_name': fun_name, 'entity': 's0'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma_cls = utils._EngineFloat64ArrayType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _sigma = _sigma_cls.to_ctypes(can_reference_contents=False)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _ivalid_cls = utils._EngineIntArrayType(
        ivalid, {'fun_name': fun_name, 'entity': 'ivalid'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _ivalid = _ivalid_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _n,
        _p,
        _k,
        _s0,
        _t,
        _r,
        _sigma,
        _mode,
        _ivalid,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 9999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry at least one input argument was invalid.',
                    'Check ivalid for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 9998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 9997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _sigma_cls.refresh_py(_sigma)
        _ivalid_cls.refresh_py(_ivalid)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def opt_lookback_fls_price(calput, m, n, sm, s, t, sigma, r, q, p):
    u"""
    Floating-strike lookback option pricing formula in the
    Black-Scholes-Merton model.

    ``opt_lookback_fls_price`` computes the price of a floating-strike
    lookback option.

    For full information please refer to the NAG Library document for
    s30ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30baf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of minimum or maximum prices to be used.

    n : int
        The number of times to expiry to be used.

    sm : float, shape (`m`)
        `sm`[i-1] must contain S_min(i), the ith minimum observed price
        of the underlying asset when `calput` = 'C', or S_max(i), the
        maximum observed price when `calput` = 'P', for i = 1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the minimum or maximum observed price S_min(i) or
        S_max(i) at expiry `t`_j for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry with a put option, `sm`[*<value>*] = *<value>*.

            Constraint: for put options, `sm`[i] >= *<value>* for all i.

        (`errno` 4)
            On entry with a call option, `sm`[*<value>*] = *<value>*.

            Constraint: for call options, `sm`[i] <= *<value>* for all
            i.

        (`errno` 4)
            On entry, `sm`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `sm`[i] <= *<value>* for all i.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>* for all i.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_lookback_fls_price`` computes the price of a floating-strike
    lookback call or put option.
    A call option of this type confers the right to buy the underlying
    asset at the lowest price, S_min, observed during the lifetime of
    the contract.
    A put option gives the holder the right to sell the underlying asset
    at the maximum price, S_max, observed during the lifetime of the
    contract.
    Thus, at expiry, the payoff for a call option is S-S_min, and for a
    put, S_max-S.

    For a given minimum value the price of a floating-strike lookback
    call with underlying asset price, S, and time to expiry, T, is

        P_call = Se^(-qT)Phi(a_1)-S_min e^(-rT)Phi(a_2)+Se^(-rT)
        (sigma^2)/2b[(S/(S_min))^(-2b/sigma^2)Phi(-a_1+2b/sigmasqrt(T))-e^(bT)Phi(-a_1)],

    where b = r-q != 0.
    The volatility, sigma, risk-free interest rate, r, and annualised
    dividend yield, q, are constants.
    When r = q, the option price is given by

        P_call = Se^(-qT)Phi(a_1)-S_min
        e^(-rT)Phi(a_2)+Se^(-rT)sigma*sqrt(T)[phi(a_1)+a_1(Phi(a_1)-1)].

    The corresponding put price is (for b != 0),

        P_put = S_max e^(-rT)Phi(-a_2)-Se^(-qT)Phi(-a_1)+Se^(-rT)
        (sigma^2)/2b[-(S/(S_max))^(-2b/sigma^2)Phi(a_1-2b/sigmasqrt(T))+e^(bT)Phi(a_1)].

    When r = q,

        P_put = S_max
        e^(-rT)Phi(-a_2)-Se^(-qT)Phi(-a_1)+Se^(-rT)sigma*sqrt(T)[phi(a_1)+a_1Phi(a_1)].

    In the above, Phi denotes the cumulative Normal distribution
    function,

        Phi(x) = int_- infinity^x phi(y)dy

    where phi denotes the standard Normal probability density function

        phi(y) = 1/(sqrt(2 pi))exp(-y^2/2)

    and

        a_1 = (ln(S/S_m)+(b+sigma^2/2)T)/(sigma*sqrt(T)); a_2 =
        a_1-sigma*sqrt(T)

    where S_m is taken to be the minimum price attained by the
    underlying asset, S_min, for a call and the maximum price, S_max,
    for a put.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    minimum or maximum observed price in a set S_min(i) or S_max(i), i =
    1,2,...,m, and for each expiry time in a set T_j, j = 1,2,...,n.

    References
    ----------
    Goldman, B M, Sosin, H B and Gatto, M A, 1979, `Path dependent
    options: buy at the low, sell at the high`, Journal of Finance (34),
    1111--1127
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_lookback_fls_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, sm[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _sm = utils._EngineFloat64ArrayType(
        sm, {'fun_name': fun_name, 'entity': 'sm'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _sm,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry with a put option,',
                    'sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: for put options, sm[i] >= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 114:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry with a call option,',
                    'sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: for call options, sm[i] <= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 204:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= sm[i] <= {:s} for'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: t[i] >= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_lookback_fls_greeks(calput, m, n, sm, s, t, sigma, r, q, p, delta, gamma, vega, theta, rho, crho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Floating-strike lookback option pricing formula with Greeks in the
    Black-Scholes-Merton model.

    ``opt_lookback_fls_greeks`` computes the price of a floating-strike
    lookback option together with its sensitivities (Greeks).

    For full information please refer to the NAG Library document for
    s30bb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30bbf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of minimum or maximum prices to be used.

    n : int
        The number of times to expiry to be used.

    sm : float, shape (`m`)
        `sm`[i-1] must contain S_min(i), the ith minimum observed price
        of the underlying asset when `calput` = 'C', or S_max(i), the
        maximum observed price when `calput` = 'P', for i = 1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        The annual risk-free interest rate, r, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        The annual continuous yield rate. Note that a rate of 8% should
        be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the minimum or maximum observed price S_min(i) or
        S_max(i) at expiry `t`_j for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    crho : float, shape (`m`, `n`), modified in place
        `On exit`: `crho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual cost of carry rate, i.e., -dP_{ij}/db, for i =
        1,2,...,`m` and j = 1,2,...,`n`, where b = r-q.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry with a put option, `sm`[*<value>*] = *<value>*.

            Constraint: for put options, `sm`[i] >= *<value>* for all i.

        (`errno` 4)
            On entry with a call option, `sm`[*<value>*] = *<value>*.

            Constraint: for call options, `sm`[i] <= *<value>* for all
            i.

        (`errno` 4)
            On entry, `sm`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `sm`[i] <= *<value>* for all i.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>* for all i.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

        (`errno` 12)
            On entry, `r` = *<value>* and `q` = *<value>*.

            Constraint: abs(`r`-`q`) > 10*eps*max(abs(`r`),1), where eps
            is the machine precision.

    Notes
    -----
    ``opt_lookback_fls_greeks`` computes the price of a floating-strike
    lookback call or put option, together with the Greeks or
    sensitivities, which are the partial derivatives of the option price
    with respect to certain of the other input parameters.
    A call option of this type confers the right to buy the underlying
    asset at the lowest price, S_min, observed during the lifetime of
    the contract.
    A put option gives the holder the right to sell the underlying asset
    at the maximum price, S_max, observed during the lifetime of the
    contract.
    Thus, at expiry, the payoff for a call option is S-S_min, and for a
    put, S_max-S.

    For a given minimum value the price of a floating-strike lookback
    call with underlying asset price, S, and time to expiry, T, is

        P_call = Se^(-qT)Phi(a_1)-S_min e^(-rT)Phi(a_2)+Se^(-rT)
        (sigma^2)/2b[(S/(S_min))^(-2b/sigma^2)Phi(-a_1+2b/sigmasqrt(T))-e^(bT)Phi(-a_1)],

    where b = r-q != 0.
    The volatility, sigma, risk-free interest rate, r, and annualised
    dividend yield, q, are constants.

    The corresponding put price is

        P_put = S_max e^(-rT)Phi(-a_2)-Se^(-qT)Phi(-a_1)+Se^(-rT)
        (sigma^2)/2b[-(S/(S_max))^(-2b/sigma^2)Phi(a_1-2b/sigmasqrt(T))+e^(bT)Phi(a_1)].

    In the above, Phi denotes the cumulative Normal distribution
    function,

        Phi(x) = int_- infinity^x phi(y)dy

    where phi denotes the standard Normal probability density function

        phi(y) = 1/(sqrt(2 pi))exp(-y^2/2)

    and

        a_1 = (ln(S/S_m)+(b+sigma^2/2)T)/(sigma*sqrt(T)); a_2 =
        a_1-sigma*sqrt(T)

    where S_m is taken to be the minimum price attained by the
    underlying asset, S_min, for a call and the maximum price, S_max,
    for a put.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    minimum or maximum observed price in a set S_min(i) or S_max(i), i =
    1,2,...,m, and for each expiry time in a set T_j, j = 1,2,...,n.

    References
    ----------
    Goldman, B M, Sosin, H B and Gatto, M A, 1979, `Path dependent
    options: buy at the low, sell at the high`, Journal of Finance (34),
    1111--1127
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_lookback_fls_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, sm[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, crho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (crho, 'crho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _sm = utils._EngineFloat64ArrayType(
        sm, {'fun_name': fun_name, 'entity': 'sm'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _crho_cls = utils._EngineFloat64ArrayType(
        crho, {'fun_name': fun_name, 'entity': 'crho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _crho = _crho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _sm,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _crho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry with a put option,',
                    'sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: for put options, sm[i] >= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 114:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry with a call option,',
                    'sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: for call options, sm[i] <= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 204:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sm[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= sm[i] <= {:s} for'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: t[i] >= {:s} for all i.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s} and q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(r-q) > 10*eps*max(abs(r),1),',
                    'where eps is the machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _crho_cls.refresh_py(_crho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data

def opt_binary_con_price(calput, m, n, x, s, k, t, sigma, r, q, p):
    u"""
    Binary option, cash-or-nothing pricing formula.

    ``opt_binary_con_price`` computes the price of a binary or digital
    cash-or-nothing option.

    For full information please refer to the NAG Library document for
    s30ca

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30caf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    k : float
        The amount, K, to be paid at expiration if the option is
        in-the-money, i.e., if `s` > `x`[i-1] when `calput` = 'C', or if
        `s` < `x`[i-1] when `calput` = 'P', for i = 1,2,...,m.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.0.

        (`errno` 7)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 8)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 9)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 10)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_binary_con_price`` computes the price of a binary or digital
    cash-or-nothing option which pays a fixed amount, K, at expiration
    if the option is in-the-money (see [the S Introduction]).
    For a strike price, X, underlying asset price, S, and time to
    expiry, T, the payoff is, therefore, K, if S > X for a call or S < X
    for a put.
    Nothing is paid out when this condition is not met.

    The price of a call with volatility, sigma, risk-free interest rate,
    r, and annualised dividend yield, q, is

        P_call = Ke^(-rT)Phi(d_2)

    and for a put,

        P_put = Ke^(-rT)Phi(-d_2)

    where Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^x(-y^2/2)dy,

    and

        d_2 = (ln(S/X)+(r-q-sigma^2/2)T)/(sigma*sqrt(T)).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Reiner, E and Rubinstein, M, 1991, `Unscrambling the binary code`,
    Risk (4)
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_binary_con_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _k = utils._EngineFloat64ScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _k,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_binary_con_greeks(calput, m, n, x, s, k, t, sigma, r, q, p, delta, gamma, vega, theta, rho, crho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Binary option, cash-or-nothing pricing formula with Greeks.

    ``opt_binary_con_greeks`` computes the price of a binary or digital
    cash-or-nothing option together with its sensitivities (Greeks).

    For full information please refer to the NAG Library document for
    s30cb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30cbf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    k : float
        The amount, K, to be paid at expiration if the option is
        in-the-money, i.e., if `s` > `x`[i-1] when `calput` = 'C', or if
        `s` < `x`[i-1] when `calput` = 'P', for i = 1,2,...,m.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    crho : float, shape (`m`, `n`), modified in place
        `On exit`: `crho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual cost of carry rate, i.e., -dP_{ij}/db, for i =
        1,2,...,`m` and j = 1,2,...,`n`, where b = r-q.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.0.

        (`errno` 7)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 8)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 9)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 10)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_binary_con_greeks`` computes the price of a binary or digital
    cash-or-nothing option, together with the Greeks or sensitivities,
    which are the partial derivatives of the option price with respect
    to certain of the other input parameters.
    This option pays a fixed amount, K, at expiration if the option is
    in-the-money (see [the S Introduction]).
    For a strike price, X, underlying asset price, S, and time to
    expiry, T, the payoff is, therefore, K, if S > X for a call or S < X
    for a put.
    Nothing is paid out when this condition is not met.

    The price of a call with volatility, sigma, risk-free interest rate,
    r, and annualised dividend yield, q, is

        P_call = Ke^(-rT)Phi(d_2)

    and for a put,

        P_put = Ke^(-rT)Phi(-d_2)

    where Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^xexp(-y^2/2)dy,

    and

        d_2 = (ln(S/X)+(r-q-sigma^2/2)T)/(sigma*sqrt(T)).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Reiner, E and Rubinstein, M, 1991, `Unscrambling the binary code`,
    Risk (4)
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_binary_con_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, crho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (crho, 'crho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _k = utils._EngineFloat64ScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _crho_cls = utils._EngineFloat64ArrayType(
        crho, {'fun_name': fun_name, 'entity': 'crho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _crho = _crho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _k,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _crho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _crho_cls.refresh_py(_crho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data

def opt_binary_aon_price(calput, m, n, x, s, t, sigma, r, q, p):
    u"""
    Binary option, asset-or-nothing pricing formula.

    ``opt_binary_aon_price`` computes the price of a binary or digital
    asset-or-nothing option.

    For full information please refer to the NAG Library document for
    s30cc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30ccf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_binary_aon_price`` computes the price of a binary or digital
    asset-or-nothing option which pays the underlying asset itself, S,
    at expiration if the option is in-the-money (see [the S
    Introduction]).
    For a strike price, X, underlying asset price, S, and time to
    expiry, T, the payoff is, therefore, S, if S > X for a call or S < X
    for a put.
    Nothing is paid out when this condition is not met.

    The price of a call with volatility, sigma, risk-free interest rate,
    r, and annualised dividend yield, q, is

        P_call = Se^(-qT)Phi(d_1)

    and for a put,

        P_put = Se^(-qT)Phi(-d_1)

    where Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^xexp(-y^2/2)dy,

    and

        d_1 = (ln(S/X)+(r-q+sigma^2/2)T)/(sigma*sqrt(T)).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Reiner, E and Rubinstein, M, 1991, `Unscrambling the binary code`,
    Risk (4)
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_binary_aon_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_binary_aon_greeks(calput, m, n, x, s, t, sigma, r, q, p, delta, gamma, vega, theta, rho, crho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Binary option, asset-or-nothing pricing formula with Greeks.

    ``opt_binary_aon_greeks`` computes the price of a binary or digital
    asset-or-nothing option together with its sensitivities (Greeks).

    For full information please refer to the NAG Library document for
    s30cd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30cdf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    crho : float, shape (`m`, `n`), modified in place
        `On exit`: `crho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual cost of carry rate, i.e., -dP_{ij}/db, for i =
        1,2,...,`m` and j = 1,2,...,`n`, where b = r-q.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Notes
    -----
    ``opt_binary_aon_greeks`` computes the price of a binary or digital
    asset-or-nothing option, together with the Greeks or sensitivities,
    which are the partial derivatives of the option price with respect
    to certain of the other input parameters.
    This option pays the underlying asset itself, S, at expiration if
    the option is in-the-money (see [the S Introduction]).
    For a strike price, X, underlying asset price, S, and time to
    expiry, T, the payoff is, therefore, S, if S > X for a call or S < X
    for a put.
    Nothing is paid out when this condition is not met.

    The price of a call with volatility, sigma, risk-free interest rate,
    r, and annualised dividend yield, q, is

        P_call = Se^(-qT)Phi(d_1)

    and for a put,

        P_put = Se^(-qT)Phi(-d_1)

    where Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^xexp(-y^2/2)dy,

    and

        d_1 = (ln(S/X)+(r-q+sigma^2/2)T)/(sigma*sqrt(T)).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Reiner, E and Rubinstein, M, 1991, `Unscrambling the binary code`,
    Risk (4)
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_binary_aon_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, crho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (crho, 'crho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _crho_cls = utils._EngineFloat64ArrayType(
        crho, {'fun_name': fun_name, 'entity': 'crho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _crho = _crho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _crho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _crho_cls.refresh_py(_crho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data

def opt_barrier_std_price(calput, btype, m, n, x, s, h, k, t, sigma, r, q, p):
    u"""
    Standard barrier option pricing formula.

    ``opt_barrier_std_price`` computes the price of a standard barrier
    option.

    For full information please refer to the NAG Library document for
    s30fa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30faf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    btype : str, length 2
        Indicates the barrier type as **In** or **Out** and its relation
        to the price of the underlying asset as **Up** or **Down**.

        `btype` = 'DI'

            Down-and-In.

        `btype` = 'DO'

            Down-and-Out.

        `btype` = 'UI'

            Up-and-In.

        `btype` = 'UO'

            Up-and-Out.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    h : float
        The barrier price.

    k : float
        The value of a possible cash rebate to be paid if the option has
        not been knocked in (or out) before expiration.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `btype` = *<value>*.

            Constraint: `btype` = 'DI', 'DO', 'UI' or 'UO'.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 4)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 5)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 6)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 7)
            On entry, `h` = *<value>*.

            Constraint: `h` >= *<value>* and `h` <= *<value>*.

        (`errno` 8)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.0.

        (`errno` 9)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 10)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 11)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 12)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

        (`errno` 15)
            On entry, `s` and `h` are inconsistent with `btype`: `s` =
            *<value>* and `h` = *<value>*.

    Notes
    -----
    ``opt_barrier_std_price`` computes the price of a standard barrier
    option, where the exercise, for a given strike price, X, depends on
    the underlying asset price, S, reaching or crossing a specified
    barrier level, H.
    Barrier options of type **In** only become active (are knocked in)
    if the underlying asset price attains the pre-determined barrier
    level during the lifetime of the contract.
    Those of type **Out** start active and are knocked out if the
    underlying asset price attains the barrier level during the lifetime
    of the contract.
    A cash rebate, K, may be paid if the option is inactive at
    expiration.
    The option may also be described as **Up** (the underlying price
    starts below the barrier level) or **Down** (the underlying price
    starts above the barrier level).
    This gives the following options which can be specified as put or
    call contracts.

    **Down-and-In:** the option starts inactive with the underlying
    asset price above the barrier level.
    It is knocked in if the underlying price moves down to hit the
    barrier level before expiration.

    **Down-and-Out:** the option starts active with the underlying asset
    price above the barrier level.
    It is knocked out if the underlying price moves down to hit the
    barrier level before expiration.

    **Up-and-In:** the option starts inactive with the underlying asset
    price below the barrier level.
    It is knocked in if the underlying price moves up to hit the barrier
    level before expiration.

    **Up-and-Out:** the option starts active with the underlying asset
    price below the barrier level.
    It is knocked out if the underlying price moves up to hit the
    barrier level before expiration.

    The payoff is max(S-X,0) for a call or max(X-S,0) for a put, if the
    option is active at expiration, otherwise it may pay a pre-specified
    cash rebate, K.
    Following Haug (2007), the prices of the various standard barrier
    options can be written as shown below.
    The volatility, sigma, risk-free interest rate, r, and annualised
    dividend yield, q, are constants.
    The integer parameters, j and k, take the values +/-1, depending on
    the type of barrier.

        A = j S e^(-qT) Phi (jx_1) - j X e^(-rT) Phi
        (j[x_1-sigma*sqrt(T)]); B = j S e^(-qT) Phi (jx_2) - j X e^(-rT)
        Phi (j[x_2-sigma*sqrt(T)]); C = j S e^(-qT) (H/S)^(2*(mu+1)) Phi
        (ky_1) - j X e^(-rT) (H/S)^(2*mu) Phi (k[y_1-sigma*sqrt(T)]); D
        = j S e^(-qT) (H/S)^(2*(mu+1)) Phi (ky_2) - j X e^(-rT)
        (H/S)^(2*mu) Phi (k[y_2-sigma*sqrt(T)]); E = K e^(-rT)
        {Phi(k[x_2-sigma*sqrt(T)])-(H/S)^(2
        mu)Phi(k[y_2-sigma*sqrt(T)])} ; F = K
        {(H/S)^(mu+lambda)Phi(kz)+(H/S)^(mu-lambda)Phi(k[z-sigma*sqrt(T)])}

    with

        x_1 = ln(S/X)/(sigma*sqrt(T)) + (1+mu) sigma*sqrt(T); x_2 =
        ln(S/H)/(sigma*sqrt(T)) + (1+mu) sigma*sqrt(T); y_1 =
        ln(H^2/(SX))/(sigma*sqrt(T)) + (1+mu)sigma*sqrt(T); y_2 =
        ln(H/S)/(sigma*sqrt(T)) + (1+mu)sigma*sqrt(T); z =
        ln(H/S)/(sigma*sqrt(T)) + lambdasigma*sqrt(T); mu =
        (r-q-sigma^2/2)/(sigma^2); lambda = sqrt(mu^2+2r/(sigma^2))

    and where Phi denotes the cumulative Normal distribution function,

        Phi(x) = 1/(sqrt(2 pi))int_- infinity^xexp(-y^2/2)dy.

    **Down-and-In (S > H)**:

        When X >= H, with j = k = 1,

                P_call = C+E

        and with j = -1, k = 1

                P_put = B-C+D+E

        When X < H, with j = k = 1

            P_call = A-B+D+E

        and with j = -1, k = 1

            P_put = A+E.

    **Down-and-Out (S > H)**:

        When X >= H, with j = k = 1,

                P_call = A-C+F

        and with j = -1, k = 1

                P_put = A-B+C-D+F

        When X < H, with j = k = 1,

            P_call = B-D+F

        and with j = -1, k = 1

            P_put = F.

    **Up-and-In (S < H)**:

        When X >= H, with j = 1, k = -1,

                P_call = A+E

        and with j = k = -1,

                P_put = A-B+D+E

        When X < H, with j = 1, k = -1,

            P_call = B-C+D+E

        and with j = k = -1,

            P_put = C+E.

    **Up-and-Out (S < H):**

        When X >= H, with j = 1, k = -1,

                P_call = F

        and with j = k = -1,

                P_put = B-D+F

        When X < H, with j = 1, k = -1,

            P_call = A-B+C-D+F

        and with j = k = -1,

            P_put = A-C+F.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Haug, E G, 2007, `The Complete Guide to Option Pricing Formulas`,
    (2nd Edition), McGraw-Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_barrier_std_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30faft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, itype.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, h.
        utils._BYREF_FLOAT64, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _btype = utils._EngineCharScalarType(
        btype, {'fun_name': fun_name, 'entity': 'btype'}, exp_chrlen=2,
    ).to_ctypes()
    _btype_py = utils._EngineCharScalarType.to_py(_btype)
    if _btype_py.upper() == 'DI':
        _itype = utils._EngineIntScalarType.c_type(1)
    elif _btype_py.upper() == 'DO':
        _itype = utils._EngineIntScalarType.c_type(2)
    elif _btype_py.upper() == 'UI':
        _itype = utils._EngineIntScalarType.c_type(3)
    elif _btype_py.upper() == 'UO':
        _itype = utils._EngineIntScalarType.c_type(4)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, btype = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_btype_py),
                ),
                'Constraint: btype = \'DI\', \'DO\', \'UI\' or \'UO\'.',
            ],
            errcodes=(2,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _h = utils._EngineFloat64ScalarType(
        h, {'fun_name': fun_name, 'entity': 'h'},
    ).to_ctypes()
    _k = utils._EngineFloat64ScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _itype,
        _m,
        _n,
        _x,
        _s,
        _h,
        _k,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: h >= {:s} and h <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 15 and _ifmt == 15:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s and h are inconsistent with btype:',
                    's = {:s} and h = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_jumpdiff_merton_price(calput, m, n, x, s, t, sigma, r, lamda, jvol, p):
    u"""
    Jump-diffusion, Merton's model, option pricing formula.

    ``opt_jumpdiff_merton_price`` computes the European option price
    using the Merton jump-diffusion model.

    For full information please refer to the NAG Library document for
    s30ja

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30jaf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the annual total volatility, including jumps.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    lamda : float
        lambda, the number of expected jumps per year.

    jvol : float
        The proportion of the total volatility associated with jumps.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `lamda` = *<value>*.

            Constraint: `lamda` > 0.0.

        (`errno` 10)
            On entry, `jvol` = *<value>*.

            Constraint: `jvol` >= 0.0 and `jvol` < 1.0.

    Notes
    -----
    ``opt_jumpdiff_merton_price`` uses Merton's jump-diffusion model
    (Merton (1976)) to compute the price of a European option.
    This assumes that the asset price is described by a Brownian motion
    with drift, as in the Black--Scholes--Merton case, together with a
    compound Poisson process to model the jumps.
    The corresponding stochastic differential equation is,

        dS/S = (alpha-lambda k)dt+sigma^dW_t+dq_t.

    Here alpha is the instantaneous expected return on the asset price,
    S; sigma^^2 is the instantaneous variance of the return when the
    Poisson event does not occur; dW_t is a standard Brownian motion;
    q_t is the independent Poisson process and k = E[Y-1] where Y-1 is
    the random variable change in the stock price if the Poisson event
    occurs and E is the expectation operator over the random variable Y.

    This leads to the following price for a European option (see Haug
    (2007))

        P_call = sum_j = 0^infinity e^(-lambda T)(lambda
        T)^j/j!C_j(S,X,T,r,sigma'_j),

    where T is the time to expiry; X is the strike price; r is the
    annual risk-free interest rate; C_j(S,X,T,r,sigma'_j) is the
    Black--Scholes--Merton option pricing formula for a European call
    (see ``opt_bsm_price``).

        sigma'_j = sqrt(z^2+delta^2(j/T)), ; z^2 = sigma^2 - lambda
        delta^2, ; delta^2 = gammasigma^2/lambda,

    where sigma is the total volatility including jumps; lambda is the
    expected number of jumps given as an average per year; gamma is the
    proportion of the total volatility due to jumps.

    The value of a put is obtained by substituting the
    Black--Scholes--Merton put price for C_j(S,X,T,r,sigma'_j).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Haug, E G, 2007, `The Complete Guide to Option Pricing Formulas`,
    (2nd Edition), McGraw-Hill

    Merton, R C, 1976, `Option pricing when underlying stock returns are
    discontinuous`, Journal of Financial Economics (3), 125--144
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_jumpdiff_merton_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30jaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, lamda.
        utils._BYREF_FLOAT64, # Input, jvol.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _lamda = utils._EngineFloat64ScalarType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
    ).to_ctypes()
    _jvol = utils._EngineFloat64ScalarType(
        jvol, {'fun_name': fun_name, 'entity': 'jvol'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _lamda,
        _jvol,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lamda > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, jvol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: jvol >= 0.0 and jvol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_jumpdiff_merton_greeks(calput, m, n, x, s, t, sigma, r, lamda, jvol, p, delta, gamma, vega, theta, rho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Jump-diffusion, Merton's model, option pricing formula with Greeks.

    ``opt_jumpdiff_merton_greeks`` computes the European option price
    together with its sensitivities (Greeks) using the Merton
    jump-diffusion model.

    For full information please refer to the NAG Library document for
    s30jb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30jbf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the annual total volatility, including jumps.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    lamda : float
        lambda, the number of expected jumps per year.

    jvol : float
        The proportion of the total volatility associated with jumps.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `lamda` = *<value>*.

            Constraint: `lamda` > 0.0.

        (`errno` 10)
            On entry, `jvol` = *<value>*.

            Constraint: `jvol` >= 0.0 and `jvol` < 1.0.

    Notes
    -----
    ``opt_jumpdiff_merton_greeks`` uses Merton's jump-diffusion model
    (Merton (1976)) to compute the price of a European option, together
    with the Greeks or sensitivities, which are the partial derivatives
    of the option price with respect to certain of the other input
    parameters.
    Merton's model assumes that the asset price is described by a
    Brownian motion with drift, as in the Black--Scholes--Merton case,
    together with a compound Poisson process to model the jumps.
    The corresponding stochastic differential equation is,

        dS/S = (alpha-lambda k)dt+sigma^dW_t+dq_t.

    Here alpha is the instantaneous expected return on the asset price,
    S; sigma^^2 is the instantaneous variance of the return when the
    Poisson event does not occur; dW_t is a standard Brownian motion;
    q_t is the independent Poisson process and k = E[Y-1] where Y-1 is
    the random variable change in the stock price if the Poisson event
    occurs and E is the expectation operator over the random variable Y.

    This leads to the following price for a European option (see Haug
    (2007))

        P_call = sum_j = 0^infinity e^(-lambda T)(lambda
        T)^j/j!C_j(S,X,T,r,sigma'_j),

    where T is the time to expiry; X is the strike price; r is the
    annual risk-free interest rate; C_j(S,X,T,r,sigma'_j) is the
    Black--Scholes--Merton option pricing formula for a European call
    (see ``opt_bsm_price``).

        sigma'_j = sqrt(z^2+delta^2(j/T)), ; z^2 = sigma^2 - lambda
        delta^2, ; delta^2 = gammasigma^2/lambda,

    where sigma is the total volatility including jumps; lambda is the
    expected number of jumps given as an average per year; gamma is the
    proportion of the total volatility due to jumps.

    The value of a put is obtained by substituting the
    Black--Scholes--Merton put price for C_j(S,X,T,r,sigma'_j).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Haug, E G, 2007, `The Complete Guide to Option Pricing Formulas`,
    (2nd Edition), McGraw-Hill

    Merton, R C, 1976, `Option pricing when underlying stock returns are
    discontinuous`, Journal of Financial Economics (3), 125--144
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_jumpdiff_merton_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30jbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, lamda.
        utils._BYREF_FLOAT64, # Input, jvol.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _lamda = utils._EngineFloat64ScalarType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
    ).to_ctypes()
    _jvol = utils._EngineFloat64ScalarType(
        jvol, {'fun_name': fun_name, 'entity': 'jvol'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _lamda,
        _jvol,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lamda > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, jvol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: jvol >= 0.0 and jvol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data

def opt_heston_price(calput, m, n, x, s, t, sigmav, kappa, corr, var0, eta, grisk, r, q, p):
    u"""
    Heston's model option pricing formula.

    ``opt_heston_price`` computes the European option price given by
    Heston's stochastic volatility model.

    For full information please refer to the NAG Library document for
    s30na

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30naf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigmav : float
        The volatility, sigma_v, of the volatility process, sqrt(v_t).
        Note that a rate of 20% should be entered as 0.2.

    kappa : float
        kappa, the long term mean reversion rate of the volatility.

    corr : float
        The correlation between the two standard Brownian motions for
        the asset price and the volatility.

    var0 : float
        The initial value of the variance, v_t, of the asset price.

    eta : float
        eta, the long term mean of the variance of the asset price.

    grisk : float
        The risk aversion parameter, gamma, of the representative agent.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i-1] >= *<value>* and `x`[i-1] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i-1] >= *<value>*.

        (`errno` 7)
            On entry, `sigmav` = *<value>*.

            Constraint: `sigmav` > 0.0.

        (`errno` 8)
            On entry, `kappa` = *<value>*.

            Constraint: `kappa` > 0.0.

        (`errno` 9)
            On entry, `corr` = *<value>*.

            Constraint: abs(`corr`) <= 1.0.

        (`errno` 10)
            On entry, `var0` = *<value>*.

            Constraint: `var0` >= 0.0.

        (`errno` 11)
            On entry, `eta` = *<value>*.

            Constraint: `eta` > 0.0.

        (`errno` 12)
            On entry, `grisk` = *<value>*, `sigmav` = *<value>* and
            `kappa` = *<value>*.

            Constraint: 0.0 <= `grisk` <= 1.0 and
            `grisk`*(1.0-`grisk`)*`sigmav`^2 <= `kappa`^2.

        (`errno` 13)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 14)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 17)
            Quadrature has not converged to the specified accuracy.
            However, the result should be a reasonable approximation.

        (`errno` 18)
            Solution cannot be computed accurately. Check values of
            input arguments.

    Notes
    -----
    ``opt_heston_price`` computes the price of a European option using
    Heston's stochastic volatility model.
    The return on the asset price, S, is

        dS/S = (r-q)dt+sqrt(v_t)dW_t^(1)

    and the instantaneous variance, v_t, is defined by a mean-reverting
    square root stochastic process,

        dv_t = kappa(eta-v_t)dt+sigma_v sqrt(v_t)dW_t^(2),

    where r is the risk free annual interest rate; q is the annual
    dividend rate; v_t is the variance of the asset price; sigma_v is
    the volatility of the volatility, sqrt(v_t); kappa is the mean
    reversion rate; eta is the long term variance. dW_t^(i), for i =
    1,2,...,2, denotes two correlated standard Brownian motions with

        ℂov[dW_t^(1),dW_t^(2)] = rho dt.

    The option price is computed by evaluating the integral transform
    given by Lewis (2000) using the form of the characteristic function
    discussed by Albrecher `et al.` (2007), see also Kilin (2006).

        P_call = Se^(-qT)-Xe^(-rT)1/piRe[int_0+i/2^infinity
        +i/2e^(-ikX_)H^(k,v,T)/(k^2-ik)dk],

    where X_= ln(S/X)+(r-q)T and

        H^(k,v,T) = exp(2 kappaeta/(sigma_v^2)[tg -ln((1-he^(-xi
        t))/(1-h))]+v_t g[(1-e^(-xi t))/(1-he^(-xi t))]),

        g = 1/2(b-xi), h = (b-xi)/(b+xi), t = sigma_v^2T/2,

        xi = [b^2+4(k^2-ik)/sigma_v^2]^(1/2),

        b =
        2/sigma_v^2[(1-gamma+ik)rhosigma_v+sqrt(kappa^2-gamma(1-gamma)sigma_v^2)]

    with t = sigma_v^2T/2.
    Here gamma is the risk aversion parameter of the representative
    agent with 0 <= gamma <= 1 and gamma(1-gamma)sigma_v^2 <= kappa^2.
    The value gamma = 1 corresponds to lambda = 0, where lambda is the
    market price of risk in Heston (1993) (see Lewis (2000) and Rouah
    and Vainberg (2007)).

    The price of a put option is obtained by put-call parity.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Albrecher, H, Mayer, P, Schoutens, W and Tistaert, J, 2007, `The
    little Heston trap`, Wilmott Magazine (January 2007), 83--92

    Heston, S, 1993, `A closed-form solution for options with stochastic
    volatility with applications to bond and currency options`, Review
    of Financial Studies (6), 327--343

    Kilin, F, 2006, `Accelerating the calibration of stochastic
    volatility models`, MPRA Paper No. 2975,
    https://mpra.ub.uni-muenchen.de/2975/

    Lewis, A L, 2000, `Option valuation under stochastic volatility`,
    Finance Press, USA

    Rouah, F D and Vainberg, G, 2007, `Option Pricing Models and
    Volatility using Excel-VBA`, John Wiley and Sons, Inc
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_heston_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30naft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigmav.
        utils._BYREF_FLOAT64, # Input, kappa.
        utils._BYREF_FLOAT64, # Input, corr.
        utils._BYREF_FLOAT64, # Input, var0.
        utils._BYREF_FLOAT64, # Input, eta.
        utils._BYREF_FLOAT64, # Input, grisk.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigmav = utils._EngineFloat64ScalarType(
        sigmav, {'fun_name': fun_name, 'entity': 'sigmav'},
    ).to_ctypes()
    _kappa = utils._EngineFloat64ScalarType(
        kappa, {'fun_name': fun_name, 'entity': 'kappa'},
    ).to_ctypes()
    _corr = utils._EngineFloat64ScalarType(
        corr, {'fun_name': fun_name, 'entity': 'corr'},
    ).to_ctypes()
    _var0 = utils._EngineFloat64ScalarType(
        var0, {'fun_name': fun_name, 'entity': 'var0'},
    ).to_ctypes()
    _eta = utils._EngineFloat64ScalarType(
        eta, {'fun_name': fun_name, 'entity': 'eta'},
    ).to_ctypes()
    _grisk = utils._EngineFloat64ScalarType(
        grisk, {'fun_name': fun_name, 'entity': 'grisk'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigmav,
        _kappa,
        _corr,
        _var0,
        _eta,
        _grisk,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i-1] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigmav = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigmav > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kappa > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(corr) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var0 >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eta > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, grisk = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'sigmav = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0.0 <= grisk <= 1.0',
                    'and grisk*(1.0-grisk)*sigmav^2 <= kappa^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 17 and _ifmt == 17:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Quadrature has not converged to the specified accuracy.',
                    'However, the result should be a reasonable approximation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Solution cannot be computed accurately. Check values of input arguments.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def opt_heston_greeks(calput, m, n, x, s, t, sigmav, kappa, corr, var0, eta, grisk, r, q, p, delta, gamma, vega, theta, rho, vanna, charm, speed, zomma, vomma):
    u"""
    Heston's model option pricing formula with Greeks.

    ``opt_heston_greeks`` computes the European option price given by
    Heston's stochastic volatility model together with its sensitivities
    (Greeks).

    For full information please refer to the NAG Library document for
    s30nb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30nbf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigmav : float
        The volatility, sigma_v, of the volatility process, sqrt(v_t).
        Note that a rate of 20% should be entered as 0.2.

    kappa : float
        kappa, the long term mean reversion rate of the volatility.

    corr : float
        The correlation between the two standard Brownian motions for
        the asset price and the volatility.

    var0 : float
        The initial value of the variance, v_t, of the asset price.

    eta : float
        eta, the long term mean of the variance of the asset price.

    grisk : float
        The risk aversion parameter, gamma, of the representative agent.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i-1] >= *<value>* and `x`[i-1] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i-1] >= *<value>*.

        (`errno` 7)
            On entry, `sigmav` = *<value>*.

            Constraint: `sigmav` > 0.0.

        (`errno` 8)
            On entry, `kappa` = *<value>*.

            Constraint: `kappa` > 0.0.

        (`errno` 9)
            On entry, `corr` = *<value>*.

            Constraint: abs(`corr`) <= 1.0.

        (`errno` 10)
            On entry, `var0` = *<value>*.

            Constraint: `var0` >= 0.0.

        (`errno` 11)
            On entry, `eta` = *<value>*.

            Constraint: `eta` > 0.0.

        (`errno` 12)
            On entry, `grisk` = *<value>*, `sigmav` = *<value>* and
            `kappa` = *<value>*.

            Constraint: 0.0 <= `grisk` <= 1.0 and
            `grisk`*(1.0-`grisk`)*`sigmav`^2 <= `kappa`^2.

        (`errno` 13)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 14)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 17)
            Quadrature has not converged to the required accuracy.
            However, the result should be a reasonable approximation.

        (`errno` 18)
            Solution cannot be computed accurately. Check values of
            input arguments.

    Notes
    -----
    ``opt_heston_greeks`` computes the price and sensitivities of a
    European option using Heston's stochastic volatility model.
    The return on the asset price, S, is

        dS/S = (r-q)dt+sqrt(v_t)dW_t^(1)

    and the instantaneous variance, v_t, is defined by a mean-reverting
    square root stochastic process,

        dv_t = kappa(eta-v_t)dt+sigma_v sqrt(v_t)dW_t^(2),

    where r is the risk free annual interest rate; q is the annual
    dividend rate; v_t is the variance of the asset price; sigma_v is
    the volatility of the volatility, sqrt(v_t); kappa is the mean
    reversion rate; eta is the long term variance. dW_t^(i), for i =
    1,2,...,2, denotes two correlated standard Brownian motions with

        ℂov[dW_t^(1),dW_t^(2)] = rho dt.

    The option price is computed by evaluating the integral transform
    given by Lewis (2000) using the form of the characteristic function
    discussed by Albrecher `et al.` (2007), see also Kilin (2006).

        P_call = Se^(-qT)-Xe^(-rT)1/piRe[int_0+i/2^infinity
        +i/2e^(-ikX_)H^(k,v,T)/(k^2-ik)dk],

    where X_= ln(S/X)+(r-q)T and

        H^(k,v,T) = exp(2 kappaeta/(sigma_v^2)[tg -ln((1-he^(-xi
        t))/(1-h))]+v_t g[(1-e^(-xi t))/(1-he^(-xi t))]),

        g = 1/2(b-xi), h = (b-xi)/(b+xi), t = sigma_v^2T/2,

        xi = [b^2+4(k^2-ik)/sigma_v^2]^(1/2),

        b =
        2/sigma_v^2[(1-gamma+ik)rhosigma_v+sqrt(kappa^2-gamma(1-gamma)sigma_v^2)]

    with t = sigma_v^2T/2.
    Here gamma is the risk aversion parameter of the representative
    agent with 0 <= gamma <= 1 and gamma(1-gamma)sigma_v^2 <= kappa^2.
    The value gamma = 1 corresponds to lambda = 0, where lambda is the
    market price of risk in Heston (1993) (see Lewis (2000) and Rouah
    and Vainberg (2007)).

    The price of a put option is obtained by put-call parity.

        P_put = P_call+Xe^(-rT)-Se^(-qT).

    Writing the expression for the price of a call option as

        P_call = Se^(-qT)-Xe^(-rT)1/piRe[int_0+i/2^infinity
        +i/2I(k,r,S,T,v)dk]

    then the sensitivities or Greeks can be obtained in the following
    manner,

    Delta

            dP_call/dS = e^(-qT)+Xe^(-rT)/S1/piRe[int_0+i/2^infinity
            +i/2(ik)I(k,r,S,T,v)dk],

    Vega

            dP/dv = -Xe^(-rT)1/piRe[int_0-i/2^0+i/2f_2
            I(k,r,j,S,T,v)dk], where f_2 = g[(1-e^(-xi t))/(1-he^(-xi
            t))],

    Rho

            dP_call/dr = TXe^(-rT)1/piRe[int_0+i/2^infinity
            +i/2(1+ik)I(k,r,S,T,v)dk].

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Albrecher, H, Mayer, P, Schoutens, W and Tistaert, J, 2007, `The
    little Heston trap`, Wilmott Magazine (January 2007), 83--92

    Heston, S, 1993, `A closed-form solution for options with stochastic
    volatility with applications to bond and currency options`, Review
    of Financial Studies (6), 327--343

    Kilin, F, 2006, `Accelerating the calibration of stochastic
    volatility models`, MPRA Paper No. 2975,
    https://mpra.ub.uni-muenchen.de/2975/

    Lewis, A L, 2000, `Option valuation under stochastic volatility`,
    Finance Press, USA

    Rouah, F D and Vainberg, G, 2007, `Option Pricing Models and
    Volatility using Excel-VBA`, John Wiley and Sons, Inc
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_heston_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30nbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigmav.
        utils._BYREF_FLOAT64, # Input, kappa.
        utils._BYREF_FLOAT64, # Input, corr.
        utils._BYREF_FLOAT64, # Input, var0.
        utils._BYREF_FLOAT64, # Input, eta.
        utils._BYREF_FLOAT64, # Input, grisk.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = {:s}.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigmav = utils._EngineFloat64ScalarType(
        sigmav, {'fun_name': fun_name, 'entity': 'sigmav'},
    ).to_ctypes()
    _kappa = utils._EngineFloat64ScalarType(
        kappa, {'fun_name': fun_name, 'entity': 'kappa'},
    ).to_ctypes()
    _corr = utils._EngineFloat64ScalarType(
        corr, {'fun_name': fun_name, 'entity': 'corr'},
    ).to_ctypes()
    _var0 = utils._EngineFloat64ScalarType(
        var0, {'fun_name': fun_name, 'entity': 'var0'},
    ).to_ctypes()
    _eta = utils._EngineFloat64ScalarType(
        eta, {'fun_name': fun_name, 'entity': 'eta'},
    ).to_ctypes()
    _grisk = utils._EngineFloat64ScalarType(
        grisk, {'fun_name': fun_name, 'entity': 'grisk'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigmav,
        _kappa,
        _corr,
        _var0,
        _eta,
        _grisk,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _vanna,
        _charm,
        _speed,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i-1] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigmav = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigmav > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kappa > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(corr) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var0 >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eta > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, grisk = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'sigmav = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0.0 <= grisk <= 1.0',
                    'and grisk*(1.0-grisk)*sigmav^2 <= kappa^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 17 and _ifmt == 17:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Quadrature has not converged to the required accuracy.',
                    'However, the result should be a reasonable approximation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Solution cannot be computed accurately. Check values of input arguments.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def opt_heston_term(calput, m, numts, x, fwd, disc, ts, t, alpha, lamda, corr, sigmat, var0, p):
    u"""
    Heston's model option pricing with term structure.

    ``opt_heston_term`` computes the European option price given by
    Heston's stochastic volatility model with term structure.

    For full information please refer to the NAG Library document for
    s30nc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30ncf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        m, the number of strike prices to be used.

    numts : int
        n_s, the number of subintervals into which the time to expiry,
        T, is divided.

    x : float, shape (`m`)
        `x`[i-1] contains the ith strike price, for i = 1,2,...,m.

    fwd : float
        The forward price of the asset.

    disc : float
        The discount factor, where the current price of the underlying
        asset, S_0, is given as S_0 = `disc`*`fwd`.

    ts : float, shape (`numts`)
        `ts`[i-1] must contain the length of the time intervals for
        which the corresponding element of `alpha`, `lamda`, `corr` and
        `sigmat` apply. These should be ordered as they occur in time
        i.e., Delta t_i = t_i-t_{i-1}.

    t : float
        `t` contains the time to expiry. If T > sum Delta t_i then the
        parameters associated with the last time interval are extended
        to the expiry time. If T < sum Delta t_i then the parameters
        specified are used up until the expiry time. The rest are
        ignored.

    alpha : float, shape (`numts`)
        `alpha`[i-1] must contain the value of alpha_t, the value of the
        volatility of the scaled volatility, sqrt(nu), over time
        subinterval Delta t_i.

    lamda : float, shape (`numts`)
        `lamda`[i-1] must contain the value, lambda_t, of the mean
        reversion parameter over the time subinterval Delta t_i.

    corr : float, shape (`numts`)
        `corr`[i-1] must contain the value, rho_t, of the correlation
        parameter over the time subinterval Delta t_i.

    sigmat : float, shape (`numts`)
        `sigmat`[i-1] must contain the value, sigma_t, of the variance
        scale factor over the time subinterval Delta t_i.

    var0 : float
        nu_0, the initial scaled variance.

    p : float, shape (`m`), modified in place
        `On exit`: `p`[i-1] contains the computed option price at the
        expiry time, T, corresponding to strike `x`[i-1] for the
        specified term structure, for i = 1,2,...,`m`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` != 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `numts` = *<value>*.

            Constraint: `numts` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `x`[i-1] <= *<value>*.

        (`errno` 5)
            On entry, `fwd` = *<value>*.

            Constraint: *<value>* <= `fwd` <= *<value>*.

        (`errno` 6)
            On entry, `disc` = *<value>*.

            Constraint: *<value>* <= `disc` <= *<value>*.

        (`errno` 7)
            On entry, `ts`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `ts`[i-1] <= *<value>*.

        (`errno` 8)
            On entry, `t` = *<value>*.

            Constraint: `t` >= *<value>*.

        (`errno` 9)
            On entry, `alpha`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `alpha`[i-1] <= *<value>*.

        (`errno` 10)
            On entry, `lamda`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `lamda`[i-1] <= *<value>*.

        (`errno` 11)
            On entry, `corr`[*<value>*] = *<value>*.

            Constraint: abs(`corr`[i-1]) <= 1.0.

        (`errno` 12)
            On entry, `sigmat`[*<value>*] = *<value>*.

            Constraint: *<value>* <= `sigmat`[i-1] <= *<value>*.

        (`errno` 13)
            On entry, `var0` = *<value>*.

            Constraint: `var0` > 0.0.

        (`errno` 14)
            Quadrature has not converged to the specified accuracy.
            However, the result should be a reasonable approximation.

        (`errno` 15)
            Solution cannot be computed accurately. Check values of
            input arguments.

    Notes
    -----
    ``opt_heston_term`` computes the price of a European option for
    Heston's stochastic volatility model with time-dependent parameters
    which are piecewise constant.
    Starting from the stochastic volatility model given by the
    Stochastic Differential Equation (SDE) system defined by Heston
    (1993), a scaling of the variance process is introduced, together
    with a normalization, setting the long run variance, eta, equal to
    1.
    This leads to

        dS_t/S_t = mu_t dt+sigma_t sqrt(nu_t)dW_t^(1),

        d nu_t = lambda_t(1-nu_t)dt+alpha_t sqrt(nu_t)dW_t^(2),

        Cov[dW_t^(1),dW_t^(2)] = rho_t dt,

    where mu_t = r_t-q_t is the drift term representing the contribution
    of interest rates, r_t, and dividends, q_t, while sigma_t is the
    scaling parameter, nu_t is the scaled variance, lambda_t is the mean
    reversion rate and alpha_t is the volatility of the scaled
    volatility, sqrt(nu_t).
    Then, W_t^(i), for i = 1,2,...,2, are two standard Brownian motions
    with correlation parameter rho_t.
    Without loss of generality, the drift term, mu_t, is eliminated by
    modelling the forward price, F_t, directly, instead of the spot
    price, S_t, with

        F_t = S_0 exp(int_0^t mu_s ds).

    If required, the spot can be expressed as, S_0 = DF_t, where D is
    the discount factor.

    The option price is computed by dividing the time to expiry, T, into
    n_s subintervals [t_0,t_1],...,[t_{i-1},t_i],...,[t_{n_s-1},T] and
    applying the method of characteristic functions to each subinterval,
    with appropriate initial conditions.
    Thus, a pair of ordinary differential equations (one of which is a
    Riccati equation) is solved on each subinterval as outlined in
    Elices (2008) and Mikhailov and Nögel (2003).
    Reversing time by taking tau = T-t, the characteristic function
    solution for the first time subinterval, starting at tau = 0, is
    given by Heston (1993), while the solution on each following
    subinterval uses the solution of the preceding subinterval as
    initial condition to compute the value of the characteristic
    function.

    In the case of a 'flat' term structure, i.e., the parameters are
    constant over the time of the option, T, the form of the SDE system
    given by Heston (1993) can be recovered by setting kappa = lambda_t,
    eta = sigma_t^2, sigma_v = sigma_talpha_t and V_0 = sigma_t^2V_0.

    Conversely, given the Heston form of the SDE pair, to get the term
    structure form set lambda_t = kappa, sigma_t = sqrt(eta), alpha_t =
    (sigma_v)/(sqrt(eta)) and V_0 = (V_0)/eta.

    References
    ----------
    Bain, A, 2011, `Private communication`

    Elices, A, 2008, `Models with time-dependent parameters using
    transform methods: application to Heston’s model`, arXiv:0708.2020v2

    Heston, S, 1993, `A closed-form solution for options with stochastic
    volatility with applications to bond and currency options`, Review
    of Financial Studies (6), 327--343

    Mikhailov, S and Nögel, U, 2003, `Heston’s Stochastic Volatility
    Model Implementation, Calibration and Some Extensions`, Wilmott
    Magazine July/August, 74--79
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_heston_term'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30ncft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, numts.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, fwd.
        utils._BYREF_FLOAT64, # Input, disc.
        utils._EngineFloat64ArrayType.c_type, # Input, ts[numts].
        utils._BYREF_FLOAT64, # Input, t.
        utils._EngineFloat64ArrayType.c_type, # Input, alpha[numts].
        utils._EngineFloat64ArrayType.c_type, # Input, lamda[numts].
        utils._EngineFloat64ArrayType.c_type, # Input, corr[numts].
        utils._EngineFloat64ArrayType.c_type, # Input, sigmat[numts].
        utils._BYREF_FLOAT64, # Input, var0.
        utils._EngineFloat64ArrayType.c_type, # Output, p[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput != \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _numts = utils._EngineIntScalarType(
        numts, {'fun_name': fun_name, 'entity': 'numts'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _fwd = utils._EngineFloat64ScalarType(
        fwd, {'fun_name': fun_name, 'entity': 'fwd'},
    ).to_ctypes()
    _disc = utils._EngineFloat64ScalarType(
        disc, {'fun_name': fun_name, 'entity': 'disc'},
    ).to_ctypes()
    _numts_py = utils._EngineIntScalarType.to_py(_numts)
    _ts = utils._EngineFloat64ArrayType(
        ts, {'fun_name': fun_name, 'entity': 'ts'},
        exp_shape=(_numts_py,),
    ).to_ctypes()
    _t = utils._EngineFloat64ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ArrayType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
        exp_shape=(_numts_py,),
    ).to_ctypes()
    _lamda = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_numts_py,),
    ).to_ctypes()
    _corr = utils._EngineFloat64ArrayType(
        corr, {'fun_name': fun_name, 'entity': 'corr'},
        exp_shape=(_numts_py,),
    ).to_ctypes()
    _sigmat = utils._EngineFloat64ArrayType(
        sigmat, {'fun_name': fun_name, 'entity': 'sigmat'},
        exp_shape=(_numts_py,),
    ).to_ctypes()
    _var0 = utils._EngineFloat64ScalarType(
        var0, {'fun_name': fun_name, 'entity': 'var0'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _numts,
        _x,
        _fwd,
        _disc,
        _ts,
        _t,
        _alpha,
        _lamda,
        _corr,
        _sigmat,
        _var0,
        _p,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, numts = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: numts >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= x[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fwd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: {:s} <= fwd <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, disc = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: {:s} <= disc <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ts[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= ts[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: t >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= alpha[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= lamda[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corr[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(corr[i-1]) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigmat[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: {:s} <= sigmat[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var0 > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Quadrature has not converged to the specified accuracy.',
                    'However, the result should be a reasonable approximation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 15 and _ifmt == 15:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Solution cannot be computed accurately. Check values of input arguments.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_heston_more_greeks(calput, m, n, x, s, t, sigmav, kappa, corr, var0, eta, grisk, r, q, p, delta, gamma, vega, theta, rho, vanna, charm, speed, zomma, vomma, dp_dx, dp_dq, dp_deta, dp_dkappa, dp_dsigmav, dp_dcorr, dp_dgrisk):
    u"""
    Heston's model option pricing formula with Greeks.

    ``opt_heston_more_greeks`` computes the European option price given
    by Heston's stochastic volatility model together with its
    sensitivities (Greeks).

    For full information please refer to the NAG Library document for
    s30nd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30ndf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigmav : float
        The volatility, sigma_v, of the volatility process, sqrt(v_t).
        Note that a rate of 20% should be entered as 0.2.

    kappa : float
        kappa, the long term mean reversion rate of the volatility.

    corr : float
        The correlation between the two standard Brownian motions for
        the asset price and the volatility.

    var0 : float
        The initial value of the variance, v_t, of the asset price.

    eta : float
        eta, the long term mean of the variance of the asset price.

    grisk : float
        The risk aversion parameter, gamma, of the representative agent.

    r : float, shape (`n`)
        `r`[i-1] must contain R_i, the ith annual risk-free interest
        rate, continuously compounded, for i = 1,2,...,`n`. Note that a
        rate of 5% should be entered as 0.05.

    q : float, shape (`n`)
        `q`[i-1] must contain Q_i, the ith annual continuous yield rate,
        for i = 1,2,...,`n`. Note that a rate of 8% should be entered as
        0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/dv_t,
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr_i, for
        i = 1,2,...,`m` and j = 1,2,...,`n`.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    dp_dx : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dx`[i-1,j-1], contains the derivative measuring
        the sensitivity of the option price P_{ij} to change in the
        strick price, X_i, i.e., dP_{ij}/dX_i, for i = 1,2,...,`m` and j
        = 1,2,...,`n`.

    dp_dq : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dq`[i-1,j-1], contains the derivative measuring
        the sensitivity of the option price P_{ij} to change in the
        annual continuous yield rate, q_j, i.e., dP_{ij}/dq_j, for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    dp_deta : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_deta`[i-1,j-1], contains the derivative measuring
        the sensitivity of the option price P_{ij} to change in the long
        term mean of the variance eta of the asset price, i.e.,
        dP_{ij}/d eta, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    dp_dkappa : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dkappa`[i-1,j-1], contains the derivative
        measuring the sensitivity of the option price P_{ij} to change
        in the long term mean reversion rate kappa of the volatility,
        i.e., dP_{ij}/d kappa, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    dp_dsigmav : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dsigmav`[i-1,j-1], contains the derivative
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility sigma_v of the volatility process, sqrt(v_t),
        i.e., dP_{ij}/d sigma_v, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    dp_dcorr : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dcorr`[i-1,j-1], contains the derivative
        measuring the sensitivity of the option price P_{ij} to change
        in the correlation rho between the two standard Brownian motions
        for the asset price and the volatility, i.e., dP_{ij}/d rho, for
        i = 1,2,...,`m` and j = 1,2,...,`n`.

    dp_dgrisk : float, shape (`m`, `n`), modified in place
        `On exit`: `dp_dgrisk`[i-1,j-1], contains the derivative
        measuring the sensitivity of the option price P_{ij} to change
        in the risk aversion parameter gamma of the representative
        agent,, i.e., dP_{ij}/d gamma, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i-1] >= *<value>* and `x`[i-1] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i-1] >= *<value>*.

        (`errno` 7)
            On entry, `sigmav` = *<value>*.

            Constraint: `sigmav` > 0.0.

        (`errno` 8)
            On entry, `kappa` = *<value>*.

            Constraint: `kappa` > 0.0.

        (`errno` 9)
            On entry, `corr` = *<value>*.

            Constraint: abs(`corr`) <= 1.0.

        (`errno` 10)
            On entry, `var0` = *<value>*.

            Constraint: `var0` >= 0.0.

        (`errno` 11)
            On entry, `eta` = *<value>*.

            Constraint: `eta` > 0.0.

        (`errno` 12)
            On entry, `grisk` = *<value>*, `sigmav` = *<value>* and
            `kappa` = *<value>*.

            Constraint: 0.0 <= `grisk` <= 1.0 and
            `grisk`*(1.0-`grisk`)*`sigmav`^2 <= `kappa`^2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 17)
            Quadrature has not converged to the required accuracy.
            However, the result should be a reasonable approximation.

        (`errno` 18)
            Solution cannot be computed accurately. Check values of
            input arguments.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``opt_heston_more_greeks`` computes the price and sensitivities of a
    European option using Heston's stochastic volatility model.
    The return on the asset price, S, is

        dS/S = (r-q)dt+sqrt(v_t)dW_t^(1)

    and the instantaneous variance, v_t, is defined by a mean-reverting
    square root stochastic process,

        dv_t = kappa(eta-v_t)dt+sigma_v sqrt(v_t)dW_t^(2),

    where r is the risk free annual interest rate; q is the annual
    dividend rate; v_t is the variance of the asset price; sigma_v is
    the volatility of the volatility, sqrt(v_t); kappa is the mean
    reversion rate; eta is the long term variance. dW_t^(i), for i =
    1,2,...,2, denotes two correlated standard Brownian motions with

        ℂov[dW_t^(1),dW_t^(2)] = rho dt.

    The option price is computed by evaluating the integral transform
    given by Lewis (2000) using the form of the characteristic function
    discussed by Albrecher `et al.` (2007), see also Kilin (2006).

        P_call = Se^(-qT)-Xe^(-rT)1/piRe[int_0+i/2^infinity
        +i/2e^(-ikX_)H^(k,v,T)/(k^2-ik)dk],

    where X_= ln(S/X)+(r-q)T and

        H^(k,v,T) = exp(2 kappaeta/(sigma_v^2)[tg -ln((1-he^(-xi
        t))/(1-h))]+v_t g[(1-e^(-xi t))/(1-he^(-xi t))]),

        g = 1/2(b-xi), h = (b-xi)/(b+xi), t = sigma_v^2T/2,

        xi = [b^2+4(k^2-ik)/sigma_v^2]^(1/2),

        b =
        2/sigma_v^2[(1-gamma+ik)rhosigma_v+sqrt(kappa^2-gamma(1-gamma)sigma_v^2)]

    with t = sigma_v^2T/2.
    Here gamma is the risk aversion parameter of the representative
    agent with 0 <= gamma <= 1 and gamma(1-gamma)sigma_v^2 <= kappa^2.
    The value gamma = 1 corresponds to lambda = 0, where lambda is the
    market price of risk in Heston (1993) (see Lewis (2000) and Rouah
    and Vainberg (2007)).

    The price of a put option is obtained by put-call parity.

        P_put = P_call+Xe^(-rT)-Se^(-qT).

    Writing the expression for the price of a call option as

        P_call = Se^(-qT)-Xe^(-rT)1/piRe[int_0+i/2^infinity
        +i/2I(k,r,S,T,v)dk]

    then the sensitivities or Greeks can be obtained in the following
    manner,

    Delta

            dP_call/dS = e^(-qT)+Xe^(-rT)/S1/piRe[int_0+i/2^infinity
            +i/2(ik)I(k,r,S,T,v)dk],

    Vega

            dP/dv = -Xe^(-rT)1/piRe[int_0-i/2^0+i/2f_2
            I(k,r,j,S,T,v)dk], where f_2 = g[(1-e^(-xi t))/(1-he^(-xi
            t))],

    Rho

            dP_call/dr = TXe^(-rT)1/piRe[int_0+i/2^infinity
            +i/2(1+ik)I(k,r,S,T,v)dk].

    The option price P_{ij} = P(X = X_i,T = T_j,r = r_j,q = q_j) is
    computed for each strike price in a set X_i, i = 1,2,...,m, and for
    each expiry time in a set T_j, j = 1,2,...,n.
    The continuously compounded annual risk-free interest rate used is
    r_j, j = 1,2,...,n, and the annual continuous yield rate used is
    q_j, j = 1,2,...,n.

    References
    ----------
    Albrecher, H, Mayer, P, Schoutens, W and Tistaert, J, 2007, `The
    little Heston trap`, Wilmott Magazine (January 2007), 83--92

    Heston, S, 1993, `A closed-form solution for options with stochastic
    volatility with applications to bond and currency options`, Review
    of Financial Studies (6), 327--343

    Kilin, F, 2006, `Accelerating the calibration of stochastic
    volatility models`, MPRA Paper No. 2975,
    https://mpra.ub.uni-muenchen.de/2975/

    Lewis, A L, 2000, `Option valuation under stochastic volatility`,
    Finance Press, USA

    Rouah, F D and Vainberg, G, 2007, `Option Pricing Models and
    Volatility using Excel-VBA`, John Wiley and Sons, Inc

    See Also
    --------
    :meth:`naginterfaces.library.examples.specfun.opt_heston_more_greeks_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_heston_more_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30ndft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigmav.
        utils._BYREF_FLOAT64, # Input, kappa.
        utils._BYREF_FLOAT64, # Input, corr.
        utils._BYREF_FLOAT64, # Input, var0.
        utils._BYREF_FLOAT64, # Input, eta.
        utils._BYREF_FLOAT64, # Input, grisk.
        utils._EngineFloat64ArrayType.c_type, # Input, r[n].
        utils._EngineFloat64ArrayType.c_type, # Input, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dx[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dq[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_deta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dkappa[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dsigmav[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dcorr[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, dp_dgrisk[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (zomma, 'zomma'), (vomma, 'vomma'), (dp_dx, 'dp_dx'), (dp_dq, 'dp_dq'), (dp_deta, 'dp_deta'), (dp_dkappa, 'dp_dkappa'), (dp_dsigmav, 'dp_dsigmav'), (dp_dcorr, 'dp_dcorr'), (dp_dgrisk, 'dp_dgrisk')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = {:s}.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigmav = utils._EngineFloat64ScalarType(
        sigmav, {'fun_name': fun_name, 'entity': 'sigmav'},
    ).to_ctypes()
    _kappa = utils._EngineFloat64ScalarType(
        kappa, {'fun_name': fun_name, 'entity': 'kappa'},
    ).to_ctypes()
    _corr = utils._EngineFloat64ScalarType(
        corr, {'fun_name': fun_name, 'entity': 'corr'},
    ).to_ctypes()
    _var0 = utils._EngineFloat64ScalarType(
        var0, {'fun_name': fun_name, 'entity': 'var0'},
    ).to_ctypes()
    _eta = utils._EngineFloat64ScalarType(
        eta, {'fun_name': fun_name, 'entity': 'eta'},
    ).to_ctypes()
    _grisk = utils._EngineFloat64ScalarType(
        grisk, {'fun_name': fun_name, 'entity': 'grisk'},
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _q = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _dp_dx_cls = utils._EngineFloat64ArrayType(
        dp_dx, {'fun_name': fun_name, 'entity': 'dp_dx'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dx = _dp_dx_cls.to_ctypes(can_reference_contents=False)
    _dp_dq_cls = utils._EngineFloat64ArrayType(
        dp_dq, {'fun_name': fun_name, 'entity': 'dp_dq'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dq = _dp_dq_cls.to_ctypes(can_reference_contents=False)
    _dp_deta_cls = utils._EngineFloat64ArrayType(
        dp_deta, {'fun_name': fun_name, 'entity': 'dp_deta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_deta = _dp_deta_cls.to_ctypes(can_reference_contents=False)
    _dp_dkappa_cls = utils._EngineFloat64ArrayType(
        dp_dkappa, {'fun_name': fun_name, 'entity': 'dp_dkappa'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dkappa = _dp_dkappa_cls.to_ctypes(can_reference_contents=False)
    _dp_dsigmav_cls = utils._EngineFloat64ArrayType(
        dp_dsigmav, {'fun_name': fun_name, 'entity': 'dp_dsigmav'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dsigmav = _dp_dsigmav_cls.to_ctypes(can_reference_contents=False)
    _dp_dcorr_cls = utils._EngineFloat64ArrayType(
        dp_dcorr, {'fun_name': fun_name, 'entity': 'dp_dcorr'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dcorr = _dp_dcorr_cls.to_ctypes(can_reference_contents=False)
    _dp_dgrisk_cls = utils._EngineFloat64ArrayType(
        dp_dgrisk, {'fun_name': fun_name, 'entity': 'dp_dgrisk'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _dp_dgrisk = _dp_dgrisk_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigmav,
        _kappa,
        _corr,
        _var0,
        _eta,
        _grisk,
        _r,
        _q,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _vanna,
        _charm,
        _speed,
        _zomma,
        _vomma,
        _dp_dx,
        _dp_dq,
        _dp_deta,
        _dp_dkappa,
        _dp_dsigmav,
        _dp_dcorr,
        _dp_dgrisk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i-1] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-1] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigmav = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigmav > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kappa > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: abs(corr) <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var0 >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eta > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, grisk = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'sigmav = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'kappa = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0.0 <= grisk <= 1.0',
                    'and grisk*(1.0-grisk)*sigmav^2 <= kappa^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 17 and _ifmt == 17:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Quadrature has not converged to the required accuracy.',
                    'However, the result should be a reasonable approximation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Solution cannot be computed accurately. Check values of input arguments.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        _dp_dx_cls.refresh_py(_dp_dx)
        _dp_dq_cls.refresh_py(_dp_dq)
        _dp_deta_cls.refresh_py(_dp_deta)
        _dp_dkappa_cls.refresh_py(_dp_dkappa)
        _dp_dsigmav_cls.refresh_py(_dp_dsigmav)
        _dp_dcorr_cls.refresh_py(_dp_dcorr)
        _dp_dgrisk_cls.refresh_py(_dp_dgrisk)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def opt_amer_bs_price(calput, m, n, x, s, t, sigma, r, q, p):
    u"""
    American option, Bjerksund and Stensland pricing formula.

    ``opt_amer_bs_price`` computes the Bjerksund and Stensland (2002)
    approximation to the price of an American option.

    For full information please refer to the NAG Library document for
    s30qc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30qcf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    q : float
        q, the annual continuous yield rate. Note that a rate of 8%
        should be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

        (`errno` 9)
            On entry, `q` = *<value>*.

            Constraint: `q` >= 0.0.

        (`errno` 14)
            On entry, `s` = *<value>* and beta = *<value>*.

            Constraint: `s`^beta < *<value>*.

    Notes
    -----
    ``opt_amer_bs_price`` computes the price of an American option using
    the closed form approximation of Bjerksund and Stensland (2002).
    The time to maturity, T, is divided into two periods, each with a
    flat early exercise boundary, by choosing a time t in [0,T], such
    that t = 1/2(sqrt(5)-1)T.
    The two boundary values are defined as x~ = X~(t), X~ = X~(T) with

        X~(tau) = B_0+(B_infinity -B_0)(1-exp{h(tau)}),

    where

        h(tau) = -(b tau+2 sigma*sqrt(tau))((X^2)/(B_infinity -B_0)B_0),

        B_infinity = beta/(beta-1)X, B_0 = max{X,(r/(r-b))X},

        beta = (1/2-b/(sigma^2))+sqrt((b/(sigma^2)-1/2)^2+2r/(sigma^2)).

    with b = r-q, the cost of carry, where r is the risk-free interest
    rate and q is the annual dividend rate.
    Here X is the strike price and sigma is the annual volatility.

    The price of an American call option is approximated as

        [table omitted]

    where alpha, phi and Psi are as defined in Bjerksund and Stensland
    (2002).

    The price of a put option is obtained by the put-call
    transformation,

        P_put(X,S,T,sigma,r,q) = P_call(S,X,T,sigma,q,r).

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Bjerksund, P and Stensland, G, 2002, `Closed form valuation of
    American options` (Discussion Paper 2002/09), NHH Bergen Norway

    Genz, A, 2004, `Numerical computation of rectangular bivariate and
    trivariate Normal and t probabilities`, Statistics and Computing
    (14), 151--160
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_amer_bs_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30qcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, q.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _q = utils._EngineFloat64ScalarType(
        q, {'fun_name': fun_name, 'entity': 'q'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _q,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: q >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s} and beta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s^beta < {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_asian_geom_price(calput, m, n, x, s, t, sigma, r, b, p):
    u"""
    Asian option, geometric continuous average rate pricing formula.

    ``opt_asian_geom_price`` computes the Asian geometric continuous
    average-rate option price.

    For full information please refer to the NAG Library document for
    s30sa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30saf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    b : float
        b, the annual cost of carry rate. Note that a rate of 8% should
        be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

    Notes
    -----
    ``opt_asian_geom_price`` computes the price of an Asian geometric
    continuous average-rate option for constant volatility, sigma,
    risk-free rate, r, and cost of carry, b (see Kemna and Vorst
    (1990)).
    For a given strike price, X, the price of a call option with
    underlying price, S, and time to expiry, T, is

        P_call = Se^((b_-r)T)Phi(d__1)-Xe^(-rT)Phi(d__2),

    and the corresponding put option price is

        P_put = Xe^(-rT)Phi(-d__2)-Se^((b_-r)T)Phi(-d__1),

    where

        d__1 = (ln(S/X)+(b_+sigma_^2/2)T)/sigma_sqrt(T)

    and

        d__2 = d__1-sigma_sqrt(T),

    with

        sigma_= sigma/(sqrt(3)), b_= 1/2(r-(sigma^2)/6).

    Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^xexp(-y^2/2)dy.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Kemna, A and Vorst, A, 1990, `A pricing method for options based on
    average asset values`, Journal of Banking and Finance (14), 113--129
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_asian_geom_price'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _b,
        _p,
        _ldp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def opt_asian_geom_greeks(calput, m, n, x, s, t, sigma, r, b, p, delta, gamma, vega, theta, rho, crho, vanna, charm, speed, colour, zomma, vomma):
    u"""
    Asian option, geometric continuous average rate pricing formula with
    Greeks.

    ``opt_asian_geom_greeks`` computes the Asian geometric continuous
    average-rate option price together with its sensitivities (Greeks).

    For full information please refer to the NAG Library document for
    s30sb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/s/s30sbf.html

    Parameters
    ----------
    calput : str, length 1
        Determines whether the option is a call or a put.

        `calput` = 'C'

            A call; the holder has a right to buy.

        `calput` = 'P'

            A put; the holder has a right to sell.

    m : int
        The number of strike prices to be used.

    n : int
        The number of times to expiry to be used.

    x : float, shape (`m`)
        `x`[i-1] must contain X_i, the ith strike price, for i =
        1,2,...,`m`.

    s : float
        S, the price of the underlying asset.

    t : float, shape (`n`)
        `t`[i-1] must contain T_i, the ith time, in years, to expiry,
        for i = 1,2,...,`n`.

    sigma : float
        sigma, the volatility of the underlying asset. Note that a rate
        of 15% should be entered as 0.15.

    r : float
        r, the annual risk-free interest rate, continuously compounded.
        Note that a rate of 5% should be entered as 0.05.

    b : float
        b, the annual cost of carry rate. Note that a rate of 8% should
        be entered as 0.08.

    p : float, shape (`m`, `n`), modified in place
        `On exit`: `p`[i-1,j-1] contains P_{ij}, the option price
        evaluated for the strike price `x`_i at expiry `t`_j for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    delta : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `delta`
        contains the sensitivity, dP/dS, of the option price to change
        in the price of the underlying asset.

    gamma : float, shape (`m`, `n`), modified in place
        `On exit`: the leading `m`*`n` part of the array `gamma`
        contains the sensitivity, d^2P/dS^2, of `delta` to change in the
        price of the underlying asset.

    vega : float, shape (`m`, `n`), modified in place
        `On exit`: `vega`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the volatility of the underlying asset, i.e., dP_{ij}/d
        sigma, for i = 1,2,...,`m` and j = 1,2,...,`n`.

    theta : float, shape (`m`, `n`), modified in place
        `On exit`: `theta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in time, i.e., -dP_{ij}/dT, for i = 1,2,...,`m` and j =
        1,2,...,`n`, where b = r-q.

    rho : float, shape (`m`, `n`), modified in place
        `On exit`: `rho`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the annual risk-free interest rate, i.e., -dP_{ij}/dr, for i
        = 1,2,...,`m` and j = 1,2,...,`n`.

    crho : float, shape (`m`, `n`), modified in place
        `On exit`: `delta`[i-1,j-1], contains the first-order Greek
        measuring the sensitivity of the option price P_{ij} to change
        in the price of the underlying asset, i.e., -dP_{ij}/dS, for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vanna : float, shape (`m`, `n`), modified in place
        `On exit`: `vanna`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the asset price, i.e., -d
        Delta_{ij}/dT = -d^2P_{ij}/(dSd sigma), for i = 1,2,...,`m` and
        j = 1,2,...,`n`.

    charm : float, shape (`m`, `n`), modified in place
        `On exit`: `charm`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the time, i.e., -d Delta_{ij}/dT = -d^2P_{ij}/(dSdT),
        for i = 1,2,...,`m` and j = 1,2,...,`n`.

    speed : float, shape (`m`, `n`), modified in place
        `On exit`: `speed`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the price of the underlying asset, i.e., -d
        Gamma_{ij}/dS = -d^3P_{ij}/dS^3, for i = 1,2,...,`m` and j =
        1,2,...,`n`.

    colour : float, shape (`m`, `n`), modified in place
        `On exit`: `colour`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the time, i.e., -d Gamma_{ij}/dT =
        -d^3P_{ij}/(dSdT), for i = 1,2,...,`m` and j = 1,2,...,`n`.

    zomma : float, shape (`m`, `n`), modified in place
        `On exit`: `zomma`[i-1,j-1], contains the third-order Greek
        measuring the sensitivity of the second-order Greek Gamma_{ij}
        to change in the volatility of the underlying asset, i.e., -d
        Gamma_{ij}/d sigma = -d^3P_{ij}/(dS^2d sigma), for i =
        1,2,...,`m` and j = 1,2,...,`n`.

    vomma : float, shape (`m`, `n`), modified in place
        `On exit`: `vomma`[i-1,j-1], contains the second-order Greek
        measuring the sensitivity of the first-order Greek Delta_{ij} to
        change in the volatility of the underlying asset, i.e., -d
        Delta_{ij}/d sigma = -d^2P_{ij}/d sigma^2, for i = 1,2,...,`m`
        and j = 1,2,...,`n`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `calput` = *<value>*.

            Constraint: `calput` = 'C' or 'P'.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `x`[*<value>*] = *<value>*.

            Constraint: `x`[i] >= *<value>* and `x`[i] <= *<value>*.

        (`errno` 5)
            On entry, `s` = *<value>*.

            Constraint: `s` >= *<value>* and `s` <= *<value>*.

        (`errno` 6)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i] >= *<value>*.

        (`errno` 7)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 8)
            On entry, `r` = *<value>*.

            Constraint: `r` >= 0.0.

    Notes
    -----
    ``opt_asian_geom_greeks`` computes the price of an Asian geometric
    continuous average-rate option, together with the Greeks or
    sensitivities, which are the partial derivatives of the option price
    with respect to certain of the other input parameters.
    The annual volatility, sigma, risk-free rate, r, and cost of carry,
    b, are constants (see Kemna and Vorst (1990)).
    For a given strike price, X, the price of a call option with
    underlying price, S, and time to expiry, T, is

        P_call = Se^((b_-r)T)Phi(d__1)-Xe^(-rT)Phi(d__2),

    and the corresponding put option price is

        P_put = Xe^(-rT)Phi(-d__2)-Se^((b_-r)T)Phi(-d__1),

    where

        d__1 = (ln(S/X)+(b_+sigma_^2/2)T)/sigma_sqrt(T)

    and

        d__2 = d__1-sigma_sqrt(T),

    with

        sigma_= sigma/(sqrt(3)), b_= 1/2(b-(sigma^2)/6).

    Phi is the cumulative Normal distribution function,

        Phi(x) = 1/sqrt(2 pi)int_- infinity^xexp(-y^2/2)dy.

    The option price P_{ij} = P(X = X_i,T = T_j) is computed for each
    strike price in a set X_i, i = 1,2,...,m, and for each expiry time
    in a set T_j, j = 1,2,...,n.

    References
    ----------
    Kemna, A and Vorst, A, 1990, `A pricing method for options based on
    average asset values`, Journal of Banking and Finance (14), 113--129
    """

    excs = []
    fun_name = 'naginterfaces.base.specfun.opt_asian_geom_greeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('s30sbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, option.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, n].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, delta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, gamma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vega[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, rho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, crho[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vanna[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, charm[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, speed[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, colour[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, zomma[ldp, n].
        utils._EngineFloat64ArrayType.c_type, # Output, vomma[ldp, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (delta, 'delta'), (gamma, 'gamma'), (vega, 'vega'), (theta, 'theta'), (rho, 'rho'), (crho, 'crho'), (vanna, 'vanna'), (charm, 'charm'), (speed, 'speed'), (colour, 'colour'), (zomma, 'zomma'), (vomma, 'vomma')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _calput = utils._EngineCharScalarType(
        calput, {'fun_name': fun_name, 'entity': 'calput'}, exp_chrlen=1,
    ).to_ctypes()
    _calput_py = utils._EngineCharScalarType.to_py(_calput)
    if _calput_py.upper() == 'C':
        _option = utils._EngineIntScalarType.c_type(0)
    elif _calput_py.upper() == 'P':
        _option = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, calput = \'{:s}\'.'.format(
                    utils._STR_FMT_STR.format(_calput_py),
                ),
                'Constraint: calput = \'C\' or \'P\'.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_shape(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2)[0]
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _delta_cls = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _delta = _delta_cls.to_ctypes(can_reference_contents=False)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _gamma = _gamma_cls.to_ctypes(can_reference_contents=False)
    _vega_cls = utils._EngineFloat64ArrayType(
        vega, {'fun_name': fun_name, 'entity': 'vega'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vega = _vega_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _rho_cls = utils._EngineFloat64ArrayType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rho = _rho_cls.to_ctypes(can_reference_contents=False)
    _crho_cls = utils._EngineFloat64ArrayType(
        crho, {'fun_name': fun_name, 'entity': 'crho'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _crho = _crho_cls.to_ctypes(can_reference_contents=False)
    _vanna_cls = utils._EngineFloat64ArrayType(
        vanna, {'fun_name': fun_name, 'entity': 'vanna'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vanna = _vanna_cls.to_ctypes(can_reference_contents=False)
    _charm_cls = utils._EngineFloat64ArrayType(
        charm, {'fun_name': fun_name, 'entity': 'charm'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _charm = _charm_cls.to_ctypes(can_reference_contents=False)
    _speed_cls = utils._EngineFloat64ArrayType(
        speed, {'fun_name': fun_name, 'entity': 'speed'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _speed = _speed_cls.to_ctypes(can_reference_contents=False)
    _colour_cls = utils._EngineFloat64ArrayType(
        colour, {'fun_name': fun_name, 'entity': 'colour'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _colour = _colour_cls.to_ctypes(can_reference_contents=False)
    _zomma_cls = utils._EngineFloat64ArrayType(
        zomma, {'fun_name': fun_name, 'entity': 'zomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _zomma = _zomma_cls.to_ctypes(can_reference_contents=False)
    _vomma_cls = utils._EngineFloat64ArrayType(
        vomma, {'fun_name': fun_name, 'entity': 'vomma'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vomma = _vomma_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _option,
        _m,
        _n,
        _x,
        _s,
        _t,
        _sigma,
        _r,
        _b,
        _p,
        _ldp,
        _delta,
        _gamma,
        _vega,
        _theta,
        _rho,
        _crho,
        _vanna,
        _charm,
        _speed,
        _colour,
        _zomma,
        _vomma,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: x[i] >= {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i] <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: s >= {:s} and s <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i] >= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: r >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        _delta_cls.refresh_py(_delta)
        _gamma_cls.refresh_py(_gamma)
        _vega_cls.refresh_py(_vega)
        _theta_cls.refresh_py(_theta)
        _rho_cls.refresh_py(_rho)
        _crho_cls.refresh_py(_crho)
        _vanna_cls.refresh_py(_vanna)
        _charm_cls.refresh_py(_charm)
        _speed_cls.refresh_py(_speed)
        _colour_cls.refresh_py(_colour)
        _zomma_cls.refresh_py(_zomma)
        _vomma_cls.refresh_py(_vomma)
        return_data = None
    _err_class.manage()
    return return_data
